/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     VarFactory_i
// @module  VarFactory_i | This is the implementation of the interface to
//  the TestManager Variable Factory for Global and Script type scope variables
//
///////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
#include "GlobalVariable_i.h"
#include "VarFactory_i.h"
#include "CECVarOptions.h"
#include "mdbexcept.h"
#include "mdbparse.h"
#include "TPLog.h"
#include <mutex.hpp>
#include "CECInetInfo.h"

const string TESTWIDEDOT="TEST-WIDE.";
const string SCRIPTDOT="SCRIPT.";
const string GLOBALDOT="GLOBAL.";
#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif

#include <iostream>
using namespace std;
///////////////////////////////////////////////////////////////////////////////
//
// @class   VarFactoryImpl
//
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor
// @parm    CORBA server Object
// @parm    Name to use in registration with name service
//      This should be derived from the name of the TestRun object associated
//      with the TestManager that creates this Variable Factory
VarFactoryImpl::VarFactoryImpl(CORBAServer* pServer,const string &regname)
{
    ConstructInit(pServer,regname);
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Used by constructors for common functionality
// @parm    CORBA server Object
// @parm    Name to use in registration with name service
//      This should be derived from the name of the TestRun object associated
//      with the TestManager that creates this Variable Factory
void VarFactoryImpl::ConstructInit(CORBAServer* pServer,const string &regname)
{
    m_pServer = pServer;
    m_regname = regname;
    // Register object with NameService
    m_pServer->ActivateObject(this);   
    m_pServer->NameObject(this->_this(), VarFactory::CLASS_NAME, m_regname.c_str());   
    m_RegisteredScriptList.clear();
    m_mutexlist.clear();
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Destructor
VarFactoryImpl::~VarFactoryImpl()
{
    m_pServer->UnNameObject(VarFactory::CLASS_NAME, m_regname.c_str() );   
    ClearMutexList();
    ClearSemaphoreList();
    ClearRegisteredScriptList();
    ClearGlobalVarList();
    ClearScriptVarList();
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   RegisterScript - Register a script for use invariable creation
// @parm    script to register
CORBA::Long  VarFactoryImpl::RegisterScript(const char*  scriptname)
{
    string script=scriptname;
    ScriptMap_t::iterator i=m_RegisteredScriptList.find(script);
    // if not the first instance of this scriptname add to previous map
    if(i!=m_RegisteredScriptList.end())
    {
        (*i).second++;
    }
    else
    {
        //set number of scripts with this name to 1
        m_RegisteredScriptList.insert(ScriptMap_t::value_type(script,1));
    }
    return 0;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Get a Variable, create and/or return pointer to new CORBA object
//      as defined by parameters
//      This method is a bit long at the moment , should be broken down
// @parm    simple variable name
// @parm    size of variable
// @parm    file name for retreiving variable values from
// @parm    file id for above file
// @parm    dimensions for arrays
// @parm    values for lists
// @parm    option bitmask
// @parm    seed for repeatable random variables
// @parm    scriptname that this variable is for
// @rdesc   CORBA ptr to new variable object
CORBA::Object_ptr  VarFactoryImpl::GetVariable(const char*  name,
        CORBA::Long  size,
        const char*  file,
        const char*  fileid,
        const char*  dimensions,
        const char*  values,
        CORBA::Long  options,
        CORBA::Long  seed,
        const char*  scriptname)
{
	cout<<"In GetVariable"<<endl;

    //Create a Variable declaration object
	cout<<"Create a Variable declaration object"<<endl;
    VFVarDecl VFVDcl(name,size,file,fileid,dimensions,values,options,seed,scriptname);
	cout<<"Created a Variable declaration object"<<endl;
    // return object
    CORBA::Object_var obj=CORBA::Object::_nil();
    //Check if Global type or Script
    if( options & VCGLOBAL )
    {
        //Check if in Global List
        VFVarMap_t::iterator i;
        i=m_GlobalVarList.find(VFVDcl.GetUniqueName());
        if(i!=m_GlobalVarList.end())
        {
            //if we already have that variable
            //check declarations match
            bool match=(*i).second->compare(VFVDcl);
            if(match)
            {
                obj=(*i).second->GetVariable_ptr();
            }
            else
            {
                //throw some horrible exception as there is conflict between these 2 variables
                throw mdbParseException("Variable declaration mismatch");
            }
        }
        else
        {
            //make a new variable and add to map
            VFVar* NewVar=new VFVar(m_pServer,m_regname,VFVDcl);
            m_GlobalVarList.insert(VFVarMap_t::value_type(VFVDcl.GetUniqueName(), NewVar));
            obj=NewVar->GetVariable_ptr();
        }
    }
    else if (options & VCSCRIPT )
    {

        VFVarMapMap_t::iterator iter;
        //Check for existence of map of variables for this scriptname
        iter=m_ScriptVarList.find(scriptname);
        VFVarMap_t* this_script_list;//pointer to map of variables for this script
        if(iter!=m_ScriptVarList.end())
        {
            this_script_list=(*iter).second;
            VFVarMap_t::iterator   viter;
            //Check for existence of required variable within this map
            viter=this_script_list->find(VFVDcl.GetUniqueName());
            if(viter!=this_script_list->end())
            {
                //If we have this variable already then set returned
                //variable pointer to this variable
                // no need to check declaration as
                // script variables by their definition can only be declared in one way
                //(O.K if someone edits a script part way through a test things could go wrong)
                obj=(*viter).second->GetVariable_ptr();
            }
            else
            {
                //otherwise make a new variable and insert it into this script map
                //make a new variable and add to map
                VFVar* NewVar=new VFVar(m_pServer,m_regname,VFVDcl);
                this_script_list->insert(VFVarMap_t::value_type(VFVDcl.GetUniqueName(), NewVar));
                obj=NewVar->GetVariable_ptr();
            }

        }
        else//No script wide variables have been created for this script yet so
            //create new map
        {
			cout<<"Create VFVar variable object"<<endl;
            this_script_list=new VFVarMap_t;
            VFVar* NewVar=new VFVar(m_pServer,m_regname,VFVDcl);
            this_script_list->insert(VFVarMap_t::value_type(VFVDcl.GetUniqueName(), NewVar));
            obj=NewVar->GetVariable_ptr();
            m_ScriptVarList[scriptname]=this_script_list;
			cout<<"Created variable."<<endl;
        }
    }


    return CORBA::Object::_duplicate(obj);
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Create a new CORBA mutex
// @parm    name that mutex is to register with
// @rdesc   Object_ptr to mutex object (narrow will need to be done anyway)
CORBA::Object_ptr  VarFactoryImpl::CreateMutex(const char*  name)
{
    TPMutexImpl* mutex = new TPMutexImpl(m_pServer,string(name));
    m_mutexlist.push_back(mutex);
    return CORBA::Object::_duplicate(mutex->_this());    
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Create a new CORBA Simple Semaphore (flag)
// @parm    name that Semaphore is to register with
// @rdesc   Object_ptr to Semaphore object (narrow will need to be done anyway)
CORBA::Object_ptr  VarFactoryImpl::CreateSimpleSemaphore(const char*  name)
{
    TPCYRSemaphore* Semaphore = new TPCYRSemaphore(m_pServer,string(name));
    m_simplesemaphorelist.push_back(Semaphore);
    return CORBA::Object::_duplicate(Semaphore->_this());   
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Create a new CORBA Semphore
// @parm    name that Semaphore is to register with
// @rdesc   Object_ptr to Semaphore object (narrow will need to be done anyway)
CORBA::Object_ptr  VarFactoryImpl::CreateRealSemaphore(const char*  name)
{
    TPSemaphoreImpl* Semaphore = new TPSemaphoreImpl(m_pServer,string(name));
    m_realsemaphorelist.push_back(Semaphore);
    return CORBA::Object::_duplicate(Semaphore->_this());   
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   DeregisterScript
// @parm    The script to deregister
// @parm    is Keep-Alive option set true/false
// @rdesc   -1 for failure, more likely to throw a CORBA::OBJECT_NOT_EXIST
CORBA::Long  VarFactoryImpl::DeregisterScript(const char*  scriptname, CORBA::Boolean  keepalive)
{
    static CTMutex mutex;
    CTLockGuard<CTMutex> guard(mutex);
    string script=scriptname;
    CORBA::Long  status=-1;
    ScriptMap_t::iterator i=m_RegisteredScriptList.find(script);
    // check that it is a valid script name
    if(i!=m_RegisteredScriptList.end())
    {
        if(!keepalive)status=DisposeVariables(script);
    }
    else
    {
        throw CORBA::OBJECT_NOT_EXIST();// seems fair enough
    }

    return status;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Check to see if it is possible to dispose of variables for a scriptname
//      Checks all ScriptObjects Still pointed to to see if they exist
//      If no valid scriptobjects left then dispose of all variables associated
//      with that scriptname.
// @parm    The script to dispose of variables for
// @rdesc   -1 for failure
CORBA::Long  VarFactoryImpl::DisposeVariables(const string& scriptname)
{
    int valid_count=0;
    string lmsg="VarFactoryImpl::DisposeVariables - "+scriptname;
    logr->AddTrace(DBGSTD,DETAILLEVEL,lmsg.c_str());
    ScriptMap_t::iterator map_iter=m_RegisteredScriptList.find(scriptname);
    if(map_iter!=m_RegisteredScriptList.end())
    {
        valid_count=(*map_iter).second;
    }
    if(valid_count==0)//no valid scripts running
    {
        string msg="No Scripts called "+scriptname+" running";
        logr->AddTrace(DBGSTD,DETAILLEVEL,msg.c_str());
        VFVarMapMap_t::iterator i;
        i=m_ScriptVarList.find(scriptname);
        if(i!=m_ScriptVarList.end())//have some variable for this script
        {
            VFVarMap_t::iterator j;
            // for each variable pointer for this script dispose of variable
            for(j=(*i).second->begin();j!=(*i).second->end();j++)
            {
                delete (*j).second;
            }
            (*i).second->clear();
        }
    }
    return 0;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Clear down all ScriptObjects Registered with this factory
//      (No longer regsitering CORBA objects so not required)
//
void    VarFactoryImpl::ClearRegisteredScriptList()
{
    m_RegisteredScriptList.clear();
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Clear down all Global Variables for this factory
//
void VarFactoryImpl::ClearGlobalVarList()
{
    VFVarMap_t::iterator i;
    for(i=m_GlobalVarList.begin();i!=m_GlobalVarList.end();i++)
    {
        delete (*i).second;
    }
    m_GlobalVarList.clear();
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Clear down all Script Variables for this factory
//
void VarFactoryImpl::ClearScriptVarList()
{
    VFVarMapMap_t::iterator i;
    for(i=m_ScriptVarList.begin();i!=m_ScriptVarList.end();i++)
    {
        VFVarMap_t::iterator j;
        for(j=(*i).second->begin();j!=(*i).second->end();j++)
        {
            delete (*j).second;
        }
        (*i).second->clear();
        delete (*i).second;
    }
    m_ScriptVarList.clear();
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Clear down all Mutexes for this factory
//
void    VarFactoryImpl::ClearMutexList()
{
    mutexlist_t::iterator i;
    for(i=m_mutexlist.begin();i!=m_mutexlist.end();i++)
    {
        try
        {
            (*i)->_remove_ref();    
        }
        catch(...)
        {
            //Do nothing just go to next mutex
        }

    }
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Clear down all Semaphores for this factory
//
void    VarFactoryImpl::ClearSemaphoreList()
{
    simplesemaphorelist_t::iterator i;
    for(i=m_simplesemaphorelist.begin();i!=m_simplesemaphorelist.end();i++)
    {
        try
        {
            (*i)->_remove_ref();    
        }
        catch(...)
        {
            //Do nothing just go to next Semaphore
        }

    }
    m_simplesemaphorelist.clear();
    realsemaphorelist_t::iterator it;
    for(it=m_realsemaphorelist.begin();it!=m_realsemaphorelist.end();it++)
    {
        try
        {
            (*it)->_remove_ref();
        }
        catch(...)
        {
            //Do nothing just go to next Semaphore
        }

    }
    m_realsemaphorelist.clear();
}
///////////////////////////////////////////////////////////////////////////////
//
// @class   VarFactoryImpl::VFVarDecl, Details of A CORBA variable derived from the declaration
//
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// @mfunc Constructor, assign member variables and create unique name for variable described by
// members
VarFactoryImpl::VFVarDecl::VFVarDecl( const char*  name,
        CORBA::Long  size,
        const char*  file,
        const char*  fileid,
        const char*  dimensions,
        const char*  values,
        CORBA::Long  options,
        CORBA::Long  seed,
        const char*  scriptname)
        :m_name(name),m_size(size),m_file(file),m_fileid(fileid),m_dimensions(dimensions),
        m_values(values),m_options(options),m_seed(seed),m_scriptname(scriptname)
{
    m_unique_name="";
    if(options & VCTESTWIDE)m_unique_name+=TESTWIDEDOT;
    if(options & VCSCRIPT)
    {
        m_unique_name+=SCRIPTDOT;
    }
    else if (options & VCGLOBAL )
    {
        m_unique_name+=GLOBALDOT;
    }
    else
    {
        // throw some bad exception, should only be dealing with script and global scope
        throw mdbParseException("illegal Variable declaration");
    }
    m_unique_name+=m_name;
    // if we are a script variable  then use the name else blank it out
    if(options & VCSCRIPT)m_unique_name+=( "." + m_scriptname );
    else (m_scriptname="");
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Destructor
VarFactoryImpl::VFVarDecl::~VFVarDecl()
{}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Return the Unique name of variable for this declaration
//      If 2 declarations share the same name then they should be the same variable
//      more checks can then be performed to see if the other details of a variable match

string VarFactoryImpl::VFVarDecl::GetUniqueName()
{
    return m_unique_name;
}
///////////////////////////////////////////////////////////////////////////////
// @func    compare VFVarDecl with 'this'
// @rdesc   return true if the same
// @parm    VFVarDecl to compare to
bool VarFactoryImpl::VFVarDecl::compare(const VFVarDecl& rhs)
{
    if((m_name==rhs.m_name)&&
            (m_size==rhs.m_size)&&
            (m_file==rhs.m_file)&&
            (m_fileid==rhs.m_fileid)&&
            (m_dimensions==rhs.m_dimensions)&&
            (m_values==rhs.m_values)&&
            (m_options==rhs.m_options)&&
            (m_seed==rhs.m_seed)&&
            (m_scriptname==rhs.m_scriptname))return true;
    else return false;
}
///////////////////////////////////////////////////////////////////////////////
// @func    operator overloads for VFVarDecl
bool VarFactoryImpl::VFVarDecl::operator==(const VFVarDecl& rhs)
{
    return compare(rhs);
}
bool VarFactoryImpl::VFVarDecl::operator!=(const VFVarDecl& rhs)
{
    return !(compare(rhs));
}
///////////////////////////////////////////////////////////////////////////////
//
// @class   VarFactoryImpl::VFVar, Details and instance of A CORBA variable derived from the declaration
//
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor
VarFactoryImpl::VFVar::VFVar( CORBAServer* pServer,
                                    string VFRegName,
                                    const char*  name,
                                    CORBA::Long  size,
                                    const char*  file,
                                    const char*  fileid,
                                    const char*  dimensions,
                                    const char*  values,
                                    CORBA::Long  options,
                                    CORBA::Long  seed,
                                    const char*  scriptname)
        :VFVarDecl(name,size,file,fileid,dimensions,
                     values,options,seed,scriptname)
{
    CreateVariable(pServer,VFRegName);
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor
VarFactoryImpl::VFVar::VFVar(CORBAServer* pServer ,string VFRegName,const VFVarDecl& VFVD)
        :VFVarDecl(VFVD)
{
    CreateVariable(pServer,VFRegName);
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Destructor
VarFactoryImpl::VFVar::~VFVar()
{
    m_CORBAvariable->_remove_ref(); 
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Return the CORBA variable contained within class
// @rdesc   CORBA Variable
CORBA::Object_ptr VarFactoryImpl::VFVar::GetVariable_ptr()
{
    if (CORBA::is_nil(m_CORBAvariable->_this()))
        return CORBA::Object::_nil();
    else
        return CORBA::Object::_duplicate(m_CORBAvariable->_this());  
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Create the CORBA variable for this class based upon decalration
// @parm    A CORBAServer so that the Variables can be registered
// @parm    The Name that This TestManager Variable Factory is using to register itself.
//      This will have appended the Unique name of the variable and will be used by
//      the variables to register themselves with the nameservice
void VarFactoryImpl::VFVar::CreateVariable(CORBAServer* pServer,string VFRegName)
{
    m_pServer=pServer;
    m_regname=VFRegName+"."+GetUniqueName();

    if(m_options & VSLIST)
    {
        //Value List variable
        m_CORBAvariable=CreateNewListVariable();
    }

    else if(m_options & VSFILE)
    {
        //File variable
        m_CORBAvariable=CreateNewFileVariable();
    }

    else if(m_options & VSRANGE)
    {
        //Range Variable
        m_CORBAvariable=CreateNewRangeVariable();
    }
    else if(m_dimensions!="")
    {
        //Array Variable
        m_CORBAvariable=CreateNewArrayVariable();
    }
    else
    {

        //Simple Variable
        m_CORBAvariable=CreateNewSimpleVariable();

    }

}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   Create and initialise a List CORBA variable
GlobalVariableImpl* VarFactoryImpl::VFVar::CreateNewListVariable()  
{
    GlobalListVarImpl* Var=new GlobalListVarImpl(m_pServer,m_name,m_size,m_options,m_values,m_seed);
    return Var; 
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Create and initialise a File CORBA variable
GlobalVariableImpl* VarFactoryImpl::VFVar::CreateNewFileVariable()  
{
    GlobalFileVarImpl* Var = new GlobalFileVarImpl(m_pServer,m_name,m_size,m_options,m_file);
    return Var; 
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Create and initialise a Array CORBA variable
GlobalVariableImpl* VarFactoryImpl::VFVar::CreateNewArrayVariable() 
{
    GlobalArrayVarImpl* Var = new GlobalArrayVarImpl(m_pServer,m_name,m_size,m_options,m_dimensions);
    return Var; 
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Create and initialise a Simple CORBA variable
GlobalVariableImpl* VarFactoryImpl::VFVar::CreateNewSimpleVariable()
{
    GlobalSimpleVarImpl* Var = new GlobalSimpleVarImpl(m_pServer,m_name,m_size,m_options);
    return Var; 
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Create and initialise a Range CORBA variable
GlobalVariableImpl* VarFactoryImpl::VFVar::CreateNewRangeVariable() 
{
    GlobalRangeVarImpl* Var=new GlobalRangeVarImpl(m_pServer,m_name,m_options,m_values,m_seed);
    return Var; 
}

void VarFactoryImpl::ClearAll()
{
    ClearMutexList();
    ClearSemaphoreList();
    ClearRegisteredScriptList();
    ClearGlobalVarList();
    ClearScriptVarList();
}
