/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrWrapLib
// @module  File Wrapper | This module contains the implementation of
//          the CECOrbFile wrapper class.  This class simply wraps around the
//          CORBA interface CyrStFile and hides the CORBA code.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CyrWrapLib/CECOrbFile.h>
#include <Repository/CyrStFileIt_i.h>
#include <Factory/ParamList_i.h>
#include <fstream>
using std::fstream;
///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor retrieves a File from the OpenSTARep database, based
//          on its name and its parent Folder.
// @parm    Pointer to the parent Folder object.
// @parm    The name of the File to retrieve.
CECOrbFile::CECOrbFile( CECOrbFolder* folder, const char* name )
        : CECOrbLink( folder->GetHelper(), folder->GetFactory() ),
        m_cacheName(name)
{
    // Error information
    const char* where = "CECOrbFile::CECOrbFile";

    try
    {
        // Retrieve underlying CyrStServer reference from wrapper class
        m_hidden = folder->GetHiddenServer();
        m_folder = folder;

        // Create a new parameter list for File.  Remember that all returned
        // strings must be released so we copy them into a CORBA::String_var
        // variables before passing them as parameters.  This will cause them
        // to be released as soon as the variable goes out of scope.
        FactoryParamList* params =
            new FactoryParamList( CyrStFile::CLASS_NAME, name, DEFAULT_PERIOD );
        PortableServer::ObjectId* id = m_corba->ActivateObject( params );

        string paramFolder = folder->GetRegisteredName();
        params->AddValue( CyrStFile::PARAM_FOLDER, paramFolder.c_str() );

        // Create CORBA object in Factory
        CORBA::Object_ptr obj = m_factory->New( params->_this() );
        m_servant = CyrStFile::_narrow( obj );

        // Register name so that object can be borrowed
        CORBA::String_var regName = m_servant->GetRegisteredName();
        Register( regName, CyrStFile::CLASS_NAME );

        // Destroy parameter object
        m_corba->DeActivateObject(*id);
        params->_remove_ref();
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates a new File object in the specified Folder.
// @parm    Pointer to the parent Folder.
// @parm    The name of the File to create.
// @parm    The description of the File to create.
CECOrbFile::CECOrbFile( CECOrbFolder* folder, const char* name, const char* desc )
        : CECOrbLink( folder->GetHelper(), folder->GetFactory() ),
        m_cacheName(name)
{

    // Error information
    const char* where = "CECOrbFile::CECOrbFile";

    try
    {
        // Retrieve underlying CyrStServer reference from wrapper class
        m_hidden = folder->GetHiddenServer();
        m_folder = folder;

        // Create a new parameter list for File.  Remember that all returned
        // strings must be released so we copy them into a CORBA::String_var
        // variables before passing them as parameters.  This will cause them
        // to be released as soon as the variable goes out of scope.
        FactoryParamList* params =
            new FactoryParamList( CyrStFile::CLASS_NAME, name, DEFAULT_PERIOD );
        PortableServer::ObjectId* id = m_corba->ActivateObject( params );

        string paramFolder = folder->GetRegisteredName();
        params->AddValue( CyrStFile::PARAM_FOLDER, paramFolder.c_str() );
        params->AddValue( CyrStFile::PARAM_DESC, desc );

        // Create CORBA object in Factory
        CORBA::Object_ptr obj = m_factory->New( params->_this() );
        m_servant = CyrStFile::_narrow( obj );

        // Register name so that object can be borrowed
        CORBA::String_var regName = m_servant->GetRegisteredName();
        Register( regName, CyrStFile::CLASS_NAME );

        // Destroy parameter object and file iterator
        m_corba->DeActivateObject(*id);
        params->_remove_ref();
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the OpenSTARep File.  This is the
//          same name that was passed into the constructors, but is repeated
//          here for completeness.
// @rdesc   The name string.
string CECOrbFile::GetName() const
{
    // Call underlying CORBA object's method, but store result in
    // a CORBA string so that memory is handled correctly.
    //CORBA::String_var s = m_servant->GetName();

    // Convert the CORBA string into something else so that the CORBA
    // mechanism remains hidden.
    //string ret( s );
    //return ret;
    // return cached name from constructor rather than use method calls
    return m_cacheName;
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the current status of the file in question.
// @rdesc   The current status.
long CECOrbFile::GetCurrentStatus() const
{
    return m_servant->GetCurrentStatus();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a copy of the underlying CORBA servant object
//          to the calling code.  This will increment the reference count and
//          so will need to be released.
// @rdesc   The reference to the servant object.
CyrStFile_ptr CECOrbFile::GetServant()
{
    return CyrStFile::_duplicate( m_servant );
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   Returns a pointer to the server that spawned this object.
// @rdesc   A server pointer.
CECOrbServer* CECOrbFile::GetParentServer()
{
    return GetParentProject()->GetParentServer();
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   Returns a pointer to the project that spawned this object.
// @rdesc   A project pointer.
CECOrbProject* CECOrbFile::GetParentProject()
{
    return GetParentFolder()->GetParentProject();
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   Returns a pointer to the folder that spawned this object.
// @rdesc   A folder pointer.
CECOrbFolder* CECOrbFile::GetParentFolder()
{
    return m_folder;
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method tranfers the entire contents of a file to the remote
//          OpenSTARep database.
// @parm    Fully specified path to the file to check in.
// @parm    If true will place a lock on the file.  if false will leave lock
//          status unchanged.
void CECOrbFile::CheckIn( const char* sourcePath, bool lock )
{
    // Create file iterator
    CyrStFileItImpl* fileIt = new CyrStFileItImpl( sourcePath );
    PortableServer::ObjectId* id = m_corba->ActivateObject( fileIt );

    // Copy file to remote machine
    m_servant->PutFile( fileIt->_this(), lock );

    // dispose of file iterator
    m_corba->DeActivateObject(*id);
    fileIt->_remove_ref();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method transfers the entire contents of the File object to
//          the specified location.  Note that the directory to the file MUST
//          exist as this method does not create directories.
// @parm    The full path of the file to create.
// @parm    If true the checked out file is locked exclusively by this user,
//          otherwise the lock status is unchanged.
void CECOrbFile::CheckOut( const char* destPath, bool lock )
{
    // Get File iterator
    CyrStByteIt_var fileIt = m_servant->GetFile( lock );

    // Open new file
    fstream* f = new fstream( destPath, std::ios::binary|std::ios::out );

    // Loop through iterator
    ByteSeq_var byteSeq = fileIt->ReadBytes();
    while ( byteSeq->length() != 0 )
    {
        // Write contents to file and get next sequence
        f->write((const char*) byteSeq->get_buffer(), byteSeq->length() );
        byteSeq = fileIt->ReadBytes();
    }

    // Close file and flush contents
    delete f;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method locks the remote File so that no other user can modify
//          it.  This method should be called before updating any changes to
//          the File.
// @rdesc   Returns true if the lock request was successful, otherwise false.
bool CECOrbFile::Lock()
{
    return m_servant->Lock();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method removes a lock placed previously by this user.
void CECOrbFile::Unlock()
{
    m_servant->Unlock();
}

