/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_SPAWN                   */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains the PSL_SPAWN routine.              */
/*                                       */
/*===========================================================================*/

/*    Define useful Include structures   */

#include <psl_macros.h>         /* PSL Standard Macros */
#include <stddef.h>         /* Define NULL */
#include <string.h>         /* String Routines */

/*    Define Function Prototypes for all external routines */
/*    used by the module */

#include  psl_common_file(psl_rtndef.h) /* PSL Routine Function Prototypes */

#ifdef  VMS
#pragma nostandard          /* Keep Alpha Happy! */
#include <lib$routines.h>       /* LIB$SPAWN */
#pragma standard
#elif UNIX
#include <pwd.h>            /* Password file - for getpwuid() */
#include <errno.h>          /* Errors */
#include <fcntl.h>
#include <stdlib.h>         /* getenv and such-like */
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>          /* System types for getxxx() */
#include <sys/wait.h>           /* For wait() - Wait for child */
#include <signal.h>         /* For kill() - send signal */
#endif  /* VMS */

/*   Define Data Structures */

#include psl_common_file(pslspn_prmdef.h)
/* Symbolic Parameters */
#include psl_common_file(shrmsg.h)  /* Message Statuses */

#ifdef VMS
#pragma nostandard          /* Keep Alpha Happy! */
#include <descrip.h>            /* VAX Descriptor Format */
#pragma standard
#endif


/************************************************************************/
/*                                  */
/*  PSL_SPAWN - Spawn a DCL/shell command               */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_SPAWN (diag_ptr, cmd_ptr, flags, outfil_ptr); */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long PSL Return Status         */
/*     diag_ptr   = Diagnostics Stack Pointer           */
/*     cmd_ptr    = Pointer to null-terminated command line     */
/*     flags      = Unsigned Long Control Flags:            */
/*          PSLSPN_IGNSTS => ignore status of spawned image */
/*     outfil_ptr = Pointer to buffer containg name of file in  */
/*          which output is to be trapped           */
/*                                  */
/*  This routine is called to Spawn a DCL/shell command (or just    */
/*  to create a spawned sub/child process).             */
/*                                  */
/************************************************************************/

unsigned long PSL_SPAWN (struct diagstk *diag_ptr,
                         const char *cmd_ptr,
                         unsigned long flags,
                         const char *outfil_ptr)
{

#ifndef WIN32

    /*   Local Data   */

    unsigned long       iret;       /* Routine Return Status */
    unsigned long       diag_save;      /* Saved diagnostic context */

    /*   Platform-specific Local Data  */

#ifdef  VMS
    struct   dsc$descriptor_s spawn_dsc;    /* Spawn Descriptor */
    struct   dsc$descriptor_s outfil_dsc;   /* Output file Descriptor */
    struct   dsc$descriptor_s *outfil_dsc_ptr;
    /* Pointer to Output file Descriptor */
#else   /* VMS */
    unsigned long       cmdlen;     /* Command length */
    long       userid;     /* Current user ID */
    pid_t      ret_pid;        /* PID Returned */
    pid_t      parent_pid;     /* PID of parent */
    struct   passwd     *passwd_ptr;    /* Pointer to Password Structure */
    char       *arg_ptr[4];    /* EXECL Arguments */
    char       *buf_ptr;       /* Pointer to scan command line */
    static   char       cmdlin[2048];   /* Copy of command line */
    unsigned long       argcnt;     /* Argument Count */
    long       status;     /* Status from UNIX functions */
    int        child_status;   /* Status returned from wait*/
    Boolean    exec_failed;    /* Child's execn() failed */
    char       err_msg[30];    /* Error message */
    char       shell[256];     /* Name of shell */
    int        fds;        /* File descriptor */
    char       exp_file [256]; /* Expanded filespec */
    static   char       default_shell[] = "/bin/csh";
#endif  /* VMS */

    /*    First add a diagnostic to the stack */

    diag_save = PSL_ADD_STATDIAG (diag_ptr, "During PSL_SPAWN");

#ifdef  VMS
    /*   On VMS, just set up a descriptor for the command and possibly  */
    /*   the output trap file and issue the spawn               */

    spawn_dsc.dsc$b_class   = DSC$K_CLASS_S;
    /* Set up String Class */
    spawn_dsc.dsc$b_dtype   = DSC$K_DTYPE_T;
    /* and data type */
    spawn_dsc.dsc$a_pointer = (char *) cmd_ptr;
    /* and pointer */
    spawn_dsc.dsc$w_length  = strlen (spawn_dsc.dsc$a_pointer);
    /* and length */

    if (outfil_ptr != NULL)
    {
        outfil_dsc.dsc$b_class   = DSC$K_CLASS_S;
        /* Set up String Class */
        outfil_dsc.dsc$b_dtype   = DSC$K_DTYPE_T;
        /* and data type */
        outfil_dsc.dsc$a_pointer = (char *) outfil_ptr;
        /* and pointer */
        outfil_dsc.dsc$w_length  = strlen (outfil_dsc.dsc$a_pointer);
        /* and length */
        outfil_dsc_ptr = &outfil_dsc;
        /* Set up pointer to descriptor */
    }
    else
    {
        outfil_dsc_ptr = NULL;  /* Clear pointer to descriptor */
    }


    iret = lib$spawn (      /* Spawn the command */
               &spawn_dsc, /* Command string */
               NULL,       /* No input file */
               outfil_dsc_ptr, /* Output file */
               0,      /* No flags */
               NULL,       /* No process name */
               NULL,       /* No process id */
               NULL,       /* No completion status */
               NULL,       /* No event flag number */
               NULL,       /* No AST */
               0,      /* No AST parameter */
               NULL,       /* No prompt string */
               NULL);      /* No CLI */

    if (!psl_test_status(iret))
    {
        PSL_PUTDIAG (diag_ptr, 2, PSL_SYSERVERR, "lib$spawn", iret);

        goto end;
    }
    /* Exit if an error */
#else   /* VMS */

/*   Calculate length of command line */

    cmdlen = strlen (cmd_ptr);  /* Get length of string */

    /*   Initialize file descriptor so that we can drop out at any time */

    fds = -1;

    /*   In UNIX we first need to get the name of the users shell.      */
    /*   To locate this we first try to decode the environment variable */
    /*   SHELL; failing that we try to get a password information       */
    /*   structure for this user; if all else fails we default to       */
    /*   /bin/csh.                              */



    buf_ptr = getenv("SHELL");  /* Translate SHELL */
    if (buf_ptr == NULL)        /* If that failed */
    {
        userid = geteuid ();    /* Get effective user id */
        setpwent();         /* Reset password file */
        passwd_ptr = getpwuid (userid);
        /* Get password entry */
        endpwent ();        /* Close the password file */
        if (passwd_ptr != NULL) /* If it Succeeded */
        {
            buf_ptr = (*passwd_ptr).pw_shell;
            /* Set up to run shell */
        }
        else            /* Else (no password entry) */
        {
            buf_ptr = default_shell;
            /* Default to C-Shell */
        }               /* End of (if password entry) */
    }               /* End of (if environment var) */
    if  (strlen(buf_ptr) >= sizeof(shell))
        /* If shell name too long */
    {
        iret = PSL_BADPARAM;    /* Return an error */
        (void) PSL_ADD_STATDIAG (diag_ptr, "Shell name too long");
        /* Add diagnostic to stack */
        PSL_PUTDIAG (diag_ptr, 2, PSL_SYSERVERR, "FDR_SPAWN", iret);
        /* Log a PSL_SYSERVERR error */
        goto end;           /* And exit */
    }
    (void)strcpy(shell, buf_ptr);   /* Save name of shell */



    /*   Throw out commands that are too long               */

    if  (cmdlen >= sizeof(cmdlin))  /* If command line too long */
    {
        iret = PSL_BADPARAM;    /* Return an error */
        (void) PSL_ADD_STATDIAG (diag_ptr, "Command Line Too Long");
        /* Add diagnostic to stack */
        PSL_PUTDIAG (diag_ptr, 2, PSL_SYSERVERR, "FDR_SPAWN", iret);
        /* Log a PSL_SYSERVERR error */
        goto end;           /* And exit */
    }

    for (argcnt = 0; argcnt < 4; argcnt++)
    {
        arg_ptr[argcnt] = NULL; /* Initialize all pointers */
    }

    /*   Note that to execute a command in the context of a shell we    */
    /*   must ensure the first argument to the shell is "-c". The       */
    /*   command itself MUST be passed as a single argument to execn()  */
    /*   or the shell will consume later arguments.             */

    if (cmdlen != 0)        /* If there is a command line */
    {



        (void) strcpy (cmdlin, cmd_ptr);
        /* Copy into local buffer */
        arg_ptr[0] = shell;     /* Save shell as first argument */
        arg_ptr[1] = "-c";      /* shell -c (execute command) */
        arg_ptr[2] = cmdlin;    /* Command to execute */



    }               /* End of (if command line) */

    /*   If there is no command line, then we spawn the user's      */
    /*   preferred shell instead.                       */

    else                /* Else (no command line) */
    {
        arg_ptr[0] = shell;     /* Name of shell */
    }               /* End of (if command) */



    /*   If we have been asked to trap the output from the command in   */
    /*   a file then we must first re-direct stdout and stderr.     */

    if (outfil_ptr != NULL)
    {
        PSL_CLEARERR ();        /* Clear error numbers */

        iret = PSL_FILE_PARSE ( /* Parse filespec */
                   outfil_ptr, /* For this file */
                   NULL,       /* No default */
                   NULL,       /* No related */
                   NULL,       /* No expected related extension */
                   exp_file,   /* Put result here */
                   sizeof (exp_file),
                   /* Which is this big */
                   0);     /* Flags */
        if (!(psl_test_status (iret)))
            /* If it failed */
        {
            PSL_PUTDIAG (       /* Put diagnostic */
                diag_ptr,   /* Use this stack */
                2,      /* Number of message segments */
                PSL_OPENOUT,    /* First message */
                exp_file,   /* Parameter */
                iret);      /* Second message */
            iret = PSL_OPENOUT; /* Set up return status */
            goto end;       /* And exit */
        }

        /*   Create the file */

        fds = creat (exp_file, 0777);
        /* Create file for stdout */
        if (fds < 0)        /* If it failed */
        {
            PSL_LOG_CIOERR (    /* Log I/O error */
                diag_ptr,   /* Use this stack */
                PSL_OPENOUT,    /* Message */
                exp_file);  /* Parameter */
            iret = PSL_OPENOUT; /* Set up return status */
            goto end;       /* And exit */
        }

        /*   Get the fully expanded filespec of the file */

        PSL_GET_FDNAME (fds, exp_file, sizeof (exp_file));
    }


    /*   Next we need to FORK a process to EXECute the command. */

    parent_pid = getpid();      /* Get this processes PID */
    exec_failed = PFALSE;       /* Show child's execn() OK */
    ret_pid = fork ();      /* Fork a process */
    if (ret_pid < 0)        /* If it failed */
    {
        iret = PSL_DIAG_ERRNO(diag_ptr, "fork()");
        /* Output a diagnostic */
        goto end;           /* And exit */
    }

    /*   If we are the child process then we want to EXECute the        */
    /*   command (or the shell if no command was specified).        */

    if (ret_pid == 0)       /* If child process */
    {



        /*   If we have been asked to trap the output from the command in   */
        /*   a file then we must first re-direct stdout and stderr.     */

        if (outfil_ptr != NULL)
        {
            (void) close (1);   /* Close stdout */
            (void) close (2);   /* Close stderr */

            fds = dup2 (fds, 1);    /* Send stdout to file */
            if (fds < 0)        /* If it failed */
            {
                (void) kill(parent_pid, SIGUSR1);
                /* Send SIGUSR1 to parent */
                _exit(errno);   /* Terminate the child with error */
            }

            fds = dup2 (fds, 2);    /* Send stderr to file */
            if (fds < 0)        /* If it failed */
            {
                (void) kill(parent_pid, SIGUSR1);
                /* Send SIGUSR1 to parent */
                _exit(errno);   /* Terminate the child with error */
            }
        }


        /*   We use execl if the command starts with an explicit pathname   */
        /*   and execlp otherwise.  Note also that we replicate the first   */
        /*   argument as it acts both as image name and as command name to  */
        /*   the image.                             */

        if  (arg_ptr[0][1] == '/')  /* If an explicit pathname */
        {
            (void) execl (arg_ptr[0],
                          arg_ptr[0], arg_ptr[1],
                          arg_ptr[2], NULL);
            /* Execute the command via execl */
        }
        else            /* Else (no explicit pathname) */
        {
            (void) execlp (arg_ptr[0],
                           arg_ptr[0], arg_ptr[1],
                           arg_ptr[2], NULL);
            /* Execute the command via execpl */
        }

        /*   EXECL/EXECLP will terminate the (child) process if it      */
        /*   succeeds.  If we get here, therefore, we have an error and     */
        /*   must report it.  Note that we cannot signal an error from here */
        /*   as that will keep both child and parent process in the     */
        /*   utility. When the parent detects that the child has terminated */
        /*   and an error has occurred it does not know whether the execn() */
        /*   failed or the image EXECed returned an error. We help the      */
        /*   parent differentiate between the two by terminating its wait() */
        /*   with a SIGUSR signal. We then exit the child with the error    */
        /*   from execn() in _exit().                       */

        (void) kill(parent_pid, SIGUSR1);
        /* Send SIGUSR1 to parent */
        _exit(errno);       /* Terminate the child with error */
    }

    /*   If we get here then we are the parent process and just want to */
    /*   wait for the child to exit. However, two possibilities exist   */
    /*   at this point.                         */
    /*   [1] if the wait() is interrupted by the child sending a        */
    /*       SIGUSR1 signal then the childs call to execn() failed and  */
    /*   we perform another wait() on the child to obtain the       */
    /*   failure state of the execn() call.             */
    /*   [2] if the wait() is not interrupted then the child's execn()  */
    /*   succeeded and we just want the status from the EXECed      */
    /*   image.                             */
    /*   This method allows us to differentiate between the execn()     */
    /*   failing and the EXECed image returning a failure status.       */



wait_again:

    status = waitpid (ret_pid, &child_status, 0);
    /* Wait for child to complete */

    if ((status == -1) &&       /* Interrupted by signal from child */
            (errno == EINTR))
    {
        status = waitpid(ret_pid, &child_status, WNOHANG);
        /* Get reason for wait() failure */

        /*   If the child has not terminated (i.e. waitpid() returned       */
        /*   because we received a signal) then we ignore the signal and    */
        /*   wait again.                            */

        if (status == 0)        /* Child has not terminated */
        {
            goto wait_again;
        }



        /*  If the child has terminated then we assume the exec failed      */

        exec_failed = PTRUE;    /* Show execn() in child failed */
        (void) strcpy(err_msg, "wait()");
        /* Setup error message */
    }
    else
    {
        (void) strcpy(err_msg, "unknown() in child process");
    }

    /*   If the execn() succeeded and PSLSPN_IGNSTS is set then we      */
    /*   ignore the status returned from the EXECed image.          */

    if (!exec_failed && ((flags & PSLSPN_IGNSTS) != 0))
    {
        iret = PSL_NORMAL;
        goto skip_error;
    }

    if (status == -1)       /* If wait() failed report & exit */
    {
        iret = PSL_DIAG_ERRNO(diag_ptr, "wait()");
        goto end;
    }

    /*   Because child status is not a PSL failure status then we have  */
    /*   to go through the following convoluted macro tests to see what */
    /*   happened.                              */

    iret = PSL_NORMAL;      /* Assume child succeeded */

    /*   If child process terminated normally but returned an error     */
    /*   status then move the status into errno so PSL_DIAG_ERRNO can   */
    /*   report the problem.                        */

    if (WIFEXITED(child_status))    /* If child terminated normally */
    {
        if (WEXITSTATUS(child_status) != 0)
            /* ... but with an error status */
        {
            errno = WEXITSTATUS(child_status);
            /* Set errno from child status */
            iret = PSL_DIAG_ERRNO(diag_ptr, err_msg);
            /* Log a diagnostic */
            goto end;       /* And exit */
        }
    }

    /*   See if child terminated because it received a signal */

    else if (WIFSIGNALED(child_status))
        /* If failure was signal */
    {
        (void) PSL_ADD_DIAGNOSTIC(  /* Add a descriptive diagnostic */
            diag_ptr,
            "child process failed with errno %d",
            WTERMSIG(child_status));
#ifndef IBM
        if (WCOREDUMP(child_status)) /* If there was a core dump */
        {
            (void) PSL_ADD_DIAGNOSTIC(
                /* Say so */
                diag_ptr,
                "core dump produced");
        }
#endif
        PSL_PUTDIAG(        /* And add normal diagnostic */
            diag_ptr,
            1, PSL_SIGTERMPRC,
            WTERMSIG(child_status));
        iret = PSL_SYSERVERR;   /* Set failure status */
        goto end;           /* And Exit */
    }

    /*   See if child process was stopped */

    else if (WIFSTOPPED(child_status))
        /* If child was stopped */
    {
        (void) PSL_ADD_DIAGNOSTIC(
            diag_ptr,       /* Add a descriptive diagnostic */
            "child process failed with errno %d",
            WSTOPSIG(child_status));
        PSL_PUTDIAG(        /* And normal diagnostic */
            diag_ptr,
            1, PSL_SIGTERMPRC,
            WSTOPSIG(child_status));
        iret = PSL_SYSERVERR;   /* Set failure status */
        goto end;           /* And Exit */
    }

    /*   This label is branched to if the child process successfully    */
    /*   EXECed the specified image AND the PSLSPN_IGNSTS flag is set.  */

skip_error:

#endif  /* VMS */

    /*   All OK - Remove diagnostic and return success. */

    PSL_REMOVE_DIAGNOSTICS (diag_ptr, diag_save);
    /* Remove Diagnostic */
    iret = PSL_NORMAL;      /* And set Success Status */

    /*   END: */

    /*   At this point IRET will contain the final return status. */

end:

#ifdef UNIX
    if (fds != -1)          /* If file open */
    {
        (void) close (fds);      /* Close file */
    }

#endif /* UNIX */

    return iret;            /* Return Status to Caller */
#endif /*WIN32*/

#ifdef WIN32
    return PSL_NORMAL;
#endif
}





/************************************************************************/
/*                                  */
/*  PSLOLD_SPAWN - Spawn a DCL/shell command            */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSLOLD_SPAWN (diag_ptr, cmd_ptr, flags);      */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret     = Unsigned Long PSL Return Status           */
/*     diag_ptr = Diagnostics Stack Pointer             */
/*     cmd_ptr  = Pointer to null-terminated command line       */
/*     flags    = Unsigned Long Control Flags:          */
/*            PSLSPN_IGNSTS => ignore status of spawned image   */
/*                                  */
/*  This routine is called to Spawn a DCL/shell command (or just    */
/*  to create a spawned sub/child process).             */
/*                                  */
/*  It is a shell around PSL_SPAWN that has the old calling     */
/*  sequence.                           */
/*                                  */
/************************************************************************/

unsigned long PSLOLD_SPAWN (struct diagstk *diag_ptr,
                            const char *cmd_ptr,
                            unsigned long flags)
{
    return PSL_SPAWN (diag_ptr, cmd_ptr, flags, NULL);
}


