/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_VM_ROUTINES                 */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module is a direct translation of the origional PSL         */
/*  VM routines.                                 */
/*                                       */
/*  It contains a number of routines used for manipulating a         */
/*  linked list of fixed length datastructures allocated out         */
/*  of virtual memory.                           */
/*                                       */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <string.h>
#include <stdlib.h>

/*
 *  Define Function Prototypes for all routines in this module.
 */
#include <psl_rtndef.h>         /* PSL Function Prototypes */

static void INSQUE (struct psl_vm_llb *element_ptr,
                    struct psl_vm_llb *before_ptr);

static void REMQUE (struct psl_vm_llb *element_ptr);

/*
 *  Structure Definitions. 
 */
#include <shrmsg.h>         /* PSL Messages */
#include <psl_vm_prmdef.h>      /* PSL_VM Routine Symbolic values */
#include <psl_vm_llbdef.h>      /* PSL_VM_LLB structure */


/************************************************************************/
/*                                  */
/*    INSQUE                                */
/*    ======                                */
/*                                  */
/*    This routine inserts a queue entry into the list BEFORE the   */
/*    specified position.                       */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      INSQUE (element_ptr, before_ptr);               */
/*                                  */
/*    Where:                                */
/*                                  */
/*  element_ptr = Pointer to the linked list element to be inserted */
/*                                  */
/*      before_ptr  = Pointer to the list element before which this     */
/*            element is to be inserted.            */
/*                                  */
/************************************************************************/

static void INSQUE (struct psl_vm_llb *element_ptr,
                    struct psl_vm_llb *before_ptr)

{
    /*
     *    Local data
     */
    struct   psl_vm_llb *temp_ptr;      /* temporary pointer to queue header */

    /*
     *    Set the forward pointer of the new element to point to the element
     *    we are inserting before.
     */
    (*element_ptr).fwdptr = before_ptr;

    /*
     *    Set the backward pointer of the new element to be the same as the
     *    current backward pointer of the element we are inserting before
     *    and point the forward pointer of the element which the new elements 
     *    backward pointer now points to, to point to the new element.
     */
    (*element_ptr).bckptr = (*before_ptr).bckptr;
    temp_ptr = (*element_ptr).bckptr;
    (*temp_ptr).fwdptr = element_ptr;

    /*
     *    Set the backward pointer of the item we are inserting before to point
     *    to the new element
     */
    (*before_ptr).bckptr = element_ptr;

    return;
}


/************************************************************************/
/*                                  */
/*    REMQUE                                */
/*    ======                                */
/*                                  */
/*    This routine removes a specified queue entry from the list.   */
/*    It also resets the associated header to nulls to prevent reuse.   */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      REMQUE (element_ptr);                       */
/*                                  */
/*    Where:                                */
/*                                  */
/*       element_ptr = Pointer to the list element to be removed    */
/*                                  */
/************************************************************************/

static void REMQUE (struct psl_vm_llb *element_ptr)
{
    /*
     *    Local data
     */
    struct   psl_vm_llb    *temp_ptr;   /* temp pointer to queue header */

    /*
     *    Set the forward pointer of the element to which this elements backwards
     *    pointer points, point to the element to which this elements forward
     *    pointer points.
     */
    temp_ptr = (*element_ptr).bckptr;
    (*temp_ptr).fwdptr = (*element_ptr).fwdptr;

    /*
     *    Set the backward pointer of the element to which this elements forward
     *    pointer points to point to the element to which this elements backward
     *    pointer points.
     */
    temp_ptr = (*element_ptr).fwdptr;
    (*temp_ptr).bckptr = (*element_ptr).bckptr;

    /*
     *    Over-write the header information to prevent accidental reuse
     */
    (void) memset (element_ptr, '\0', sizeof (struct psl_vm_llb));



    return;
}


/************************************************************************/
/*                                  */
/*    PSL_VM_INITIALIZE                         */
/*    =================                         */
/*                                  */
/*    This routine initializes the linked list queue header.        */
/*    It is called as follows:                      */
/*                                  */
/*      iret = PSL_VM_INITIALIZE (size, list_ptr);          */
/*                                  */
/*   Where:                             */
/*                                  */
/*  iret     = Unsigned long status value.              */
/*       = PSL_INSVIRMEM - If memory allocation failed      */
/*           = PSL_NORMAL - If success.                         */
/*                                  */
/*      size     = Unsigned long size of each block in bytes.       */
/*                                  */
/*      list_ptr = Variable to hold a pointer to the list header.   */
/*         Passed by reference.                 */
/*         This should be the address of a pointer to STRUCT    */
/*         psl_vm_llb.                      */
/*                                  */
/************************************************************************/

unsigned long PSL_VM_INITIALIZE (unsigned long size,
                                 struct psl_vm_llb **list_ptr)
{
    /*
     *    Local Data
     */
    static  struct  psl_vm_llb *vm_llb_ptr; /* Pointer to queue header */

    /*
     *    Calculate the size of and allocate memory for the queue header record
     */
    vm_llb_ptr = (struct psl_vm_llb *) malloc (sizeof (struct psl_vm_llb));
    /* Allocate a queue header  */
    if (vm_llb_ptr == NULL) /* ...if allocation failed  */
    {
        return PSL_INSVIRMEM;   /* ...return the exit status    */
    }

    /*
     *    Initialize the queue header and set the size of each entry in it.
     */
    (*vm_llb_ptr).blk_size = size;
    /* Define the size of each entry in list  */
    (*vm_llb_ptr).fwdptr = vm_llb_ptr;
    /* ...point fwd pointer to the list header */
    (*vm_llb_ptr).bckptr = vm_llb_ptr;
    /* ...and also point bck to list header */
    (*vm_llb_ptr).sanity_ptr = vm_llb_ptr;
    /* ...and also point sanity check to header */
    (*vm_llb_ptr).data_ptr = NULL;
    /* ...and clear data pointer */

    (*list_ptr) = vm_llb_ptr;   /* Return the linked list I.d. */

    /*
     *   ...and return success
     */
    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    PSL_VM_INSERT                         */
/*    =============                         */
/*                                  */
/*    This routine allocates a new entry and links it into the end  */
/*    of the list.                          */
/*                                  */
/*      iret = PSL_VM_INSERT (list_id, buff_ptr, pos_ptr,       */
/*                ctx_ptr, position);           */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret     = Unsigned long status value.              */
/*       = PSL_INSVIRMEM if MALLOC failed           */
/*                                  */
/*  list_ptr = Pointer to the linked list header returned       */
/*         by PSL_VM_INITIALIZE.                */
/*         This should be a pointer to STRUCT psl_vm_llb.   */
/*                                  */
/*      buff_ptr = Pointer to the entry to be added  to the list.       */
/*                 This must be SIZE bytes long.            */
/*                                  */
/*      pos_ptr  = Pointer to element before which we are to insert     */
/*                 This will have been returned in the CONTEXT      */
/*                 parameter of a previous call to this routine, or to  */
/*         PSL_VM_REMOVE or PSL_VM_COPY.            */
/*         If a NULL pointer is specified, POS_PTR will default */
/*         to insert at the END of the list.            */
/*                                  */
/*      ctx_ptr  = Variable to hold a pointer to the newly-inserted */
/*         item. This may be used to identify this item in a    */
/*         subsequent call to PSL_VM_INSERT, PSL_VM_REMOVE or   */
/*         PSL_VM_COPY. Passed by reference.            */
/*                                  */
/*  position = Unsigned long value to specify positioning       */
/*       = PSL_VM_POS_SPEC - Insert at position determined by   */
/*                   POS_PTR                */
/*       = PSL_VM_POS_BEG - Insert at beginning of list     */
/*       = PSL_VM_POS_END - Insert at end of list       */
/*                                  */
/************************************************************************/

unsigned long   PSL_VM_INSERT (struct psl_vm_llb *list_ptr,
                               char *buff_ptr,
                               char *pos_ptr,
                               char **ctx_ptr,
                               unsigned long position)
{
    /*
     *    Local data
     */
    struct   psl_vm_llb *inserted_ptr;  /* temporary pointer */
    struct   psl_vm_llb *before_ptr;    /* Points to element to insert before */
    struct   psl_vm_llb *copy_ptr;      /* Points to header of element  */
    void       *tmp_ptr;       /* Temporary pointer */
    unsigned long       block_size_bytes;   /* Size of block to allocate    */

    /*
     *    Check the sanity field in the linked list header to ensure that 
     *    this is a header. A further sanity check is to ensure that the size
     *    field in the list header is non-zero. This prevents us from specifying 
     *    a pointer to an element which is not the list header.
     */
    if ((list_ptr == NULL)
            ||  ((*list_ptr).sanity_ptr != list_ptr)
            ||  ((*list_ptr).blk_size == 0))
    {
        return PSL_BADPARAM;
    }

    /*
     *    In addition the buffer pointer must point to somewhere.
     */
    if (buff_ptr == NULL)
    {
        return PSL_BADPARAM;
    }

    /*
     *    Convert the POS_PTR to be a pointer to the header of the record rather
     *    than being a pointer to the data section
     */
    if (pos_ptr == NULL)
    {
        copy_ptr = list_ptr;        /* Insert before list header    */
    }
    else
    {
        tmp_ptr = pos_ptr - (sizeof (struct psl_vm_llb));
        copy_ptr = (struct psl_vm_llb *) tmp_ptr;
    }

    /*
     *    Determine from the position and POS_PTR fields exactly which item we are
     *    going to insert BEFORE.
     */
    if (position == PSL_VM_POS_SPEC)    /* Insert at pos_ptr        */
    {
        before_ptr = copy_ptr;      /* pick up pos_ptr      */
    }
    else if (position == PSL_VM_POS_BEG) /* Insert at beginning of list */
    {
        before_ptr = (*list_ptr).fwdptr; /* Pick up first element   */
    }
    else if (position == PSL_VM_POS_END) /* Insert at end of list   */
    {
        before_ptr = list_ptr;      /* Insert before list header    */
    }
    else
    {
        return PSL_BADPARAM;
    }

    /*
     *    Check the sanity field of the element to be positioned before. It must
     *    point to itself.
     */
    if ((*before_ptr).sanity_ptr != before_ptr)
    {
        return PSL_BADPARAM;
    }

    /*
     *    Allocate sufficient bytes for the new element, remembering that
     *    we require extra space for the header information.
     */
    block_size_bytes = (sizeof (struct psl_vm_llb)) + (*list_ptr).blk_size;
    tmp_ptr = malloc (block_size_bytes);
    /* Allocate space for record */
    if (tmp_ptr == NULL)    /* ...if allocation failed */
    {
        return PSL_INSVIRMEM;   /* ...return the exit status */
    }

    /*
     *    Set up the pointer to the data portion of the record.
     *    We add on the size of the header to the base address of the header to
     *    get the base address of the data.
     */
    inserted_ptr = (struct psl_vm_llb *) tmp_ptr;
    (*inserted_ptr).data_ptr = (char *) tmp_ptr + (sizeof (struct psl_vm_llb));


    /*
     *    Set the sanity checking field in the queue header to point to 
     *    the structure. We also ensure that the size field is clear as this
     *    identifies this as a list element rather than the header.
     */
    (*inserted_ptr).sanity_ptr = inserted_ptr;
    (*inserted_ptr).blk_size = 0;

    /*
     *    Return the context of the element we have just inserted to the user.
     */
    if (ctx_ptr != NULL)
    {
        (*ctx_ptr) = (*inserted_ptr).data_ptr;
    }

    /*
     *    Copy the buffer into the linked list structure and set the data pointer
     *    to point to it.
     */
    (void) memcpy ((*inserted_ptr).data_ptr, buff_ptr, (*list_ptr).blk_size);

    /*
     *    ...and insert this item into the linked list at the position determined
     *    by the position argument.
     */
    INSQUE (inserted_ptr, before_ptr);

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    PSL_VM_REMOVE                         */
/*    =============                         */
/*                                  */
/*    This routine removes an entry from the list.          */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = PSL_VM_REMOVE (list_ptr, buff_ptr,           */
/*              pos_ptr, ctx_ptr, position);        */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret     = Long Status Return                   */
/*           = PSL_BADPARAM if the parameter list is invalid    */
/*       = PSL_RECNOTFOU if no item found           */
/*                                  */
/*  list_ptr = Pointer to the linked list header returned       */
/*         by PSL_VM_INITIALIZE.                */
/*         This should be a pointer to STRUCT psl_vm_llb.   */
/*                                  */
/*      buff_ptr = Pointer to the entry removed from the list.          */
/*                 This must be SIZE bytes long and may be the NULL     */
/*         pointer if we do not want to return the information  */
/*                                  */
/*      pos_ptr  = Pointer to element which we are to remove.       */
/*                 This will have been returned in the CONTEXT      */
/*                 parameter of a previous call to this routine, or to  */
/*         PSL_VM_INSERT or PSL_VM_COPY.            */
/*                 If a NULL pointer is specified then this will default*/
/*         to the beginning of the list.                        */
/*                                  */
/*      ctx_ptr  = Variable to hold a pointer to the following      */
/*         item. This may be used to identify this item in a    */
/*         subsequent call to PSL_VM_INSERT, PSL_VM_REMOVE or   */
/*         PSL_VM_COPY. It may be the NULL pointer if we do not */
/*         want to return the context               */
/*                                  */
/*  position = Unsigned long value to specify positioning       */
/*       = PSL_VM_POS_SPEC - Remove from position determined by */
/*                   POS_PTR                */
/*       = PSL_VM_POS_BEG - Remove from beginning of list   */
/*       = PSL_VM_POS_END - Remove from end of list     */
/*                                  */
/************************************************************************/

unsigned long   PSL_VM_REMOVE (struct psl_vm_llb *list_ptr,
                               char *buff_ptr,
                               char *pos_ptr,
                               char **ctx_ptr,
                               unsigned long position)
{
    /*
     *    Local data
     */
    struct  psl_vm_llb  *remove_ptr;    /* Points to element to remove  */
    void        *tmp_ptr;       /* Temporary pointer */

    /*
     *    Check the sanity field in the linked list header to ensure that 
     *    this is a header.
     */
    if ((list_ptr == NULL)
            ||  ((*list_ptr).sanity_ptr != list_ptr)
            ||  ((*list_ptr).blk_size == 0))
    {
        return PSL_BADPARAM;
    }

    /*
     *    If a null pointer has been specified for POS_PTR then default this to
     *    the first element in the list.
     */
    if (pos_ptr == NULL)
    {
        pos_ptr = (char *) (*list_ptr).fwdptr + (sizeof (struct psl_vm_llb));

    }

    /*
     *    Determine from the POSITION and POS_PTR fields exactly which item we are
     *    going to remove
     */
    if (position == PSL_VM_POS_SPEC)    /* Remove at pos_ptr        */
    {
        tmp_ptr = pos_ptr - (sizeof (struct psl_vm_llb));
        remove_ptr = (struct psl_vm_llb *) tmp_ptr;
    }
    else if (position == PSL_VM_POS_BEG) /* Remove from begin of list   */
    {
        remove_ptr = (*list_ptr).fwdptr;
    }
    else if (position == PSL_VM_POS_END) /* Remove from end of list */
    {
        remove_ptr = (*list_ptr).bckptr;
    }
    else                /* Invalid Position     */
    {
        return PSL_BADPARAM;
    }

    /*
     *    Check the sanity field of the element to be removed. It must
     *    point to itself and have a size field which is 0.
     */
    if ((*remove_ptr).sanity_ptr != remove_ptr)
    {
        return PSL_BADPARAM;
    }
    else if ((*remove_ptr).blk_size != 0)
    {
        return PSL_RECNOTFOU;
    }

    /*
     *    Return the context of the next item in the list
     */
    if (ctx_ptr != NULL)        /* If we have asked for the context */
    {
        (*ctx_ptr) = NULL;      /* assume no next item */
        if (((*remove_ptr).fwdptr != list_ptr)
                &&   (ctx_ptr != NULL))     /* If next is not the queue header */
            /* ...and we have asked for context */
        {
            tmp_ptr = (*remove_ptr).fwdptr;
            /* Get pointer to start next record */
            (*ctx_ptr) = (char *) tmp_ptr + (sizeof (struct psl_vm_llb));
            /* ...add on size of header */
        }
    }

    /*
     *    ...also if requested return the contents of the buffer which was
     *    contained in the element
     */
    if (buff_ptr != NULL)
    {
        (void) memcpy (buff_ptr,
                       (*remove_ptr).data_ptr,
                       (*list_ptr).blk_size);
    }

    /*
     *    Remove the item from the linked list and free up the memory allocated
     */
    REMQUE (remove_ptr);
    free (remove_ptr);

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    PSL_VM_COPY                           */
/*    ===========                           */
/*                                  */
/*    This routine copies an entry into or out of the list      */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = PSL_VM_COPY (list_ptr, buff_ptr, pos_ptr, context, mode);*/
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret     = Unsigned long status return              */
/*           = PSL_BADPARAM if the parameter list is invalid    */
/*       = PSL_RECNOTFOU if no item found           */
/*                                  */
/*  list_ptr = Pointer to the linked list header returned       */
/*         by PSL_VM_INITIALIZE.                */
/*         This should be a pointer to STRUCT psl_vm_llb.   */
/*                                  */
/*      buff_ptr = Pointer to a buffer to contain the entry copied from */
/*         the list or holding the entry to be copied into the  */
/*                 list. This must be SIZE bytes long and may be the    */
/*         NULL pointer if we only want to return the context   */
/*         of an item in the list.              */
/*                                  */
/*      pos_ptr  = Pointer to element which we are to copy      */
/*                 This will have been returned in the CONTEXT      */
/*                 parameter of a previous call to this routine, or to  */
/*         PSL_VM_INSERT or PSL_VM_COPY.            */
/*         A null pointer is valid for all types.           */
/*                                  */
/*         If a null pointer is specified, POS_PTR will default */
/*         to the start of the list for PSL_VM_MODE_FIRST,  */
/*         PSL_VM_MODE_SPEC and PSL_VM_MODE_INTO, to        */
/*         the end of the list for PSL_VM_MODE_LAST and will    */
/*         cause the FIRST/LAST item to be picked up for        */
/*         PSL_VM_MODE_NEXT and PSL_VM_MODE_PREV respectively.  */
/*                                  */
/*      ctx_ptr  = Variable to hold a pointer to the copied item.   */
/*         This may be used to identify this item in a      */
/*         subsequent call to PSL_VM_INSERT, PSL_VM_REMOVE or   */
/*         PSL_VM_COPY. It may be the NULL pointer if the   */
/*         context is not required.             */
/*                                  */
/*  mode     = unsigned long flag to indicate desired operation.    */
/*       = PSL_VM_MODE_SPEC - Copy specified item into BUFF_PTR */
/*       = PSL_VM_MODE_FIRST  - Copy first item into BUFF_PTR   */
/*       = PSL_VM_MODE_LAST - Copy last item into BUFF_PTR  */
/*       = PSL_VM_MODE_NEXT - Copy next item into BUFF_PTR  */
/*       = PSL_VM_MODE_PREV - Copy previous item into BUFF_PTR  */
/*       = PSL_VM_MODE_INTO - Copy from BUFF_PTR into POS_PTR   */
/*                                  */
/************************************************************************/

unsigned long   PSL_VM_COPY (struct psl_vm_llb *list_ptr,
                             char *buff_ptr,
                             char *pos_ptr,
                             char **ctx_ptr,
                             unsigned long mode)
{
    /*
     *    Local data
     */
    struct   psl_vm_llb *element_ptr;   /* Points to element to act on  */
    struct   psl_vm_llb *copy_ptr;      /* Points to header of element  */
    void       *tmp_ptr;       /* Temporary pointer */

    /*
     *    Validate the input pointers.
     */
    if ((list_ptr == NULL) ||           /* validate the list pointer */
            ((*list_ptr).sanity_ptr != list_ptr) ||
            ((*list_ptr).blk_size == 0))
    {
        return PSL_BADPARAM;
    }

    /*
     *    A NULL pointer is NOT valid for modes of PSL_VM_MODE_SPEC and
     *    PSL_VM_MODE_INTO. It is valid in all other cases.
     */
    if (pos_ptr == NULL)
    {
        if ((mode == PSL_VM_MODE_SPEC) || (mode == PSL_VM_MODE_INTO))
            /* Check for INTO or SPEC modes */
        {
            copy_ptr = (*list_ptr).fwdptr;
        }               /* First item in the list */
        else if ((mode == PSL_VM_MODE_PREV) || (mode == PSL_VM_MODE_NEXT))

            /*
             *    We have a mode of PREV or NEXT. Seeing as how the POS_PTR has not been
             *    specified we will set the position to copy as being the start of the list.
             *    NEXT will therefore get the first item, PREV the last.
             */
        {
            copy_ptr = list_ptr;
        }
    }
    else                /* POS_PTR was specified    */
    {

        /*
         *    Convert the pointer supplied from being a pointer to the DATA section
         *    of the record to being a pointer to the header section.
         */
        tmp_ptr = pos_ptr - (sizeof (struct psl_vm_llb));
        copy_ptr = (struct psl_vm_llb *) tmp_ptr;
    }

    /*
     *    If the mode relies on the validity of the passed POS_PTR then we should
     *    validate it before we start to use it.
     */
    if (((mode != PSL_VM_MODE_FIRST) && (mode != PSL_VM_MODE_LAST))
            &&  ((*copy_ptr).sanity_ptr != copy_ptr))
    {
        return PSL_BADPARAM;
    }

    /*
     *    Determine from the MODE flag and POS_PTR fields exactly which item we are
     *    to act on.
     */
    switch (mode)
    {
    case PSL_VM_MODE_SPEC:      /* Copy item at pos_ptr     */
        {
            element_ptr = copy_ptr;     /* So set pointer to specified  */
            break;
        }
    case PSL_VM_MODE_FIRST:     /* Copy first item      */
        {
            element_ptr = (*list_ptr).fwdptr; /* So set pointer to first    */
            break;
        }
    case PSL_VM_MODE_LAST:      /* Copy last item       */
        {
            element_ptr = (*list_ptr).bckptr; /* So set pointer to last */
            break;
        }
    case PSL_VM_MODE_NEXT:      /* Copy item that follows specified */
        {
            element_ptr = (*copy_ptr).fwdptr; /* So set pointer to forward  */
            break;
        }
    case PSL_VM_MODE_PREV:      /* Copy item that precedes specified */
        {
            element_ptr = (*copy_ptr).bckptr; /* So set pointer to backward */
            break;
        }
    case PSL_VM_MODE_INTO:      /* Copy buffer to specified     */
        {
            element_ptr = copy_ptr;     /* So set pointer to specified  */
            break;
        }
    default:                /* Invalid mode flag specified  */
        {
            return PSL_BADPARAM;
        }
    }

    /*
     *    Check the sanity field of the element positioned at. It must
     *    point to itself and have a size field which is 0.
     */
    if ((*element_ptr).sanity_ptr != element_ptr)
    {
        return PSL_BADPARAM;
    }
    else if ((*element_ptr).blk_size != 0)
    {
        return PSL_RECNOTFOU;
    }

    /*
     *    If we do not have a buffer then we obviously cannot copy data to or
     *    from it so can skip straignt to the context checks.
     */
    if (buff_ptr != NULL)
    {

        /*
         *    Mode 6 is different as it involves copying from the buffer to the list.
         *    If we have specified a buffer pointer then copy the data from the list
         *    into the buffer.
         */
        if (mode == PSL_VM_MODE_INTO)   /* Copy item from position to buffer */
        {
            (void) memcpy ((*element_ptr).data_ptr,
                           buff_ptr,
                           (*list_ptr).blk_size);
        }
        else

            /*
             *    For all other modes, copy the buffer into the list at the calculated 
             *    position.
             */
        {
            (void) memcpy (buff_ptr,
                           (*element_ptr).data_ptr,
                           (*list_ptr).blk_size);
        }
    }

    /*
     *    Return the context of the item in question.
     */
    if (ctx_ptr != NULL)
    {
        (*ctx_ptr) = (*element_ptr).data_ptr;
    }

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    PSL_VM_DELETE                         */
/*    =============                         */
/*                                  */
/*    This routine deletes the entire linked list.          */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = PSL_VM_DELETE (list_id)                  */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret     = Unsigned long status return              */
/*           = PSL_BADPARAM if the parameter list is invalid    */
/*       = PSL_NORMAL if list deleted O.k.          */
/*                                  */
/*  list_ptr = Pointer to the linked list header returned       */
/*         by PSL_VM_INITIALIZE.                */
/*         This should be a pointer to STRUCT psl_vm_llb.   */
/*                                  */
/************************************************************************/

unsigned long   PSL_VM_DELETE (struct psl_vm_llb *list_ptr)
{
    /*
     *    local data
     */
    unsigned long       iret;       /* Status return variable   */

    /*
     *    Check the validity of the list pointer to ensure that it is a valid
     *    pointer to a list header.
     */
    if (((*list_ptr).sanity_ptr != list_ptr) ||
            ((*list_ptr).blk_size == 0))
    {
        return PSL_BADPARAM;
    }

    /*
     *    We must first loop round calling the remove routine to delete all of 
     *    the entries from the list.
     */
    do
    {
        iret = PSL_VM_REMOVE (list_ptr, NULL, NULL,
                              NULL, PSL_VM_POS_BEG);
    }
    while (iret == PSL_NORMAL);

    if (iret == PSL_BADPARAM)
    {
        return iret;
    }

    /*
     *    Now over-write the list header with nulls to prevent accidental
     *    reuse and then delete it.
     */
    (void) memset (list_ptr, '\0', sizeof (struct psl_vm_llb));
    free (list_ptr);
    return PSL_NORMAL;
}
