/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*                **** V T E S T ****                    */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: VTEST_CHK_ROUTINES              */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This source file contains common validation routines             */
/*  used by V-TEST.                              */
/*                                       */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <string.h>
#include <stdio.h>
#include <ctype.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines         */
#include <vtest_rtndef.h>       /* VTEST Routines       */

/*
 *    Parameter Definitions.
 */
#include <shrmsg.h>         /* PSL Messages         */

/************************************************************************/
/*                                  */
/*    VTEST_ANSWERBACK_CHK - Validation routine for ANSWERBACK string   */
/*                                  */
/*    Calling Convention (FDR Only):                    */
/*                                  */
/*  iret = VTEST_ANSWERBACK_CHK (value_ptr, translated_ptr ,maxlen) */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret = PSL return status                    */
/*       = PSL_NORMAL if answerback o.k.                */
/*       = VTEST_ANSWERTOOLONG if answerback invalid        */
/*  value_ptr = Value to be validated.                      */
/*  translated_ptr = Pointer to a buffer to hold translated string. */
/*  maxlen = Maximum length of answerback string.           */
/*                                  */
/*    This routine is called to validate an ANSWERBACK string. An   */
/*    answerback string may contain control characters which should be  */
/*    translated before the length is checked.              */
/*                                  */
/*    Optionally the translated buffer may be returned to the caller.   */
/*                                  */
/************************************************************************/

unsigned long   VTEST_ANSWERBACK_CHK    (         char *value_ptr
        ,         char *translated_ptr
        ,unsigned long max_length)

{
    /*
     *    Local data
     */
    unsigned  long    isub;         /* Loop control variable    */
    unsigned  long    length;       /* length           */
    char    wrk_string [128]; /* Work answerback string   */
    char    *wrk_ptr;     /* Work pointer to above    */

    /*
     *    Before we can tell if the input answerback string is valid we need to
     *    reformat it to remove any control sequences and replace them in our work
     *    string with a single character representation. When we have done that,
     *    only then can we check the length of the resulting string to see if it
     *    exceeds the maximum permitted length of the answerback string.
     */
    length = strlen (value_ptr);    /* Length of entered string */
    wrk_ptr = wrk_string;       /* Set pointer to string    */

    /*
     *    Loop through the input string reformatting it.
     */
    isub = 0;               /* Zeroize loop count       */

    while (isub < length)       /* While more of input to check */
    {

        /*
         *   Check for a '^' character which is NOT the last character in the buffer
         */
        if ((*value_ptr == '^') && (isub < (length - 1)))
        {               /* if we have a '^'     */
            value_ptr += 1;     /* ...point past it in value    */
            isub += 1;          /* ...increment index as well   */

            /*
             *    If we have a second consecutive '^', we shall treat it as a literal '^'
             *    and add that to the output buffer, otherwise we treat it as a control
             *    sequence and add the ASCII value of that control sequence to the buffer.
             */
            if (*value_ptr == '^')  /* ...and if we have another    */
            {
                *wrk_ptr = '^';     /* Add a '^' to output buffer   */
            }

            else
            {
                *wrk_ptr = (*value_ptr & 0X1F);
                /* Store CONTROL character  */
            }
        }

        else                /* ...else simply another char. */
        {

            /*
             *    Simply another character - add it to the output buffer.
             */
            *wrk_ptr = *value_ptr;
        }

        /*
         *    Advance the work pointer and the value pointer and increment the index
         *    so that we check the next character in the input string.
         */
        value_ptr += 1;         /* Advance input string pointer */
        wrk_ptr += 1;           /* ...and output string pointer */
        isub += 1;          /* ...and index counter     */
    }

    /*
     *    Now that we have formatted the input string we should check to see just
     *    how long the reformatted string is. If it exceeds the length of the
     *    TDF field which is to hold it we report that it is invalid, otherwise
     *    the field is copied into the EXT TDF record.
     */
    *wrk_ptr = '\0';            /* Null terminate string.   */
    if ((strlen (wrk_string)) > max_length)
    {                   /* If string is too long    */
        return VTEST_ANSWERTOOLONG; /* ...return error status   */
    }

    else                /* ...otherwise...      */
    {
        if (translated_ptr != NULL) /* If copy back translated  */
        {
            (void) strcpy (translated_ptr ,wrk_string);
        }
        return PSL_NORMAL;
    }
}

/************************************************************************/
/*                                  */
/*    VTEST_INPRATE_CHK - Validation routine for INPUT RATE string  */
/*                                  */
/*    Calling Convention (FDR Only):                    */
/*                                  */
/*  iret = VTEST_INPRATE_CHK (value_ptr);               */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret = PSL return status                    */
/*       = PSL_NORMAL if input rate o.k.                */
/*       = VTEST_INVINPRATE if input rate invalid           */
/*           = VTEST_RATEOUTRANGE if input rate out of range        */
/*  value_ptr = Value to be validated.                      */
/*                                  */
/*    This routine is called to validate an INPUT RATE string. An   */
/*    input rate should be in the form n.m where n or .m is optional.   */
/*                                  */
/************************************************************************/

unsigned long   VTEST_INPRATE_CHK   (char *value_ptr)

{
    /*
     *    Local data
     */
    unsigned long     isub;         /* Loop Control         */
    unsigned long     iret;         /* Routine Return Status    */
    long     length;       /* Length of string     */
    long     converted_rate;   /* Converted input rate     */
    char     input_rate [16];  /* Formatted input rate     */
    char     *input_rate_ptr;  /* Pointer to trimmed rate  */

    /*
     *    Valid input rates have the form n.m where either n or .m is optional
     *    The input rate may be preceded by leading spaces or zeros.
     *
     *    To cater for this we will strip off all leading spaces, nulls and zeros.
     */
    (void) memset (input_rate, '\0', sizeof (input_rate));
    /* NULL fill string initially   */
    input_rate_ptr = value_ptr;     /* Work pointer to supplied value */
    for (isub = 0; isub < 4; isub++, input_rate_ptr++)
    {
        if ((*input_rate_ptr != ' ')    /* If NOT a SPACE ...       */
                &&  (*input_rate_ptr != '\0')   /* ...and NOT a NULL ...    */
                &&  (*input_rate_ptr != '0'))   /* ...and NOT a ZERO ...    */
        {
            (void) strcpy (input_rate, input_rate_ptr);
            /* ...then copy rest of input rate */
            goto chk_decimal;       /* ...and branch to check it.   */
        }
    }

    /*
     *    CHK_DECIMAL
     *    -----------
     *
     *    We now have a string with all LEADING characters stripped off. If the
     *    last character is a decimal point then strip it off as it is not valid.
     */
chk_decimal:
    length = strlen (input_rate);   /* get length so far        */
    input_rate_ptr = input_rate + length - 1;
    /* Point to last char in rate   */
    if (*input_rate_ptr == '.')     /* If trailing decimal point    */
    {
        *input_rate_ptr = '\0';     /* ...remove it...      */
        length -= 1;            /* ...and decrement length  */
    }

    /*
     *    If we have a decimal point it must be the next to last character. If we
     *    have a leading '.' add a zero
     */
    input_rate_ptr = memchr (input_rate, '.', length);
    if (input_rate_ptr == input_rate)
    {                   /* If '.' first character   */
        (void) strcpy (input_rate_ptr + 1, input_rate_ptr);
        /* Shuffle data up 1 character  */
        *input_rate_ptr = '0';      /* ...add a leading zero    */
        length += 1;            /* ...increment length      */
        input_rate_ptr += 1;        /* ...increment pointer to '.'  */
    }

    if (input_rate_ptr != NULL)
    {
        if ((input_rate_ptr-input_rate)!=(length - 2)) return VTEST_INVINPRATE;
    }

    /*
     *    The general format of the input rate is correct. We should now strip
     *    out any decimal point (effectively multiplying the rate by 10) and pass
     *    it to PSL_ASCBIN which will convert the rate to an integer so that we
     *    can check that it is within range.
     *
     *    We can now however copy the formatted input rate back to the form driver
     *    so that it can pick up the formatted value.
     */
    (void) strcpy (value_ptr, input_rate);

    /*
     *    ...and translate the input rate into a numeric value.
     */
    iret = VTEST_TRANSLATE_RATE (input_rate ,&converted_rate);
    if (!(psl_test_status (iret))) return VTEST_INVINPRATE;

    /*
     *    The range of valid input rates is 0.1 to 99.9 characters per second. This
     *    gives us an integer value of between 1 and 999 from the conversion.
     */
    if ((converted_rate < 1) || (converted_rate > 999))
    {
        return VTEST_RATEOUTRANGE;
    }

    else
    {
        return PSL_NORMAL;
    }
}

/************************************************************************/
/*                                  */
/*    VTEST_TRANSLATE_RATE - Translate a character input rate       */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = VTEST_TRANSLATE_RATE (character_rate ,&numeric_rate) */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret       = Unsigned Long Routine Return Status        */
/*  character_rate = Null terminated character rate to translate    */
/*  numeric rate   = L to hold return translated input rate     */
/*                                  */
/*    Description:                          */
/*  This routine is called to translate a character input rate in   */
/*  the form NN.M to a numeric representation * 10          */
/*                                  */
/*  For example:                            */
/*      "99.9" input becomes 999                */
/*      "0.1"  input becomes 1                  */
/*                                  */
/************************************************************************/

unsigned long   VTEST_TRANSLATE_RATE    (         char   *character_rate
        ,         long   *numeric_rate_ptr)

{
    /*
     *  Local Data.
     */
    unsigned long     iret;         /* Routine Return Status    */
    unsigned char     radix = 10;       /* Decimal Conversion Radix     */
    unsigned char     length;       /* Number of characters to convert */
    unsigned short    chrcnt;       /* Unconverted character count  */
    char     *input_rate_ptr;  /* Pointer to trimmed rate  */
    char     input_rate[256];

    /*
     *    Copy the input rate into a local buffer so that we can modify it.
     */
    (void) strcpy (input_rate, character_rate);

    /*
     *    Locate any decimal point in the input rate specified.
     */
    input_rate_ptr = strchr (input_rate, '.');
    /* Find any decimal point   */

    /*
     *    If we do not have a decimal point then we simply add a 0 to the end of
     *    the rate to multiply it by 10 as this is what we do with rates with a
     *    decimal point.
     */
    if (input_rate_ptr == NULL)
    {                   /* If no decimal point      */
        (void) strcat (input_rate ,"0");
        /* ...add a trailing zero   */
    }

    else if (input_rate_ptr == input_rate)
    {                   /* ...else if '.' first char    */
        input_rate [0] = '0';       /* ...replace with leading zero */
    }

    else                /* ...otherwise embedded '.'    */
    {
        (void) strcpy (input_rate_ptr, input_rate_ptr + 1);
        /* Shuffle data down onto '.'   */
    }

    /*
     *    Translate this to a numeric value.
     */
    length = strlen (input_rate);   /* ...and total length of rate  */
    iret = PSL_ASCBIN (&length ,input_rate
                       ,&radix ,numeric_rate_ptr ,&chrcnt);
    /* ...perform the conversion    */
    return iret;            /* ...and return status     */
}

/************************************************************************/
/*                                  */
/*    VTEST_ANSWERBACK_FMT - Format the non-printing answerback string  */
/*              into a printable format.            */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  VTEST_ANSWERBACK_FMT (inp_string_ptr ,out_string_ptr);          */
/*                                  */
/*    Where:                                */
/*                                  */
/*  inp_str_ptr - Pointer to NULL terminated string to format   */
/*  out_str_ptr - Pointer to string to hold reformatted data    */
/*                                  */
/*    This routine is called to reformat an answerback string which */
/*    may contain control characters into a string which has these  */
/*    characters translated into a printable form.          */
/*                                  */
/************************************************************************/

void   VTEST_ANSWERBACK_FMT    (char   *inp_string_ptr
                                ,char   *out_string_ptr)

{
    /*
     *    Local data
     */
    unsigned long   isub;           /* Loop control variable    */
    unsigned long   inp_string_length;  /* Length of input string   */
    char   control_character;  /* Control character as ASCII   */

    inp_string_length = strlen (inp_string_ptr);

    /*
     *    Copy the answerback string from the input to the output string,
     *    and reformat by replacing each control character by '^c' and '^' by '^^'.
     */
    for (isub = 0; isub < inp_string_length; isub++, inp_string_ptr++)
    {

        /*
         *    Check to see if the current character is a 'Control' character and if
         *    so represent it in the form '^c' where 'c' is the ASCII non-control
         *    value of the character.
         */
        if (iscntrl (*inp_string_ptr))
        {               /* If control character     */
            control_character = (*inp_string_ptr | 64);
            /* Get control character as ASCII */
            *out_string_ptr = '^';  /* ...add '^' to output buffer    */
            out_string_ptr++;       /* ...and move pointer forwards   */
            *out_string_ptr = control_character;
            /* ...add control to output       */
            out_string_ptr++;       /* ...and move pointer forwards   */
        }

        else if (*inp_string_ptr == '^')
        {               /* ...else if '^' character */
            *out_string_ptr = '^';  /* ...add '^' to output buffer  */
            out_string_ptr++;       /* ...and move pointer forwards */
            *out_string_ptr = '^';  /* ...add '^' again     */
            out_string_ptr++;       /* ...and move pointer forwards */
        }

        else                /* ...otherwise...      */
        {
            *out_string_ptr = *inp_string_ptr;
            /* ...simply copy character */
            out_string_ptr++;       /* ...and move pointer forwards */
        }
    }

    /*
     *    ...and null terminate the string before returning.
     */
    *out_string_ptr = '\0';
    return;
}
