/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*
 * scl_tparse.c:
 * =============
 *
 * This module contains the main control routine (SCL_TPARSE) and support
 * routines used to parse an SCL source file.
 */

/*
 *  PSL Macros.
 */
#include <psl_macros.h>
#include <vtest_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines         */
#include <trc_rtndef.h>         /* Trace Routines       */
#include <target_rtndef.h>      /* Target definitions   */
#include "scl_rtndef.h"         /* SCL Routines         */

static void     INI_DEP_LL      (struct scl_com *com_ptr);

static void     INI_TOFL_LL     (struct scl_com *com_ptr);

/*
 *  Structure definitions.
 */
#include <shrmsg.h>         /* PSL Messages         */
#include <trmdef.h>         /* TRM definitions      */
#include <psldiag_prmdef.h>     /* Diagnostic Routine Symbols */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block    */
#include <psl_tpatabdef.h>      /* TPARSE Table         */
#include <scpdef.h>         /* SCP Header record        */
#include "scl_tof.h"            /* SCL TOF data */
#include "scl_tlodef.h"         /* Test definition records */
#include "scl_tbldef.h"         /* SCL Table Definitions    */
#include <psl_vm_prmdef.h>      /* LL Parameters        */
#include <vtest_prmdef.h>       /* VTEST Parameters     */
#include "scl_cmd.h"            /* Command Line Data Structure */
#include "scl_comdef.h"         /* Common Data */
#include "scl_limits.h"         /* SCL limits */
#include "scl_tpacom.h"         /* Tparse common Data       */
#include <psl_trc_prmdef.h>     /* Tracing definitions      */
#include <pslerr_prmdef.h>      /* Error Constants      */
#include <pslmsg_prmdef.h>      /* Error Constants      */
#include <tofprmdef.h>          /* TOF parameters */

/*
 *    ASCII codes for control characters that can be represented by their
 *    ANSI mnemonics in character commands.
 */
#define CODE_BEL    0X07
#define CODE_BS     0X08
#define CODE_HT     0X09
#define CODE_LF     0X0A
#define CODE_VT     0X0B
#define CODE_FF     0X0C
#define CODE_CR     0X0D
#define CODE_ESC    0X1B
#define CODE_SP     0X20
#define CODE_DEL    0X7F

/*
 *    Colour definitions. These are referenced by the TPARSE tables. We
 *    shift the TRM value into the second two bytes of a long as the colour
 *    values are held in a general attributes mask.
 */
#define     VOPT_VALUE  VOPT_LIST|VOPT_RANGE
#define     FORECOL_BLACK   (TCOL_BLACK-1) << 24
#define     FORECOL_RED (TCOL_RED-1) << 24
#define     FORECOL_GREEN   (TCOL_GREEN-1) << 24
#define     FORECOL_YELLOW  (TCOL_YELLOW-1) << 24
#define     FORECOL_BLUE    (TCOL_BLUE-1) << 24
#define     FORECOL_MAGENTA (TCOL_MAGENTA-1) << 24
#define     FORECOL_CYAN    (TCOL_CYAN-1) << 24
#define     FORECOL_WHITE   (TCOL_WHITE-1) << 24
#define     BACKCOL_BLACK   (TCOL_BLACK-1) << 28
#define     BACKCOL_RED (TCOL_RED-1) << 28
#define     BACKCOL_GREEN   (TCOL_GREEN-1) << 28
#define     BACKCOL_YELLOW  (TCOL_YELLOW-1) << 28
#define     BACKCOL_BLUE    (TCOL_BLUE-1) << 28
#define     BACKCOL_MAGENTA (TCOL_MAGENTA-1) << 28
#define     BACKCOL_CYAN    (TCOL_CYAN-1) << 28
#define     BACKCOL_WHITE   (TCOL_WHITE-1) << 28

/*
 *    Define the various tables which we will require during parsing. Pointers
 *    to these structures will be set-up in the initialization
 */
static struct   symtbl   variable_table [MAX_VARIABLES];
/* Variable symbol table    */
static struct   labtbl   label_table [MAX_LABELS];
/* Labels symbol table      */
static struct   subtbl   subroutine_table [MAX_SUBROUTINES];
/* Subroutines symbol table     */
static struct   ifstack  if_table [MAX_IF];
/* Alloc space for IF table     */
static struct   dostack  do_table [MAX_DO];
/* Alloc space for DO table     */
/*
 * Structures for switch statement
 */

static struct   swstack  sw_table [10]; /* Max switch level             */
/* Alloc space for SWITCH table */
static struct   scph     header_record; /* Alloc space for header record */

static struct   scl_tofh tofh_record;   /* TOF header record */

static          char     code_buffer [MAX_SCP_SIZE + OBJ_VALUES_SIZE];
/* Alloc space for code and vlist */
static struct   timtbl   timer_table [MAX_TIMERS];
/* Timer Symbol Table       */
static struct   shrvartbl
            shrvar_table [MAX_SHRVAR];
/* Shareable Variables Symbol Table */
static struct   prmtbl   parameter_table [MAX_PARAMETERS];
/* Space for Script Parameters  */
static struct   fattbl   file_table [MAX_FILES];
/* File Access Table        */
static struct   fattbl   pipe_table [MAX_PIPES];
/* Pipe Access Table        */
static struct   tsttbl   thistst;   /* Test data for this test file */
static struct   tsttbl   tst_table [MAX_TST];
/* Test Symbol Table */
static struct   tstthrtbl
            tstthr_table [MAX_TSTTHR];
/* Test Thread Symbol Table */
static struct   thrtbl   thr_table [MAX_THR];
/* Thread Symbol Table */
static          char     constant_buffer [MAX_CONSTANT_SIZE];
/* Alloc space for constants    */
static          long     file_dtyp_buf [MAX_FILES];
/* Space for file data type (comma)*/
static struct   tsttbl   tst_wrk;   /* Test data work buffer */
static struct   thrtbl   thr_wrk;   /* Thread data work buffer */
static struct   tstthrtbl
            tstthr_wrk;    /* Test Thread data work buffer */

static struct   tdp  tdp_wrk;   /* Thread parameter data work buffer */

static struct   tthr     tthr_wrk;  /* Test thread data work buffer */

/************************************************************************/
/*                                  */
/*    SCL_TPARSE - TPARSE Control Routine               */
/*    ==========                            */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = SCL_TPARSE (com_ptr);                    */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret          = Unsigned Long PSL Routine Return Status     */
/*  com_ptr       = Pointer to Common Data Structure        */
/*                                  */
/*    This routine invokes PSL_TPARSE.                  */
/*                                  */
/************************************************************************/

unsigned long SCL_TPARSE (struct scl_com *com_ptr)

{
    /*
     *    Local data
     */
    static struct   scl_tpa scl_tpa;    /* Tparse Data          */
    struct tpaarg   tparse_data;        /* TPARSE Argument Block    */
    struct scl_tpa  *tpa_ptr;       /* Pointer to TPA data      */
    Boolean  trace_flag;     /* Trace flag           */

    /*
     *  TPARSE state tables.
     */
#include    "scl_sub_tbl.h"     /* General subexpression tables */
#include    "scl_thrsub_tbl.h"  /* Thread script specific */
    /* subexpression tables */
#include    "scl_thr_tbl.h"     /* Thread script specific tables */

#include    "scl_tst_tbl.h"     /* Test script specific tables */

#include    "scl_cod_tbl.h"     /* Code section tables */
#include    "scl_def_tbl.h"     /* Definition section tables */
#include    "scl_env_tbl.h"     /* Environment section tables */
    /*
     *    Set up the link between the SCL_TPA structure and its associated pointer
     *    in the SCL_COM structure
     */
    tpa_ptr = &scl_tpa;         /* Get pointer to TPARSE data   */
    (*com_ptr).tpa_ptrps = tpa_ptr;     /* ...and store in common area  */

    /*
     *    Initialize TPARSE Argument Block so that the Parsing Buffer is empty
     *    (force call to GET_NEXT_SOURCE_LINE).
     */
    (void) memset (&tparse_data, '\0', sizeof (struct tpaarg));

    /*
     *    Set up the first parsable line to force a new line to be read immediately
     */
    tparse_data.stringcntl = 0;
    tparse_data.string_ptrpc = (*tpa_ptr).parsable_linec136;

    /*
     *    Set up the tracing information
     */
    tparse_data.tracertn_ptrpf = PSL_TPATRACE;
    /* Point to trace routine   */
    trace_flag = PSL_TEST_TRACE (PTRC_GEN_TPARSE);
    /* Check for tparse tracing     */
    if (trace_flag)         /* ...and if we are     */
    {
        tparse_data.tracelvll = 3;  /* Trace level 3        */
    }
    else                /* ...otherwise...      */
    {
        tparse_data.tracelvll = 0;  /* No tracing           */
    }

    /*
     *    Store pointer to Common Data in TPARSE Argument Block.
     */
    tparse_data.user_ptrp = com_ptr;

    /*
     *    Perform the parsing.
     */
    return PSL_TPARSE (&tparse_data, scl_env_table);
}

/************************************************************************/
/*                                  */
/*    SCL_GET_LINE - Get the next line from the Source File     */
/*    ============                          */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = SCL_GET_LINE (argblk_ptr);               */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret       = Unsigned Long PSL Routine Return Status        */
/*  argblk_ptr = Pointer to TPARSE Argument Block Structure     */
/*                                  */
/*    This routine gets the next line from the source file.     */
/*                                  */
/*    It first checks to see if there are any errors in the current */
/*    source line and, if so, outputs the underline and reports the */
/*    errors.                               */
/*                                  */
/************************************************************************/

unsigned long   SCL_GET_LINE (struct tpaarg *argblk_ptr)

{
#ifdef UNIX
#  define ERROR_FORMAT_STRING "\"%s\", line %ld %s\n"
#elif WIN32
#  define ERROR_FORMAT_STRING "%s(%ld) %s\n"
#else
#  error Platform not supported at line __LINE__ (Formatting errors)
#endif
    /*
     *  The maximum number of characters (in decimal) required to output
     *  a long int
     */
#define MAX_LONG_IN_DECIMAL 16
    /*
     *  The size of the buffer required to hold the ERROR_FORMAT_STRING
     */
#define ERROR_LINE_SIZE (strlen(ERROR_FORMAT_STRING)+strlen(ErrorText)+\
strlen((*com_ptr).scl_namec256)+MAX_LONG_IN_DECIMAL+1)
    /*
     *  Local Data.
     */
    unsigned long    iret = PSL_NORMAL; /* Status Return          */
    unsigned long    jsub;          /* Loop Control           */
    unsigned long    file_id;       /* "Read" file identifier*/
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    char    listing_line [256];    /* Listing line         */
    struct   scl_cmd *scl_cmd_ptr;      /* Pointer to Command Line Struct */

    int     nest_level;            /* Current nesting level indicator*/

    char     *ErrorLinep;       /* to hold the error line */
    char     ErrorText[128];    /* to hold the PSL error text */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    scl_cmd_ptr = (*com_ptr).cmd_ptrps;

    /*
     *    If we have already read a line then list it at this point.
     */
    if ((*tpa_ptr).line_read_flagl != 0)
    {
        /*
         * If there is a warning - output it now.
         */
        if (((*tpa_ptr).test_case_warnl == 1) ||
                ((*tpa_ptr).sql_warnn))
        {
            (void) fprintf (stderr, "%.*s\n",
                            (int) (*tpa_ptr).line_sizel,
                            (*tpa_ptr).source_linec136);
        }

        /*
         *    Format the listing line to be :
         *    1. The line reference number
         *    2. The nesting level
         *    3. The source line.
         */

        /*
         *       nest_level = current nesting level
         *
         * if this is greater than zero display it in listing file after line number.
         */

        nest_level = ((*tpa_ptr).do_cntl + (*tpa_ptr).ifstack_levell);

        if (nest_level == 0)
        {
            (void) sprintf (listing_line, "%4ld     %.*s"
                            ,(*tpa_ptr).line_ref_numberl
                            ,(int) (*tpa_ptr).line_sizel
                            ,(*tpa_ptr).source_linec136);
            SCL_LIST_LINE (com_ptr, listing_line);
        }
        else
        {
            (void) sprintf (listing_line, "%4ld %3d %.*s"
                            ,(*tpa_ptr).line_ref_numberl
                            ,nest_level
                            ,(int) (*tpa_ptr).line_sizel
                            ,(*tpa_ptr).source_linec136);
            SCL_LIST_LINE (com_ptr, listing_line);
        }
        (*tpa_ptr).line_ref_numberl += 1;
    }

    /*
     *    Are there any errors to report?
     */
    if ((*tpa_ptr).errors_in_linel != 0)
    {
        /*
         *    Output the source line if we are on UNIX and a listing is not required
         */

        if (!(*scl_cmd_ptr).listn)
        {
            (void) fprintf (stderr, "\n%.*s\n",
                            (int) (*tpa_ptr).line_sizel,
                            (*tpa_ptr).source_linec136);
        }

        /*
         *    ...but only output the underline buffer, if required.
         */
        if ((*tpa_ptr).underline_flagl != 0)
        {

            /*
             *    Output to the listing file. If it fails, report the error and exit.
             */
            if ((*scl_cmd_ptr).listn)
            {
                (void) sprintf (listing_line, "%.*s",
                                (int) (*tpa_ptr).line_sizel,
                                (*tpa_ptr).error_linec136);
                iret = LIST_LINE ((*com_ptr).lst_contextps, listing_line);
                if (!(psl_test_status (iret))) goto err_exit;
            }

            else
            {
                (void) fprintf (stderr, "%.*s\n",
                                (int) (*tpa_ptr).line_sizel,
                                (*tpa_ptr).error_linec136);
            }
        }

        /*
         *    Loop through the array of errors and report them.
         */
        for (jsub = 0; jsub < (*tpa_ptr).errors_in_linel; jsub++)
        {
            PSL_GET_MESSAGE ((*tpa_ptr).error_arrayl5[jsub]
                             ,ErrorText
                             ,sizeof (ErrorText)
                             ,PSLMSG_TEXT | PSLMSG_CODE);

            ErrorLinep = (char *)malloc(ERROR_LINE_SIZE);
            if (ErrorLinep == NULL)
            {
                iret = PSL_INSVIRMEM;
                goto err_exit;
            }

#ifdef  UNIX
            (void) sprintf (ErrorLinep
                            ,ERROR_FORMAT_STRING
                            ,(*com_ptr).scl_namec256
                            ,(*com_ptr).srclinl10 [(*com_ptr).include_depthl]
                            ,ErrorText);
#elif   WIN32
(void) sprintf (ErrorLinep
            ,ERROR_FORMAT_STRING
            ,(*com_ptr).scl_namec256
            ,(*com_ptr).srclinl10 [(*com_ptr).include_depthl]
            ,ErrorText);
#else
#error Platform not supported at line __LINE__ (Formatting errors)
#endif

            /*
             *    We used to call (f)printf to output the error line.  However, the
             *    latter typically contains a % (e.g. %SCL-), and although the conversion
             *    character is not valid, some versions of the C RTL appear to be trying
             *    to handle it.  As we don't actually need to format the line, we will
             *    use fputs instead.
             */
            (void) fputs (ErrorLinep, stderr);

            /*
             *    If we are listing send this to the listing file also.
             */
            if ((*scl_cmd_ptr).listn)
            {
                iret = LIST_LINE ((*com_ptr).lst_contextps ,ErrorLinep);
                if (!(psl_test_status (iret))) goto err_exit;
            }

            free(ErrorLinep);
            ErrorLinep = NULL;
        }

        /*
         *    Reset counter and underline flag.
         */
        (*tpa_ptr).errors_in_linel = 0;
        (*tpa_ptr).underline_flagl = 0;
    }

    /*
     *    Read the next line after determining which file to read from.
     */
    if ((*com_ptr).include_depthl == 0) /* If we are reading the main source */
    {
        file_id = SRC_UNIT;     /* Unit = main source unit  */
    }

    else                /* ...else if reading include   */
    {
        file_id = INCLUDE_START_UNIT + (*com_ptr).include_depthl - 1;
        /* Unit = Current include unit  */
    }

    READ_NEXT_LINE (com_ptr,            /* Pass pointer to common  */
                    &(*tpa_ptr).line_sizel,    /* ...and return line size */
                    (*tpa_ptr).source_linec136,
                    (*tpa_ptr).parsable_linec136,
                    (*tpa_ptr).error_linec136,
                    &(*com_ptr).eof_flagl,
                    file_id);

    /*
     *    Set up to parse the new line.
     */
    (*argblk_ptr).string_ptrpc = (*tpa_ptr).parsable_linec136;
    (*argblk_ptr).stringcntl = (*tpa_ptr).line_sizel;
    /* Setup to parse next line */
    (*tpa_ptr).line_read_flagl = 1; /* ...and show line read    */

    return PSL_NORMAL;

    /*
     *    Report error from Listing routine and exit.
     */
err_exit:
    PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                 1,
                 iret);
    PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);

    if (!(psl_test_status (iret)))
    {
        exit (EXIT_FAILURE);
    }
    else
    {
        exit (EXIT_SUCCESS);
    }

    return 0;
}

/************************************************************************/
/*                                  */
/*    INI_DEP_LL - Initialize script dependency linked list     */
/*    ==========                            */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  INI_DEP_LL (com_ptr);                       */
/*                                  */
/*    Where:                                */
/*                                  */
/*  com_ptr    = Pointer to SCL common data.            */
/*                                  */
/*    This routine deletes any existing script dependency linked list   */
/*    - including all associate memory - and allocates a new one.   */
/*                                  */
/************************************************************************/
static void INI_DEP_LL (struct scl_com *com_ptr)
{
    /*
     *  Local data.
     */
    struct scl_tpa   *tpa_ptr;      /* Pointer to SCL Tparse data */
    unsigned  long   iret;          /* PSL routine status */

    /*
     *  Copy pointers for ease of reference.
     */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  Delete existing script dependency linked list, if any.
     */
    DEL_DEP_LL (com_ptr);

    /*
     *  Now allocate the new list.
     */
    iret = PSL_VM_INITIALIZE (sizeof (struct sdep), &(*com_ptr).dep_llps);
    if  (!(psl_test_status (iret)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_INITIALIZE (in INI_DEP_LL)", iret);
        PSL_SIGNAL_ERROR ((*tpa_ptr).diag_ptrps, PSLERRCLS_USER1, iret);
    }

    (*com_ptr).dep_cntl = 0;

    return;
}

/************************************************************************/
/*                                  */
/*    INI_PARSE_COMMON - Common TPARSE Data Initialisation      */
/*    ================                          */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = INI_PARSE_COMMON (argblk_ptr);               */
/*                                  */
/*    Where:                                */
/*                                  */
/*      iret    = Unsigned Long PSL Routine Return Status       */
/*      = PSL_NORMAL if all O.k.                */
/*      = PSL_INSVIRMEM if error allocating structures      */
/*                                  */
/*  argblk_ptr    = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*    This routine performs all the TPARSE data initialiation that  */
/*    is required before EACH compilation and is common to ALL types    */
/*    of script, i.e. Script/Test scripts AND Thread scripts.       */
/*                                  */
/************************************************************************/
unsigned long INI_PARSE_COMMON (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com   *com_ptr;      /* Temp pointer to COM record */
    struct scl_tpa   *tpa_ptr;      /* Temp pointer to TPA record */
    struct scph  *header_ptr;       /* Temp pointer to SCPH record */
    struct scl_tofh  *tofh_ptr;     /* Temp pointer to TOFH record */
    unsigned  long   iret;          /* Status return variable */
    struct array_blk *array_blk_ptr;    /* Array Block Pointer */
    struct oph   *oph_ptr;      /* Pointer to OPH */
    struct opd   *opd_ptr;      /* Pointer to OPD */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  Save diagnostic stack pointer.
     */
    (*tpa_ptr).diag_ptrps = (*com_ptr).diag_ptrps;

    /*
     *  Set up pointer to temporary SCP header record.
     */
    (*tpa_ptr).header_begps = &header_record;
    header_ptr = (*tpa_ptr).header_begps;       /* Local copy of pointer */

    /*
     *  Set up pointer to TOF header record data buffer.
     */
    (*tpa_ptr).tofh_begps = &tofh_record;
    tofh_ptr = (*tpa_ptr).tofh_begps;           /* Local copy of pointer */

    /*
     *    Initialize global data.
     */
    (*tpa_ptr).error_in_commandn = PFALSE;
    /* Clear error flag at end of script */
    (*com_ptr).include_depthl = 0;  /* Zeroize include depth    */
    (*com_ptr).tof_srclinl = 0;     /* Zero TOF source line number */ 
    (*com_ptr).eof_flagl = 0;       /* Clear EOF flag       */
    (*tpa_ptr).errors_in_linel = 0; /* No Errors            */
    (*tpa_ptr).underline_flagl  = 0;    /* Clear Underline flag     */
    (*tpa_ptr).line_read_flagl = 0; /* Show no lines read yet   */
    (*tpa_ptr).mode_flagl = 0;      /* ...and in parsing data   */
    (*tpa_ptr).error_countl = 0;    /* Zerize error count       */
    (*tpa_ptr).env_opt_specl = 0;   /* Clear env options spec flags */
    (*tpa_ptr).scl_optl = 0;        /* Clear scp options spec flags */
    (*tpa_ptr).cmnd_chrb = (*com_ptr).cmnd_chr_valuec256 [0];
    /* Setup command character  */
    (*tpa_ptr).ctrl_chrb = (*com_ptr).ctrl_chr_valuec256 [0];
    /* Setup control character  */
    (*tpa_ptr).line_ref_numberl = 1;     /* Initialise listing count     */
    (*tpa_ptr).sclsec_flagl = SCLSEC_NONE;
    /* Current script section */

    /*
     * Set up fields to indicate which platform and OS the script
     * is being compiled under.
     */


    /*
     * Set up OS
     */
#ifdef UNIX
    (*header_ptr).scph_compile_opsys_l = SCPH_OS_UNIX;

#elif   WIN32
    (*header_ptr).scph_compile_opsys_l = SCPH_OS_WIN32;

#else
#error Platform not supported at line __LINE__ (setting OS type)
#endif

    /*
     * Set up PLATFORM
     */
#ifdef VAX
    (*header_ptr).scph_compile_plat_l = SCPH_PLAT_VAX;

#elif ALPHA
    (*header_ptr).scph_compile_plat_l = SCPH_PLAT_ALPHA;

#elif SUN
(*header_ptr).scph_compile_plat_l = SCPH_PLAT_SUN;

#elif HP
    (*header_ptr).scph_compile_plat_l = SCPH_PLAT_HP;

#elif IBM
    (*header_ptr).scph_compile_plat_l = SCPH_PLAT_IBM;

#elif NCR
    (*header_ptr).scph_compile_plat_l = SCPH_PLAT_NCR;

#elif NTI
    (*header_ptr).scph_compile_plat_l = SCPH_PLAT_NTI;

#elif OSF
    (*header_ptr).scph_compile_plat_l = SCPH_PLAT_OSF;

#else
#error platform not supported at line __LINE__
#endif



    /*
     *    We require a linked list to hold the operand headers/descriptors as they
     *    are compiled. We create a new list for each source file just in case
     *    something has gone wrong previously...
     */
    if ((*tpa_ptr).parsed_operand_llps != NULL)
    {
        do
        {
            iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_operand_llps,
                                  (char *) &oph_ptr, NULL, NULL, PSL_VM_POS_BEG);
            if (!psl_test_status (iret))
            {
                PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                             ,"PSL_VM_REMOVE oph (in INI_PARSE_COMMON)", iret);
            }

            iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_operand_llps,
                                  (char *) &opd_ptr,NULL, NULL, PSL_VM_POS_BEG);
            if (!psl_test_status (iret))
            {
                PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                             ,"PSL_VM_REMOVE opd (in INI_PARSE_COMMON)", iret);
            }

            /* Free memory associated with the operand */
            FREE_OPH_OPD (oph_ptr, opd_ptr);
        }
        while (psl_test_status (iret));     /* Loop while not an error */

        (void) PSL_VM_DELETE ((*tpa_ptr).parsed_operand_llps);
        (*tpa_ptr).parsed_operand_llps = NULL;
    }

    /*
     *    ...and allocate the new list.
     */
    iret = PSL_VM_INITIALIZE (sizeof (long *)
                              ,&(*tpa_ptr).parsed_operand_llps);
    if  (!(psl_test_status (iret)))
    {
        PSL_PUTDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_INITIALIZE (in INITIALIZE)", iret);
        PSL_SIGNAL_ERROR ((*tpa_ptr).diag_ptrps, PSLERRCLS_USER1, iret);
        return 0;           /* ...and return        */
    }

    /*
     *    We also require a linked list for use when parsing array indicis. This
     *    list holds pointers to the OPH and OPD of the last array parsed. This is
     *    required so that when we check the dimensions of an array, we can tell
     *    how many dimensions the parent array has.
     */
    if ((*tpa_ptr).parsed_array_llps != NULL)
    {
        do
        {
            iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_array_llps,
                                  (char *) &array_blk_ptr,
                                  NULL, NULL, PSL_VM_POS_BEG);
            if (psl_test_status (iret))
            {
                FREE (array_blk_ptr);
#ifdef PSL_TRACING
                if (PSL_TEST_TRACE (PTRC_USR_USER5))
                {
                    PSL_OUTPUT_TRACE ("FREE: array - '%p' - INITIALIZE\n",
                                      array_blk_ptr);
                }
#endif  /* PSL_TRACING */
            }
        }
        while (psl_test_status (iret));     /* Loop while not an error */

        PSL_VM_DELETE ((*tpa_ptr).parsed_array_llps);
        (*tpa_ptr).parsed_array_llps = NULL;
    }

    /*
     *    ...and allocate the new list.
     */
    iret = PSL_VM_INITIALIZE (sizeof (struct array_blk)
                              ,&(*tpa_ptr).parsed_array_llps);
    if  (!(psl_test_status (iret)))
    {
        PSL_PUTDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_INITIALIZE (in INITIALIZE)", iret);
        PSL_SIGNAL_ERROR ((*tpa_ptr).diag_ptrps, PSLERRCLS_USER1, iret);
        return 0;           /* ...and return        */
    }
    /*
     *  Initialise the dependency linked list
     */
    INI_DEP_LL (com_ptr);

    return PSL_NORMAL;          /* ...and return success */
}

/************************************************************************/
/*                                  */
/*    INI_PARSE_SCRIPT - Initialize Script/Test TPARSE Data     */
/*    ================                          */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = INI_PARSE_SCRIPT (argblk_ptr);               */
/*                                  */
/*    Where:                                */
/*                                  */
/*      iret    = Unsigned Long PSL Routine Return Status       */
/*      = PSL_NORMAL if all O.k.                */
/*      = PSL_INSVIRMEM if error allocating structures      */
/*                                  */
/*  argblk_ptr    = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*    This routine is invoked before script/test parsing begins to  */
/*    initialize parsing data.                      */
/*                                  */
/*    This routine is called immediately after PSL_TPARSE has been  */
/*    called to parse a Script Source File. It initializes all      */
/*    variables that need to be reset before EACH compilation.      */
/*                                  */
/************************************************************************/
unsigned long INI_PARSE_SCRIPT (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com    *com_ptr;     /* Temp pointer to COM record */
    struct scl_tpa    *tpa_ptr;     /* Temp pointer to TPA record */
    struct scph   *header_ptr;      /* Temp pointer to SCPH record */
    struct scl_tofh   *tofh_ptr;        /* Temp pointer to SCL_TOFH record */
    unsigned  long    iret;         /* Status return variable */
    struct   scl_cmd  *scl_cmd_ptr;     /* Pointer to Command Line Struct */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */
    scl_cmd_ptr = (*com_ptr).cmd_ptrps;

    /*
     *    Perform TPARSE initialization that is common to ALL types of script.
     */
    iret = INI_PARSE_COMMON (argblk_ptr);
    if  (!(psl_test_status (iret)))
    {
        PSL_PUTDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"INI_PARSE_COMMON", iret);
        PSL_SIGNAL_ERROR ((*tpa_ptr).diag_ptrps, PSLERRCLS_USER1, iret);
        return 0;           /* ...and return */
    }

    /*
     *    Copy more pointers for ease of reference - cannot setup before.
     */
    header_ptr = (*tpa_ptr).header_begps;   /* Local pointer to 'scph' record */
    tofh_ptr = (*tpa_ptr).tofh_begps;       /* Local pointer to 'tofh' record */

    /*
     *    Initialize definition data.
     */
    INI_DEF (com_ptr);

    /*
     *    Set up the pointers to the various parsing structures.
     */
    (*tpa_ptr).def_tst_wrkps = &tst_wrk;
    /* Pointer to test data work record */
    (*tpa_ptr).def_tthr_wrkps = &tthr_wrk;
    /* Pointer to test thread */
    /* data work record */
    (*tpa_ptr).def_tstthr_wrkps = &tstthr_wrk;
    /* Pointer to test thread data work */
    /* record */
    (*tpa_ptr).variable_begps = variable_table;
    /* Pointer to variable symbol table */
    (*tpa_ptr).label_begps = label_table;
    /* Pointer to label symbol table  */
    (*tpa_ptr).subroutine_begps = subroutine_table;
    /* Pointer to sub. symbol table   */
    (*tpa_ptr).dostack_begps = do_table; /* Pointer to the DO table   */
    (*tpa_ptr).ifstack_begps = if_table; /* Pointer to the IF table */

    (*tpa_ptr).swstack_begps = sw_table; /* Pointer to SWITCH table     */
    (*tpa_ptr).code_begpc = code_buffer; /* Pointer to code buffer  */
    (*tpa_ptr).code_ptrpc = code_buffer; /* Pointer to code buffer  */
    (*tpa_ptr).code_endpc = (*tpa_ptr).code_ptrpc + MAX_SCP_SIZE - 1;
    /* Calculate end of code buffer */
    (*tpa_ptr).tofl_llps = NULL;    /* No code section 'tof' lines yet */
    (*tpa_ptr).values_begpc = (*tpa_ptr).code_ptrpc + MAX_SCP_SIZE;
    /* Calculate start of values buffer */
    (*tpa_ptr).values_endpc = (*tpa_ptr).values_begpc + OBJ_VALUES_SIZE;
    /* Calculate end of values buffer */
    (*tpa_ptr).values_nxtpc = (*tpa_ptr).values_begpc;
    /* Address next FREE in values buffer */
    (*tpa_ptr).timer_begps = timer_table;
    /* Pointer to timer table */
    (*tpa_ptr).shrvar_begps = shrvar_table;
    /* Pointer to shareable variables table */
    (*tpa_ptr).parameter_begps = parameter_table;
    /* Pointer to script parameter table */
    (*tpa_ptr).fat_begps = file_table;  /* Pointer to file access table */
    (*tpa_ptr).pipe_begps = pipe_table; /* Pointer to pipe access table */
    (*tpa_ptr).pipe_begps = pipe_table; /* Pointer to pipe access table */
    (*tpa_ptr).constant_begpc = constant_buffer;
    /* Pointer to constants buffer  */
    (*tpa_ptr).file_dtyppl = file_dtyp_buf;
    /* Pointer to file dtype table  */
    (*tpa_ptr).constant_data_offl = 0;  /* Reset constants table offset */

    /*
     *    Initialize global data.
     */
    (*tpa_ptr).lab_cntl = 0;        /* Zeroise labels count     */
    (*tpa_ptr).local_data_offl = SRV_PRMTBL_SIZE;
    /* Start offset in LDA is after */
    /* ...the parameter tavle.  */
    (*tpa_ptr).shrvar_data_offl = 0;    /* Reset shareable variable data offsets */

    (*tpa_ptr).script_data_offl = 0;    /* Reset script data offsets    */

    (*tpa_ptr).shrvar_rec_offl = 0;     /* Shareable record offset reset */

    (*tpa_ptr).tim_cntl = 0;        /* Zeroise timer count      */
    (*tpa_ptr).ifstack_levell = 0;  /* Zeroise IF command count     */
    (*tpa_ptr).ifdef_levell = 0;    /* Zeroise IFDEF command level  */
    (*tpa_ptr).do_cntl = 0;     /* Zeroise DO command count     */
    (*tpa_ptr).var_cntl = 0;        /* Zeroise variable count   */
    (*tpa_ptr).rec_cntl = 0;        /* Zeroise record count     */
    (*tpa_ptr).script_val_cntl = 0; /* Zeroise script variables with*/
    /* associated values count  */
    (*tpa_ptr).sw_depthl = 0;           /* Clear switch level count     */
    (*tpa_ptr).case_countl = 0;           /* Clear case level count     */
    (*tpa_ptr).sw_startl = 0;       /* Clear parsed switch indicator*/
    (*tpa_ptr).break_fgl = 0;       /* Clear parsed break indicator */
    (*tpa_ptr).sw_error_fgl = 0;    /*Clear parsed switch error flag*/
    (*tpa_ptr).do_error_fgl = 0;    /*Clear parsed do error flag    */
    (*tpa_ptr).not_flagl = 0;           /*Clear parsed NOT flag         */
    (*tpa_ptr).test_fgl = 0;        /* Clear parsed testcase indicator */
    (*tpa_ptr).test_case_warnl = 0; /* Clear parsed testcase indicator */
    (*tpa_ptr).sql_warnn = PFALSE;  /* Clear SQL warning flag */
    (*tpa_ptr).script_flagl = 0;    /* Clear parsed call script flag*/
    (*tpa_ptr).val_cntl = 0;        /* Clear value list/range count */
    (*tpa_ptr).random_val_cntl = 0; /* Clear random value count     */
    (*tpa_ptr).shrvar_cntl = 0;     /* Clear shareable variable count */
    (*tpa_ptr).subrtn_cntl = 0;     /* Clear subroutine symtbl entry count */

    (*tpa_ptr).subrtn_def_cntl = 0; /* Clear subroutine definition count */

    (*tpa_ptr).scpprm_cntl = 0;     /* Clear script parameter count */
    (*tpa_ptr).fat_cntl = 0;        /* Clear file access count  */
    (*tpa_ptr).pipe_cntl = 0;       /* Clear pipe access count  */
    (*tpa_ptr).nonlocal_input_flagl = 0;/* shareable/script variable flag */
    (*tpa_ptr).src_fil_indxl = 0;   /* Zeroize file index       */
    (*tpa_ptr).wait_unitl = WAIT_SECONDS;
    /* Default script WAIT UNIT to SECONDS */
    (*header_ptr).scph_inpdly_w = DEF_INP_DELAY;
    /* Input rate = default     */
    (*header_ptr).scph_prot_defaulted_n = PTRUE;
    /* Set Protection Defaulted */
    (*header_ptr).scph_sprot_l = 0xFFFFFFFF;
    /* Set Protection Mask - NO ACCESS */
    /*
     *  Initialise script object ('tof') file header data.
     */
    (*tofh_ptr).tofh_scols_w = 132; /* Maximum screen columns = 132 */
    (*tofh_ptr).tofh_sline_w = 24;  /* Maximum screen lines = 24 */
    (*tofh_ptr).tofh_osspcl = TOF_GENSTA_FALSE;
    /* No operating system variants yet */
    (*tofh_ptr).tofh_pltspcl = TOF_GENSTA_FALSE;
    /* No platform variants yet */
    (*header_ptr).scph_mode_b  = 0 ;    /* Clear mode flag in header    */
    /*
     * Initialise data for Record/variable alignment
     * Initialise data about last variable
     */
    (*tpa_ptr).shrvar_align_offl = 0;
    (*tpa_ptr).local_align_offl = 0;
    (*tpa_ptr).script_align_offl = 0;

    /*
     *  Initialize 'tof' line buffer linked list.
     */
    INI_TOFL_LL (com_ptr);

    /*
     *    Perform test script specific initialisation.
     */
    (*tpa_ptr).tst_thisps = &thistst;   /* Pointer to data for the test file */
    (void) INI_TST (argblk_ptr);    /* Initialise test data work structures */
    (*tpa_ptr).tst_begps = tst_table;   /* Pointer to Test Symbol Table */
    (*tpa_ptr).tst_cntl = 0;        /* Zero test count */
    (*tpa_ptr).tstthr_begps = tstthr_table;
    /* Pointer to Test Thread Symbol Table*/
    (*tpa_ptr).tstthr_cntl = 0;     /* Zero test thread count */
    (*tpa_ptr).subtst_cntl = 0;     /* Clear sub-test count */

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                  */
/*    INI_PARSE_THREAD - Initialize Thread TPARSE data          */
/*    ================                          */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = INI_PARSE_THREAD (argblk_ptr);               */
/*                                  */
/*    Where:                                */
/*                                  */
/*      iret    = Unsigned Long PSL Routine Return Status       */
/*      = PSL_NORMAL if all O.k.                */
/*      = PSL_INSVIRMEM if error allocating structures      */
/*                                  */
/*  argblk_ptr    = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*    This routine is invoked before thread script parsing begins to    */
/*    initialize parsing data.                      */
/*                                  */
/*    This routine is called immediately after PSL_TPARSE has been  */
/*    called to parse a Thread Script Source File. It initializes all   */
/*    variables that need to be reset before EACH compilation.      */
/*                                  */
/************************************************************************/
unsigned long INI_PARSE_THREAD (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct    scl_com *com_ptr;     /* Temp pointer to COM record   */
    struct    scl_tpa *tpa_ptr;     /* Temp pointer to TPA record   */
    struct    scph    *header_ptr;      /* Temp pointer to SCPH record  */
    unsigned  long    iret;         /* Status return variable   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */
    header_ptr = (*tpa_ptr).header_begps;
    /* ...local pointer to header record */

    /*
     *    Perform TPARSE initialization that is common to ALL types of script.
     */
    iret = INI_PARSE_COMMON (argblk_ptr);
    if  (!(psl_test_status (iret)))
    {
        PSL_PUTDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"INI_PARSE_COMMON", iret);
        PSL_SIGNAL_ERROR ((*tpa_ptr).diag_ptrps, PSLERRCLS_USER1, iret);
        return 0;           /* ...and return */
    }

    /*
     *    Set up the pointers to the various parsing structures.
     */
    (*tpa_ptr).def_thr_wrkps = &thr_wrk;
    /* Pointer to thread data work record */
    (*tpa_ptr).def_tdp_wrkps = &tdp_wrk;
    /* Pointer to thread parameter */
    /* data work record */
    (*tpa_ptr).thr_begps = thr_table;   /* Pointer to Thread Symbol Table */

    /*
     *    Initialize thread script definition data.
     */
    INI_THR_DEF (com_ptr);

    return PSL_NORMAL;          /* ...and return success */
}

/************************************************************************/
/*                                  */
/*    INI_TOFL_LL - Initialize TOF line linked list         */
/*    ===========                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  INI_TOFL_LL (com_ptr);                      */
/*                                  */
/*    Where:                                */
/*                                  */
/*  com_ptr    = Pointer to SCL common data.            */
/*                                  */
/*    This routine deletes any existing 'tof' line linked list -    */
/*    including all associate memory - and allocates a new one.     */
/*                                  */
/************************************************************************/
static void INI_TOFL_LL (struct scl_com *com_ptr)
{
    /*
     *  Local data.
     */
    struct scl_tpa   *tpa_ptr;      /* Pointer to SCL Tparse data */
    struct scl_tofl  *tofl_ptr;     /* Pointer TOF line data */
    unsigned  long   iret;          /* PSL routine status */

    /*
     *  Copy pointers for ease of reference.
     */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  If we have a TOF line linked list, delete it.
     */
    if ((*tpa_ptr).tofl_llps != NULL)
    {
        do
        {
            iret = PSL_VM_REMOVE ((*tpa_ptr).tofl_llps,
                                  (char *) &tofl_ptr,
                                  NULL, NULL, PSL_VM_POS_BEG);
            if (psl_test_status (iret))
            {
                FREE (tofl_ptr->tofl_linptrpc);
                FREE (tofl_ptr);

#ifdef PSL_TRACING
                if (PSL_TEST_TRACE (PTRC_USR_USER5))
                {
                    PSL_OUTPUT_TRACE ("FREE: tofl - '%p' - INI_TOFL_LL\n",
                                      tofl_ptr);
                }
#endif  /* PSL_TRACING */

            }
        }
        while (psl_test_status (iret));     /* Loop while not an error */

        (void) PSL_VM_DELETE ((*tpa_ptr).tofl_llps);
        (*tpa_ptr).tofl_llps = NULL;
    }

    /*
     *  Now allocate the new list.
     */
    iret = PSL_VM_INITIALIZE (sizeof (long *), &(*tpa_ptr).tofl_llps);
    if  (!(psl_test_status (iret)))
    {
        PSL_PUTDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_INITIALIZE (in INI_TOFL_LL)", iret);
        PSL_SIGNAL_ERROR ((*tpa_ptr).diag_ptrps, PSLERRCLS_USER1, iret);
        return;
    }

    return;
}

/************************************************************************/
/*                                  */
/*    SCL_LIST_LINE - Output (Source) Line to Listing File      */
/*    =============                         */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  SCL_LIST_LINE (com_ptr, buf_ptr);               */
/*                                  */
/*    Where:                                */
/*                                  */
/*  com_ptr    = Pointer to Common Data Structure           */
/*  buf_ptr    = Pointer to Buffer to write             */
/*                                  */
/*  This routine outputs a line from the source file to the     */
/*  listing file.                           */
/*                                  */
/************************************************************************/

void    SCL_LIST_LINE (struct scl_com *com_ptr,
                       char *buf_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long    iret;          /* Status Return        */
    struct   scl_cmd *scl_cmd_ptr;      /* Pointer to Command Line Struct */

    /*
     *    Copy pointers for ease of reference.
     */
    scl_cmd_ptr = (*com_ptr).cmd_ptrps;

    /*
     *  Print the line if the Listing option is active.
     */
    if ((*scl_cmd_ptr).listn)
    {
        iret = LIST_TESTPAGE ((*com_ptr).lst_contextps, 3);
        if (!(psl_test_status (iret))) goto err_exit;

        iret = LIST_LINE_COLUMN_ONE ((*com_ptr).lst_contextps, buf_ptr);
        if (!(psl_test_status (iret)) ) goto err_exit;

    }

    goto end;

    /*
     *  Report error from Listing routine and exit.
     */
err_exit:
    PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                 1,
                 iret);
    PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
    PSL_EXIT ();

end:
    return;
}

/************************************************************************/
/*                                  */
/*    LIST_VARIABLES - Write variable information to Listing File   */
/*    ==============                            */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  LIST_VARIABLES (com_ptr);                   */
/*                                  */
/*    Where:                                */
/*                                  */
/*  com_ptr    = Pointer to Common Data Structure           */
/*                                  */
/*    This routine outputs information about the variable usage of a    */
/*    script to the listing file.                   */
/*                                  */
/************************************************************************/

void    LIST_VARIABLES (struct scl_com *com_ptr)

{
    /*
     *    Local data
     */
    struct   scl_tpa *tpa_ptr;      /* Temp pointer to TPA record   */
    struct   symtbl  *var_ptr;      /* Symbol table entry for variable */
    unsigned long    isub;          /* Loop control variable    */
    char    listing_line  [80];    /* Listing line         */
    static   char *datatypes [] = {"NONE", "INTEGER", "CHARACTER", "CON INTEGER"
                                   ,"CON CHARACTER", "VP$SCREEN","FILE","OPERATOR"
                                   ,"PIPE","RECORD"};
    /* Data types array     */
    static   char *addrmodes [] = {"IMMEDIATE", "LOCAL", "GLOBAL", "VP$OUTPUT"
                                   ,"NONE", "SCRIPT", "TIMER", "EXTRACT", "CONSTANT"
                                   ,"VP$SCREEN", "VP$LINE", "VP$COLUMN", "VP$RESPONSE"
                                   ,"VP$STATUS", "LOCATE", "THREAD" ,"FILE" ,"OPERATOR"
                                   ,"PIPE", "VP$RESULT", "TEST", "THREAD", "TEST-WIDE"};
    /* Addressing modes array.  */
    /*
     *    Dereference the pointers
     */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* ...and TPARSE data pointer   */

    /*
     *    Set up the sub-headings for the Variable listing page and throw a new
     *    page to force them to be output.
     */
    (void) LIST_SUBHEADING ((*com_ptr).lst_contextps
                            ,"*****  SCL VARIABLE USAGE  *****"
                            ,"---------------------------------"
                            ,"  VARIABLE NAME     DATATYPE       LENGTH    SCOPE"
                            ,"  =============     ========       ======    =====");

    (void) LIST_TESTPAGE ((*com_ptr).lst_contextps, -1);
    /* Force a page break       */


    /*
     *    Scan the variables table listing each Variable found unless there are
     *    none in which case we should report this
     */
    if ((*tpa_ptr).var_cntl == 1)   /* If only VP$STATUS defined    */
    {
        (void) LIST_LINE_COLUMN_ONE ((*com_ptr).lst_contextps,
                                     "NO USER VARIABLES DEFINED\0");
        /* ...report no variables defined */
    }

    else                /* ...else if variables defined */
    {
        for (isub = 1, var_ptr = (*tpa_ptr).variable_begps + 1;
                isub < (*tpa_ptr).var_cntl;
                isub++, var_ptr++)     /* Loop through the variables table */
        {
            (void) sprintf (listing_line,
                            "  %-16.*s  %-13s  %5ld     %-9s",
                            (int) (*var_ptr).var_nlengthl,
                            (*var_ptr).var_namec16,
                            datatypes [(*var_ptr).var_typel],
                            (*var_ptr).var_sizel,
                            addrmodes [(*var_ptr).var_addrl]);
            /* List the variables details   */
            /*
             *    ...and write this information to the listing file column one
             */
            (void) LIST_LINE_COLUMN_ONE ((*com_ptr).lst_contextps,
                                         listing_line);
        }
    }

    return;
}

/************************************************************************/
/*                                  */
/*    LIST_TIMERS - Write timer information to Listing File     */
/*    ===========                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  LIST_TIMERS (com_ptr);                      */
/*                                  */
/*    Where:                                */
/*                                  */
/*  com_ptr    = Pointer to Common Data Structure           */
/*                                  */
/*    This routine outputs information about the timer usage of a   */
/*    script to the listing file.                   */
/*                                  */
/************************************************************************/

void    LIST_TIMERS (struct scl_com *com_ptr)

{
    /*
     *    Local data
     */
    struct   scl_tpa *tpa_ptr;      /* Temp pointer to TPA record   */
    struct   timtbl  *tim_ptr;      /* Symbol table entry for timer */
    unsigned long    isub;          /* Loop control variable    */
    char    listing_line [80]; /* Listing line         */

    /*
     *    Dereference the pointers
     */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* ...and TPARSE data pointer   */

    /*
     *    Set up the sub-headings for the Timer listing page and throw a new
     *    page to force them to be output.
     */
    (void) LIST_SUBHEADING ((*com_ptr).lst_contextps
                            ,"*****  SCL TIMER USAGE  *****"
                            ,"------------------------------"
                            ," "
                            ," ");

    (void) LIST_TESTPAGE ((*com_ptr).lst_contextps, -1);
    /* Force a page break       */

    /*
     *    Scan the timers table listing each Timer found unless there are none in
     *    case we should report this
     */
    if ((*tpa_ptr).tim_cntl == 0)   /* If no timers         */
    {
        (void) LIST_LINE_COLUMN_ONE ((*com_ptr).lst_contextps,
                                     "NO TIMERS DEFINED\0");
    }

    else
    {
        for (isub = 0, tim_ptr = (*tpa_ptr).timer_begps;
                isub < (*tpa_ptr).tim_cntl;
                isub++, tim_ptr++)     /* Loop through the Timers table */
        {
            (void) sprintf (listing_line, "%64.s", (*tim_ptr).tim_namec68);
            (void) LIST_LINE_COLUMN_ONE ((*com_ptr).lst_contextps, listing_line);
        }
    }

    return;
}
