/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

////////////////////////////////////////////////////////////////////////////////
// @doc	    TCommonRcrd.h
// @module  TCommonRcrd | This module defines the interface to all classes
//          derived from the TRecord class. These classes handle the processing
//	    of TOF records.
///////////////////////////////////////////////////////////////////////////////

#ifndef CYR_TCMNRECORDS
#define CYR_TCMNRECORDS

#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include <map>
#include <list>
#include <string>

#ifdef unix
#   elif defined (WIN32)
using namespace std;
#else //unknown
#   error Unknown Platform
#endif // unknown

#include "CECTofTxt.h"
#include "TRecord.h"
#include "TExpression.h"
//#include "mdbmtx.h"
#include "tofprmdef.h"

class TContext;

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdFlowCtrl | This is an abstract class that provides generic
//	    flow control facilities common to LOOPs, IF-THENs and ELSE[-IF]s.
//
// @base    public | TRecord.
//
class TRcrdFlowCtrl : public TRecord
{
    // @access Public Members
public:
    // @cmember Constructor
    TRcrdFlowCtrl(const TParamRule* headRuleP)
        : TRecord(headRuleP), _startIndex(-1), _endIndex(-1)
    {}

    // @cmember Access _startIndex

    int StartIndex() const
    {
        return _startIndex;
    }
    // @cmember Access _endIndex
    int EndIndex()   const
    {
        return _endIndex;
    }
    // @cmember Set _endIndex
    void SetEndIndex (int index)
    {
        _endIndex = index;
    }

    // @access Protected Members
protected:
    // @cmember
    int _startIndex;
    // @cmember
    int	_endIndex;

private:
    //// Forbidden
    TRcrdFlowCtrl();
}
; // TRcrdFlowCtrl

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdLoop | This class is used to process Loop TOF records.
//
// @base    public | TRcrdFlowCtrl.
//
class TRcrdLoop : public TRcrdFlowCtrl
{
    // @access Public Members
public:
    static const AtvfKey KEY;

    // @cmember Constructor
    TRcrdLoop(const TParamRule* headRuleP) : TRcrdFlowCtrl(headRuleP),
        _initialExprP(NULL), _endExprP(NULL), _stepExprP(NULL),_varP(NULL)
    {}

    // @cmember Destructor.
    virtual ~TRcrdLoop()
    {
        delete _initialExprP;
        delete _endExprP;
        delete _stepExprP;
        if(_varP)delete _varP;
    }

    // @cmember Execute record in current context and return status.
    virtual int execute(TContext& ctxt);
    // @cmember Prepare record for execution and return status.
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);

    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdLoop(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Protected Members.
protected:
    // @cmember Pointer to loop control variable object.
    TtVarRef* _varP;
    // @cmember Pointer to control variable initialisation expression object.
    CECTofTxt* _initialExprP;
    // @cmember Pointer to loop termination value expression object.
    CECTofTxt* _endExprP;
    // @cmember Pointer to loop step expression object.
    CECTofTxt* _stepExprP;
    // @cmember Lock for member data

    // @access Private Members
private:
    // @cmember
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdLoop();
    TRcrdLoop(const TRcrdLoop&);
    operator=(const TRcrdLoop&);
}
; // TRcrdLoop

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdAbort | This class is used to process Abort (e.g. Detach
//          Thread) TOF records.
//
// @base    public | TRecord.
//
class TRcrdAbort : public TRecord
{
    // @access Public Members
public:
    // @cmember Conditional record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor.
    TRcrdAbort(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // @cmember Execute record in current context and return status.
    virtual int execute(TContext& ctxt);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdAbort(*this);
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Private Members
private:
    // @cmember
    static bool _registeredCreator;
};

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdCond | This class is used to process Condition (If ... Goto)
//	    TOF records.
//
// @base    public | TRecord.
//
class TRcrdCond : public TRecord
{
    // @access Public Members
public:
    // @cmember Conditional record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor.
    TRcrdCond(const TParamRule* headRuleP) : TRecord(headRuleP),
        _labelIndex(-1), _exprP(NULL)
    {}
    // @cmember Destructor.

    virtual ~TRcrdCond()
    {
        delete _exprP;
    }

    // @cmember Execute record in current context and return status.
    virtual int execute (TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.

    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdCond(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Protected Members
protected:
    // @cmember Destination label index in script.
    int _labelIndex;
    // @cmember Pointer to conditional expression object.
    CECTofTxt* _exprP;

    // @access Private Members
private:
    // @cmember
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdCond();
    TRcrdCond(const TRcrdCond&);
    operator=(const TRcrdCond&);
}
; // TRcrdCond

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdConvert | This class is used to process Convert TOF records.
//
// @base    public | TRecord.
//
class TRcrdConvert : public TRecord
{
    // @access Public Members
public:
    // @cmember Convert record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor
    TRcrdConvert(const TParamRule* headRuleP) : TRecord(headRuleP),
        _srcVarP(NULL), _tgtVarP(NULL), _radixExprP(NULL), _toInteger(false),
        _labelIndex(-1)
    {}
    // @cmember Destructor.

    virtual ~TRcrdConvert()
    {
        delete _srcVarP;
        delete _tgtVarP;
        if(_radixExprP)
            delete _radixExprP;
    }

    // @cmember Execute record in current context and return status.
    virtual int execute(TContext& ctxt);
    // @cmember Prepare record for execution and return status.
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.

    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdConvert(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Protected Members.
protected:
    // @cmember Pointer to source variable object.
    TtVarRef* _srcVarP;
    // @cmember Pointer to target variable object.
    TtVarRef* _tgtVarP;
    // @cmember Pointer to conversion radix expression object.
    CECTofTxt* _radixExprP;
    // @cmember Destination label index in script.
    int _labelIndex;
    // @cmember Convert to integer flag - identifies conversion type.
    bool _toInteger;

    // @access Private Members
private:
    // @cmember
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdConvert();
    TRcrdConvert(const TRcrdConvert&);
    operator=(const TRcrdConvert&);
}
; // TRcrdConvert

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdElse | This class is used to process Else (Else/ElseIf ...)
//	    TOF records.
//
// @base    public | TRcrdFlowCtrl
//
class TRcrdElse : public TRcrdFlowCtrl
{
    // @access Public Members
public:
    // @cmember Else record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor
    TRcrdElse(const TParamRule* headRuleP) : TRcrdFlowCtrl(headRuleP),
    _ifThenElse(NULL), _exprP(NULL)
    {}
    // @cmember Destructor.

    virtual ~TRcrdElse()
    {
        delete _exprP;
    }

    // @cmember Execute record in current context and return status.
    virtual int execute(TContext& ctxt);
    // @cmember Prepare record for execution and return status.
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);

    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdElse(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

    // @cmember
    void SetEndIndex (int index)
    {
        _endIndex = index;
    }

    // @access Protected Members
protected:
    // @cmember Pointer to parent IF-THEN/ELSE-IF record
    TRcrdFlowCtrl* _ifThenElse;
    // @cmember Pointer to boolean expression object - used to process ELSE-IF.
    CECTofTxt* _exprP;

    // @access Private Members
private:
    // @cmember
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdElse();
    TRcrdElse(const TRcrdElse&);
    operator=(const TRcrdElse&);
}
; // TRcrdElse

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdEndif | This class is used to process Endif (Endif)
//	    TOF records.
//
// @base    public | TRecord.
//
class TRcrdEndIF : public TRecord
{
    // @access Public Members
public:
    // @cmember Eif record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor
    TRcrdEndIF(const TParamRule* headRuleP) : TRecord(headRuleP), _ifThenElse(NULL)
    {}

    // @cmember Execute record in current context and return status.

    virtual int execute(TContext& ctxt);
    // @cmember Prepare record for execution and return status.
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);

    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdEndIF(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Protected Members
protected:
    // @cmember
    TRcrdFlowCtrl* _ifThenElse;

    // @access Private Members
private:
    // @cmember
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdEndIF();
    TRcrdEndIF(const TRcrdEndIF&);
    operator=(const TRcrdEndIF&);
}
; // TRcrdEndIF

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdEndLoop | This class is used to process Eloop (End loop) TOF
//	    records.
//
// @base    public | TRecord.
//
class TRcrdEndLoop : public TRecord
{

    // @access Public Member.
public:
    // @cmember Eloop record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor.
    TRcrdEndLoop(const TParamRule* headRuleP) : TRecord(headRuleP), _mainLoop(NULL)
    {}

    // @cmember Execute record in current context and return status.

    virtual int execute(TContext& ctxt);
    // @cmember Prepare record for execution and return status.
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);

    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdEndLoop(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Protected Members.
protected:
    TRcrdLoop* _mainLoop;

    // @access Private Members.
private:
    // @cmember
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdEndLoop();
    TRcrdEndLoop(const TRcrdEndLoop&);
    operator=(const TRcrdEndLoop&);
}
; // TRcrdEndLoop

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdFormat | This class is used to process Format TOF records.
//
// @base    public | TRecord.
//
class TRcrdFormat : public TRecord
{
    // @access Public Members.
public:
    static const AtvfKey KEY;
    // @cmember Constructor.
    TRcrdFormat(const TParamRule* headRuleP) : TRecord(headRuleP),
        _varP(0), _valListP(0), _fmtExprP(0)
    {}
    // @cmember Default destructor.

    virtual ~TRcrdFormat()
    {
        if(_varP)
            delete _varP;
        if(_fmtExprP)
            delete _fmtExprP;
        if(_valListP)
            delete _valListP;
    }

    // @cmember Execute record in current context and return status.

    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdFormat(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Private Members.
private:
    // @cmember Pointer to variable object for variable to hold formatted string.
    TtVarRef* _varP;
    // @cmember Pointer to value list object.
    TtValList* _valListP;
    // @cmember Pointer to format string expression object.
    TtChrExpr* _fmtExprP;

    // @cmember
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdFormat();
    TRcrdFormat(const TRcrdFormat&);
    operator=(const TRcrdFormat&);
}
; // TRcrdFormat

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdIF_Then | This class is used to process IfThen (If ... Then)
//	    TOF records.
//
// @base    public | TRecord.
//
class TRcrdIF_Then : public TRcrdFlowCtrl
{
    // @ access Public Members
public:
    // @cmember Else record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor
    TRcrdIF_Then(const TParamRule* headRuleP) : TRcrdFlowCtrl(headRuleP),
        _exprP(NULL)
    {}
    // @cmember Destructor

    virtual ~TRcrdIF_Then()
    {
        delete _exprP;
    }

    // @cmember Execute record in current context and return status.
    virtual int execute(TContext& ctxt);
    // @cmember Prepare record for execution and return status.
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);

    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdIF_Then(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Protected members
protected:
    // @cmember Pointer to boolean expression object - used to process IF-THEN.
    CECTofTxt* _exprP;

    // @access Private Members
private:
    // @cmember
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdIF_Then();
    TRcrdIF_Then(const TRcrdIF_Then&);
    operator=(const TRcrdIF_Then&);
}
; // TRcrdIF_Then

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdLabel | This class is used to process Label (Label)
//	    TOF records.
//
// @base    public | TRecord.
//
class TRcrdLabel : public TRecord
{
    // @access Public Members
public:
    // @cmember Label record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor
    TRcrdLabel(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // @cmember Execute record in current context and return status.

    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdLabel(*this);
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Private Members
private:
    // @cmember
    static bool _registeredCreator;
}
; // TRcrdLabel

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdLoad | This class is used to process Load TOF records.
//
// @base    public | TRecord.
//
class TRcrdLoad : public TRecord
{
    // @access Public Members
public:
    // @cmember Label record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor
    TRcrdLoad(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // @cmember Execute record in current context and return status.

    virtual int execute(TContext& ctxt);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdLoad(*this);
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Private Members
private:
    // @cmember
    static bool _registeredCreator;
}
; // TRcrdLoad

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdLogmsg | This class is used to process Logmsg (Log message)
//	    TOF records.
//
// @base    public | TRecord.
//
class TRcrdLogmsg : public TRecord
{
    // @access Public Members
public:
    // @cmember Logmsg record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor
    TRcrdLogmsg(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // @cmember Execute record in current context and return status.

    virtual int execute(TContext&);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdLogmsg(*this);
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Private Members
private:
    // @cmember
    static bool _registeredCreator;
}
; // TRcrdLogmsg

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdSetvar | This class is used to process Set Variable (Setvar)
//	    TOF records.
//
// @base    public | TRecord.
//
class TRcrdSetvar : public TRecord
{
    // @access Public Members.
public:
    // @cmember Set variable record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor.
    TRcrdSetvar(const TParamRule* headRuleP) : TRecord(headRuleP), _exprP(NULL)
    {}

    // @cmember Destructor.
    virtual ~TRcrdSetvar ()
    {
        if(_exprP)
            delete _exprP;
        if(_varP)
            delete _varP;
    }

    // @cmember Execute record in current context and return status.
    virtual int execute (TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.

    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdSetvar(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

    // @access Private Members
private:
    // @cmember Destination label index in script
    int	_labelIndex;
    // @cmember Pointer to set variable 'name' object.
    TtVarRef* _varP;
    // @cmember Pointer to set variable 'expression' object.
    CECTofTxt* _exprP;
    // @cmember
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdSetvar();
    TRcrdSetvar(const TRcrdSetvar&);
    operator=(const TRcrdSetvar&);
}
; // TRcrdSetvar

/**
 * Definition of Genvar Record class
 **/
class TRcrdGenvar : public TRecord
{
public:
    static const AtvfKey KEY;
    // Constructors & Destructor
    TRcrdGenvar(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    virtual ~TRcrdGenvar()
    {}

    // Accessors

    virtual int execute(TContext&);

    virtual TRecord* duplicate() const
    {
        return new TRcrdGenvar(*this);
    }
    static const bool _RegisterCreator();

private:

    static bool _registeredCreator;
}
; // TRcrdGenvar
/**
 * Definition of Seed Record class
 **/
class TRcrdSeed : public TRecord
{
public:
    static const AtvfKey KEY;
    // Constructors & Destructor
    TRcrdSeed(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    virtual ~TRcrdSeed()
    {}

    // Accessors

    virtual int execute(TContext&);

    virtual TRecord* duplicate() const
    {
        return new TRcrdSeed(*this);
    };
    static const bool _RegisterCreator();

private:
    static bool _registeredCreator;
}
; // TRcrdSeed



// *****************************************************************************
// * SYNCHRONISATION
// *****************************************************************************

/**
 * Definition of Wait Record class
 **/
class TRcrdWait : public TRecord
{
public:
    static const AtvfKey KEY;
    // Constructors & Destructor
    TRcrdWait(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // Accessors

    virtual int execute(TContext&);
    virtual TRecord* duplicate() const
    {
        return new TRcrdWait(*this);
    }
    static const bool _RegisterCreator();

private:
    static bool _registeredCreator;
}
; // TRcrdWait



/**
 * Definition of BreakContinue Record class
 **/
class TRcrdBreakContinue : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdBreakContinue(const TParamRule* headRuleP) : TRecord(headRuleP), _expr(NULL)
    {}

    virtual ~TRcrdBreakContinue()
    {
        delete _expr;
    }

    // Control methods
    virtual int execute(TContext& ctxt);
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);

    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdBreakContinue(*this);
        assert(0); return NULL;
    }
    static const bool _RegisterCreator();

protected:
    TExpression* _expr;

private:
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdBreakContinue();
    TRcrdBreakContinue(const TRcrdBreakContinue&);
    operator=(const TRcrdBreakContinue&);
}
; // TRcrdBreakContinue

class TRcrdGoto : public TRecord
{

public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdGoto(const TParamRule* headRuleP) : TRecord(headRuleP), _labelIndex(-1)
    {}

    virtual int execute(TContext& ctxt);
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);

    virtual TRecord* duplicate() const
    {
        return new TRcrdGoto(*this);
    }
    static const bool _RegisterCreator();

protected:
    int	_labelIndex;			// Destination label index in script

private:
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdGoto();
}
; // TRcrdGoto


//
// TRcrdSubroutine
class TRcrdSubroutine : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdSubroutine(const TParamRule* headRuleP)
        : TRecord(headRuleP)
    {}

    // Control methods

    virtual int execute(TContext& ctxt);
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);

    virtual TRecord* duplicate() const
    {
        return new TRcrdSubroutine(*this);
    }
    static const bool _RegisterCreator();

private:
    static bool _registeredCreator;
}
; // TRcrdSubroutine

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdCsub | This class is used to process Csub (Call Subroutine)
//	    TOF records.
//
// @base    public | TRecord.
//
class TRcrdCsub : public TRecord
{
    // @access Public Members
public:
    // @cmember Csub record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor
    TRcrdCsub(const TParamRule* headRuleP) : TRecord(headRuleP),
        _labelIndex(-1), _returnIndex(-1), _prmListP(0)
    {}
    virtual ~TRcrdCsub()
    {
        if(_prmListP)
            delete _prmListP;
    }


    // @cmember Execute record in current context and return status.
    virtual int execute(TContext& ctxt);
    // @cmember Prepare record for execution and return status.
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);

    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdCsub(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

    // @cmember Access - returnIndex.
    int ReturnIndex() const
    {
        return _returnIndex;
    }

    // @access Protected Members
protected:
    // @cmember Destination label index in script.
    int _labelIndex;
    // @cmember Index of TRecord to process after return from call.
    int _returnIndex;
    // @cmember Pointer to parameter list object.
    TtValList* _prmListP;

    // @access Private Members
private:
    // @cmember
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdCsub();
    TRcrdCsub(const TRcrdCsub&);
    operator=(const TRcrdCsub&);
}
; // TRcrdCsub

//
// Return Record from Subroutine
class TRcrdReturn : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdReturn(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // Control methods

    virtual int execute(TContext& ctxt);

    virtual TRecord* duplicate() const
    {
        return new TRcrdReturn(*this);
    }
    static const bool _RegisterCreator();

private:
    static bool _registeredCreator;
}
; // TRcrdReturn

// *****************************************************************************
// * ERROR HANDLING/TERMINATION COMMANDS
// *****************************************************************************
// On Error Goto
class TRcrdGoError : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdGoError(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // Control methods

    virtual int execute(TContext& ctxt);

    virtual TRecord* duplicate() const
    {
        return new TRcrdGoError(*this);
    }
    static const bool _RegisterCreator();

private:

    static bool _registeredCreator;
}
; // TRcrdGoError

//
// Disable (Thread/Error etc)
class TRcrdDisable : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdDisable(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // Control methods

    virtual int execute(TContext& ctxt);

    virtual TRecord* duplicate() const
    {
        return new TRcrdDisable(*this);
    }
    static const bool _RegisterCreator();

private:

    static bool _registeredCreator;
}
; // TRcrdDisable

// Exit Thread Command
class TRcrdExit : public TRecord
{

public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdExit(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // Control methods

    virtual int execute(TContext& ctxt);

    virtual TRecord* duplicate() const
    {
        return new TRcrdExit(*this);
    }
    static const bool _RegisterCreator();

private:

    static bool _registeredCreator;
}
; // TRcrdExit

// *****************************************************************************
// * FILE HANDLING
// *****************************************************************************

//
// TRcrdOpen
class TRcrdOpen : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdOpen(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // Control methods

    virtual int execute(TContext& ctxt);

    virtual TRecord* duplicate() const
    {
        return new TRcrdOpen(*this);
    }
    static const bool _RegisterCreator();

private:
    static bool _registeredCreator;
}
; // TRcrdOpen


//
// TRcrdRead
class TRcrdRead : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdRead(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}
    virtual ~TRcrdRead()
    {
        if(_varP)
            delete _varP;
    }

    // Control methods

    virtual int execute(TContext& ctxt);
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);

    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdRead(*this);
        assert(0); return NULL;
    }
    static const bool _RegisterCreator();

private:
    static bool _registeredCreator;
    TtVarRef* _varP;

private:
    //// Forbidden
    TRcrdRead();
    TRcrdRead(const TRcrdRead&);
    operator=(const TRcrdRead&);
}
; // TRcrdRead


//
// TRcrdWrite
class TRcrdWrite : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdWrite(const TParamRule* headRuleP) : TRecord(headRuleP), _exprP(NULL)
    {}
    virtual ~TRcrdWrite()
    {
        if(_exprP)
            delete _exprP;
    }

    // Control methods

    virtual int execute(TContext& ctxt);
    virtual int prepare(TScript* script, int objIndex, bool verbose=false);

    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdWrite(*this);
        assert(0); return NULL;
    }
    static const bool _RegisterCreator();

private:
    static bool _registeredCreator;
    TtChrExpr* _exprP;

private:
    //// Forbidden
    TRcrdWrite();
    TRcrdWrite(const TRcrdWrite&);
    operator=(const TRcrdWrite&);
}
; // TRcrdWrite


//
// TRcrdFcntl
class TRcrdFcntl : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdFcntl(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // Control methods

    virtual int execute(TContext& ctxt);

    virtual TRecord* duplicate() const
    {
        return new TRcrdFcntl(*this);
    }
    static const bool _RegisterCreator();

private:
    static bool _registeredCreator;
}
; // TRcrdFcntl



// *****************************************************************************
// * FORMAL TEST CONTROL
// *****************************************************************************

//
// TRcrdTcstart
class TRcrdTcstart : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdTcstart(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // Control methods

    virtual int prepare(TScript* script, int objIndex, bool verbose=false);
    virtual int execute(TContext& ctxt);

    virtual TRecord* duplicate() const
    {
        return new TRcrdTcstart(*this);
    }
    static const bool _RegisterCreator();

private:
    static bool _registeredCreator;
}
; // TRcrdTcstart

//
// TRcrdTcterm
class TRcrdTcterm : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdTcterm(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // Control methods

    virtual int prepare(TScript* script, int objIndex, bool verbose=false);
    virtual int execute(TContext& ctxt);

    virtual TRecord* duplicate() const
    {
        return new TRcrdTcterm(*this);
    }
    static const bool _RegisterCreator();

protected:
    int _labelIndex;			// Destination label index in script
private:
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdTcterm();
}
; // TRcrdTcterm

//
// TRcrdTcend
class TRcrdTcend : public TRecord
{
public:
    static const AtvfKey KEY;

    // Constructors & Destructor
    TRcrdTcend(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}

    // Control methods

    virtual int prepare(TScript* script, int objIndex, bool verbose=false);
    virtual int execute(TContext& ctxt);

    virtual TRecord* duplicate() const
    {
        return new TRcrdTcend(*this);
    }
    static const bool _RegisterCreator();

private:
    static bool _registeredCreator;
}
; // TRcrdTcend


#endif // CYR_TCMNRECORDS
