/* Copyright (C) 2000-2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
// TestManager_i.cpp: implementation of the TestManager_i class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ManagerMain.h"
#include "OstaTraceDef.h"
#include "TestManager_i.h"
#include "ctime.hpp"
#include "CECInetInfo.h"
#include <assert.h>
#include <process.h>
#include "VarFactory.h"

#include <iostream>
using namespace std;
CTSemaphore blocking_semaphore(0);
CTMutex startTGmutex;

//////////////////////////////////////////////////////////////////////
// Thread Functions
//////////////////////////////////////////////////////////////////////
void* TestManagerImpl_SCHEDULER(void* args)
{
	TestManagerImpl* pTM=reinterpret_cast<TestManagerImpl*>(args);
	assert(pTM);

	try
	{
		while (true)
		{
			CTTime tm;
			tm.now();
	
			for (unsigned int i = 0; i < (unsigned int)pTM->m_execInfo.GetCount(); i++)
			{
				TestManagerImpl::TGExecInfo info = pTM->m_execInfo[i];
				if (info.status == DELAYED && (unsigned)info.tgStartup.lStartTime <= tm.getSeconds())
				{
				    char strMsg[64];
				    sprintf(strMsg, "Starting scheduled TG: Group %d", info.tgStartup.execInfo.lTaskGroupId);
					LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, strMsg);
					// Make sure the IC is activated for use by this test manager instance...
					info.pIC = pTM->ConnectToInjector(info.tgStartup.strHost);
	
					TaskGroupExecInfo EI = info.tgStartup.execInfo;
					LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Calling IC->StartTaskGroup..");
					bool bStarted = info.pIC->StartTaskGroup(EI);
					if (bStarted)
					{
						// Task group has started sucessfully, set to active!
						LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Started sucessfully! Now ACTIVE");
						info.status = ACTIVE;
						pTM->m_execInfo[i] = info;
					}
					else
					{
						info.status = INACTIVE;
						pTM->m_execInfo[i] = info;
					}
	
					pTM->m_nDelayedCount--;
				}
			}
	
			::Sleep(1000);
		}
	}
	catch(...)
	{
		LOG_MESSAGE(TRCCAT_RUNNING, TRCLEV_NONE, "Exception: TestManagerImpl_SCHEDULER");
	}
	return NULL;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
TestManagerImpl::TestManagerImpl(CORBAServer* pServer)
        :m_pServer(pServer), m_lCachedTotalVUs(0)
{
	LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Constructing TestManager Interface");
	m_nPID = _getpid();
	m_pRepos = NULL;
	m_pScheduler = NULL;
	m_nDelayedCount = 0;
	m_bUseStopOnComplete = FALSE;
	m_nStopOnCompleteRemaining = 0;

    char strReg[64];
    CECInetInfo inet;
    sprintf(strReg, "%s_%8.8lX", inet.GetIPAddressAsString().c_str(), getpid());
	m_strReg = strReg;

    m_pServer->ActivateObject(this);
    m_pServer->NameObject(this->_this(),TestManager::CLASS_NAME,m_strReg);

	string strMsg = "TestManager Interface registered with name: ";
	strMsg += m_strReg.GetBuffer(0);
	LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, strMsg);
}

TestManagerImpl::~TestManagerImpl()
{

}

BOOL TestManagerImpl::Initialize(string strTest)
{
	// Create the repository
	try
	{
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Create Repository connection");
		m_strTestName = strTest;
		m_pRepos = new CLocalRepository(m_pServer, strTest, m_nPID);
		m_strTestRunString = m_pRepos->GetTestRunString();

		CECTestLog audit(0,string("TestManager Initialized for Test "+strTest).c_str());

		m_pRepos->m_pTR->AddTestOutput(CECTestLog::OUTPUT_TYPE, audit.ToString().c_str());

		// Start Schedulder thread
		m_pScheduler = new omni_thread(&TestManagerImpl_SCHEDULER ,this);
	    m_pScheduler->start();

		// Start Stop Controller thread
	}
	catch(...)
	{
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_NONE, "Exception: Create Repository connection");
	}
	return true;
}

void TestManagerImpl::CloseDown()
{
	try
	{
		LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, "Clear Variables");
	    CECInetInfo inet;
		CORBA::Object_var Obj = m_pServer->FindObjectWithName(VarFactory::CLASS_NAME, inet.GetIPAddressAsString().c_str());
		VarFactory_var VF = VarFactory::_narrow(Obj);
		VF->ClearAll();

		LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, "Deleteing repository");
		if(m_pRepos)
			delete m_pRepos;
	}
	catch(...)
	{
		LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_NONE, "Exception: Deleteing repository");
	}

	try
	{
		LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, "De-registering TestManager Interface");
	    m_pServer->UnNameObject(TestManager::CLASS_NAME,m_strReg);
	}
	catch(...)
	{
		LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_NONE, "Exception: De-registering TestManager Interface");
	}
}

char* TestManagerImpl::GetTestRunFolderName()
{
	if(m_pRepos)
		return CORBA::string_dup(m_pRepos->GetRunFolderName().c_str());
	return CORBA::string_dup("");
}

CORBA::Boolean TestManagerImpl::StartTaskGroup(TaskGroupStartupInfo& tgInfo)
{
	LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "StartTaskGroup invoked");

	string strLev;
	if(tgInfo.execInfo.lStatsLevel == stalev_none)
	{
		strLev = "NONE";
	}
	else if(tgInfo.execInfo.lStatsLevel == stalev_low)
	{
		strLev = "LOW";
	}
	else if(tgInfo.execInfo.lStatsLevel == stalev_high)
	{
		strLev = "HIGH";
	}
	LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, string("TaskGroup StatLevel = ")+strLev);

	// Look for an existing entry in the map for this task group
	TGExecInfo info;
	BOOL bLookup = m_execInfo.Lookup(tgInfo.execInfo.lTaskGroupId, info);

	// If the entry exists....
	if(bLookup)
	{
		// ... and it's active or delayed -> return
		if (info.status == ACTIVE || info.status == DELAYED || info.status == STOPPING)
			return false;

		// ... if it is STOPPED/Failed/complete then copy in the new execInfo (may have been changed)
		// and prepare to start the test
		info.tgStartup = tgInfo;
		info.status = INACTIVE;
		m_execInfo.SetAt(tgInfo.execInfo.lTaskGroupId, info);
	}
	else
	{
		if (string(tgInfo.strHost) == string("localhost"))
		{
		    CECInetInfo inet;
		    tgInfo.strHost=inet.GetHostName().c_str();
		}
		info.tgStartup = tgInfo;
		info.status = INACTIVE;

		// Connect to the InjectorControl object on the host machine
		info.pIC = NULL;
		info.pIC = ConnectToInjector(info.tgStartup.strHost);

		if (info.pIC == NULL || CORBA::is_nil(info.pIC))
		{
			string strMsg="Failed to connect to Injector Control on host "+string(info.tgStartup.strHost);
			CECErrorLog err(0, m_strTestName.c_str(), "TestManager::StartTaskGroup", strMsg.c_str());
			m_pRepos->m_pTR->AddTestOutput(CECErrorLog::OUTPUT_TYPE, err.ToString().c_str());

			info.status = FAILED;
			m_execInfo.SetAt(tgInfo.execInfo.lTaskGroupId, info);
			return false;
		}
		m_execInfo.SetAt(tgInfo.execInfo.lTaskGroupId, info);
	}


	// If no delay has been set start the task group immediately by calling the
	// Injector Control object on the target machine.
	if (tgInfo.lStartTime == 0)
	{
		try
		{
			LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Starting task group now");

			TaskGroupExecInfo EI = tgInfo.execInfo;
			bool bStarted = info.pIC->StartTaskGroup(EI);
			if (bStarted)
			{
				// Task group has started sucessfully, set to active!
				LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Started sucessfully! Now ACTIVE");
				info.status = ACTIVE;
				m_execInfo.SetAt(tgInfo.execInfo.lTaskGroupId, info);

				if (info.tgStartup.bStopOnComplete)
				{
					m_bUseStopOnComplete = true;
					m_nStopOnCompleteRemaining++;
				}
				return bStarted;
			}
			else
			{
				info.status = FAILED;
				m_execInfo.SetAt(tgInfo.execInfo.lTaskGroupId, info);
			}
			LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Failed to start!");
		}
		catch(...)
		{
			LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_NONE, "Exception: StartTaskGroup - Immediate start");
		}
	}

	// Requires scheduling! Just add the TGExecInfo into the vector and let the
	// Scheduler thread start the task group
	else
	{
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Scheduling TaskGroup. Now DELAYED");
		info.status = DELAYED;
		m_nDelayedCount++;
		m_execInfo.SetAt(tgInfo.execInfo.lTaskGroupId, info);
		if (info.tgStartup.bStopOnComplete)
		{
			m_bUseStopOnComplete = true;
			m_nStopOnCompleteRemaining++;
		}

		return true;
	}

	return false;
}

TestExecuter_ptr TestManagerImpl::GetTaskGroupExecuter(long lTaskGroupId)
{
	try
	{
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Looking for IC responsible for this task group");
		for (unsigned int i = 0; i < (unsigned int)m_execInfo.GetCount(); i++)
		{
			TGExecInfo info = m_execInfo[i];
			if (info.tgStartup.execInfo.lTaskGroupId == lTaskGroupId)
			{
				LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Getting TestExecuter");
				return info.pIC->GetTaskGroupExecuter(lTaskGroupId);
			}
		}
	}
	catch(...)
	{
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_NONE, "Exception: GetTaskGroupExecuter");
	}

	return TestExecuter::_nil();
}

CORBA::Boolean TestManagerImpl::StopTaskGroup(long lTaskGroupId)
{
	try
	{
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "StopTaskGroup: Looking for IC responsible for this task group");
		for (unsigned int i = 0; i < (unsigned int)m_execInfo.GetCount(); i++)
		{
			TGExecInfo info = m_execInfo[i];
			if (info.tgStartup.execInfo.lTaskGroupId == lTaskGroupId)
			{
				if (info.status == ACTIVE)
				{
					LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "calling StopTaskGroup on Injector control");
					if(info.pIC->StopTaskGroup(lTaskGroupId))
					{
						info.status = STOPPING;
						m_execInfo[i] = info;
						return true;
					}
					else
					{
						//it's not stopping
						return false;
					}
				}
				else if (info.status == DELAYED)
				{
					info.status = STOPPED;
					m_execInfo[i] = info;
					return true;
				}
			}
		}
	}
	catch(...)
	{
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Exception: StopTaskGroup");
	}
	return false;
}

CORBA::Boolean TestManagerImpl::StopTest(CORBA::Boolean bForce, long lTimeout)
{
	LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "In StopTest");
	for (unsigned int i = 0; i < (unsigned int)m_execInfo.GetCount(); i++)
	{
		try
		{
			TGExecInfo info = m_execInfo[i];
			info.pIC->StopTest(false,0);
		}
		catch(...)
		{
			LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_NONE, "Exception: StopTest");
		}
	}
	return false;
}

CORBA::Boolean TestManagerImpl::GetStatsSnapshot(TESnapList& TestSnapList)
{
	CTLockGuard<CTMutex> lock(m_InjListMutex);
	try
	{
	    for (int i = 0; i < (int)m_injectorsInUse.size(); i++)
		{
	        InjectorHost inj = m_injectorsInUse[i];
			TESnapList InjectorSnapList;
	        boolean retsts = inj.pIS->GetStatsSnapshot(InjectorSnapList);

			int curLen = TestSnapList.length();
			TestSnapList.length(curLen + InjectorSnapList.length());
		
			for (int snap = 0; snap < (int)InjectorSnapList.length(); snap++)
			{
				TestSnapList[curLen + snap] = InjectorSnapList[snap];
			}
	    }
	}
	catch(...)
	{
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_NONE, "Exception: GetStatsSnapshot");
	}
    return true;
}

long TestManagerImpl::RefreshTestVUs()
{
	long nTotalVUs = 0;
	try
	{
		CTLockGuard<CTMutex> lock(m_InjListMutex);
		for (int i = 0; i < (int)m_injectorsInUse.size(); i++)
		{
			InjectorHost inj = m_injectorsInUse[i];
			nTotalVUs += inj.pIS->GetInjectorVUs(m_lCachedTotalVUs);
		}
        m_lCachedTotalVUs = nTotalVUs;  // Cache the new value.
	}
	catch(...)
	{
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_NONE, "Exception: RefreshTestVUs");
	}

    return m_lCachedTotalVUs;
}

long TestManagerImpl::GetTestVUs()
{
	return m_lCachedTotalVUs;   // Just return the cached value (no locking needed under Windows for a 32-bit value).
}

void TestManagerImpl::OnTaskGroupStopped(long lTaskGroupId, TaskGroupStopReason& eReason)
{
    char strMsg[64];
    sprintf(strMsg, "TaskGroup %d has stopped", lTaskGroupId);
	LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, strMsg);

    for (int i =0; i<m_execInfo.GetCount(); i++)
	{
		if (m_execInfo[i].tgStartup.execInfo.lTaskGroupId == lTaskGroupId)
		{
			if(eReason == tgstop_stopped)
			{
				m_execInfo[i].status = STOPPED;
			}
			else if (eReason == tgstop_completed)
			{
				m_execInfo[i].status = COMPLETED;
			}

			if (m_execInfo[i].tgStartup.bStopOnComplete && m_bUseStopOnComplete)
			{
				m_nStopOnCompleteRemaining--;
				// Make sure that it does not get handled twice!
				m_execInfo[i].tgStartup.bStopOnComplete = FALSE;
			}

			break;
		}
	}

	if (m_bUseStopOnComplete && m_nStopOnCompleteRemaining == 0)
	{
		LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, "All TasksGroup with StopOnComplete have finished. Calling StopTest");
		StopTest(false, 30000);
	}
}

void TestManagerImpl::OnInjectorStopped(const char* strInjectorHost)
{
	string strMsg = "Injector has completed on host machine ";
	strMsg += strInjectorHost;
	LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, strMsg);

	CTLockGuard<CTMutex> lock(m_InjListMutex);
	for (unsigned int i = 0; i<m_injectorsInUse.size(); i++)
	{
		if (m_injectorsInUse[i].strHostName == string(strInjectorHost))
		{
			m_injectorsInUse[i].bComplete = true;
			LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, "Release Injector Status");
			CORBA::release(m_injectorsInUse[i].pIS);
			break;
		}
	}

	BOOL bAllComplete = true;
	for (i = 0; i<m_injectorsInUse.size(); i++)
	{
		if (!m_injectorsInUse[i].bComplete)
		{
			bAllComplete = false;
			LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, string("Injector ")+m_injectorsInUse[i].strHostName+" NOT COMPLETE");
		}
		else
		{
			LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, string("Injector ")+m_injectorsInUse[i].strHostName+" COMPLETE");
		}
	}

	LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, "Checking if all injectors complete");
	if(bAllComplete && m_nDelayedCount == 0)
	{
		long nCurFiles = m_pRepos->GetNumResultsFiles();
		CECInetInfo inet;
		while(m_injectorsInUse.size()>0)
		{
			try
			{
				LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, string("Transfer Results Files from injector ")+m_injectorsInUse[0].strHostName);
			    string strICReg=inet.GetIPAddressAsString(m_injectorsInUse[0].strHostName);
	
				CORBA::Object_ptr icObj = m_pServer->FindObjectWithName(InjectorControl::CLASS_NAME,strICReg.c_str());
				if (!CORBA::is_nil(icObj))
				{
					InjectorControl_var pIC = InjectorControl::_narrow(icObj);
					long nFiles = pIC->TransferResults(m_strTestName.c_str(), inet.GetHostName().c_str(), m_pRepos->GetRunFolderName().c_str());
					nCurFiles+=nFiles;
				}
			}
			catch(...)
			{
				LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_NONE, string("Exception: OnInjectorStopped Transfer files from ")+m_injectorsInUse[0].strHostName);
			}
			m_injectorsInUse.erase(m_injectorsInUse.begin());
		}

		int nRetry = 0;
		while(m_pRepos->GetNumResultsFiles() < nCurFiles && nRetry<10)
		{
			char strMsg[64];
		    sprintf(strMsg, "Waiting fo %d results files", nCurFiles-m_pRepos->GetNumResultsFiles());

			LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, string(strMsg));
			::Sleep(1000);
			nRetry++;
		}
		LOG_MESSAGE(TRCCAT_SHUTDOWN, TRCLEV_MEDIUM, "All complete. Shutting down");
		blocking_semaphore.unlock();
	}

}

InjectorControl_ptr TestManagerImpl::ConnectToInjector(const char* strHost)
{
	InjectorControl_ptr pIC = NULL;
	try
	{
		// Loop through the vector containing the TGExecInfo of task groups already started
		// for a task group on the required machine and return it's InjectorControl object.
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Looking for existing reference to IC on required host.");
		for (unsigned int i = 0; i < (unsigned int)m_execInfo.GetCount(); i++)
		{
			TGExecInfo info = m_execInfo[i];
			if (string(info.tgStartup.strHost) == string(strHost))
			{
				LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Found IC on required host.");
				pIC = info.pIC;
			}
		}
	
		if(pIC == NULL)
		{
			// No task groups running on this machine so narrow the InjectorControl and activate it!
			LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "No IC reference found. Attempting to connect...");
		    CECInetInfo inet;
		    string strICReg=inet.GetIPAddressAsString(strHost);
	
			CORBA::Object_ptr icObj = m_pServer->FindObjectWithName(InjectorControl::CLASS_NAME,strICReg.c_str());
			if (!CORBA::is_nil(icObj))
			{
				LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Found IC. Narrowing...");
				pIC = InjectorControl::_narrow(icObj);
			}
			else
			{
				LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Cannot find an InjectorControl on this host!!");
				return InjectorControl::_nil();
			}
		}
	
		// Make sure the IC is activated for use by this test manager instance...
		if (!pIC->IsActivated(m_strReg))
		{
			// ... if not attempt to activate it
			CECInetInfo inet;
			LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Activating IC..");
			bool bActivated = pIC->Activate(m_strReg, inet.GetHostName().c_str(), m_strTestRunString.c_str(), m_strTestName.c_str());
			if (!bActivated)
			{
				LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Injector control is already in use!!");
			}
			else
			{
			    CECInetInfo inet;
			    string strISReg=inet.GetIPAddressAsString(strHost);
				InjectorHost inj;
				inj.strHostName = strHost;
				inj.bComplete = false;
				CORBA::Object_ptr isObj = m_pServer->FindObjectWithName(InjectorStatus::CLASS_NAME,strISReg.c_str());
				if (!CORBA::is_nil(isObj))
				{
					LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Found IS. Narrowing...");
					inj.pIS = InjectorStatus::_narrow(isObj);
				}
				else
				{
					LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_MEDIUM, "Failed to find Injector Status Object");
					inj.pIS = InjectorStatus::_nil();
				}

				CTLockGuard<CTMutex> lock(m_InjListMutex);
				m_injectorsInUse.push_back(inj);;
			}
		}
	}
	catch(...)
	{
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_NONE, "Cannot find an InjectorControl on this host!!");
		return InjectorControl::_nil();
	}

	return InjectorControl::_duplicate(pIC);
}

char* TestManagerImpl::TestName()
{
	return CORBA::string_dup(m_strTestName.c_str());
}

TGExecStatus TestManagerImpl::GetTaskGroupStatus(long lTaskGroupId)
{
	for (unsigned int i = 0; i < (unsigned int)m_execInfo.GetCount(); i++)
	{
		TGExecInfo info = m_execInfo[i];
		if (info.tgStartup.execInfo.lTaskGroupId == lTaskGroupId)
		{
			return info.status;
		}
	}

	return INACTIVE;
}

CORBA::Boolean TestManagerImpl::GetTaskSnapshot(const TaskIdList& taskIDs, TaskSnap2List& taskSnapshots)
{
	bool bRet = false;
	// Build a list of injectors where the specified task groups are being run
	CStringArray arrInj;
	for (int i=0; i<taskIDs.length(); i++)
	{
		TGExecInfo info;
		BOOL bLookup = m_execInfo.Lookup(taskIDs[i].lTaskGroupId, info);
		if (bLookup && info.tgStartup.execInfo.tgtType == tgt_script)
		{
			CString strInj = info.tgStartup.strHost;
			bool bAdd = true;
			for (int j = 0; j<arrInj.GetSize(); j++)
			{
				if (arrInj.GetAt(j) == strInj)
				{
					bAdd = false;
					break;
				}
			}
			if (bAdd)
			{
				arrInj.Add(strInj);
			}
		}

	}

	// Find the InjectorStatus Object on the machine running this task group
	// and forward the call.
	CTLockGuard<CTMutex> lock(m_InjListMutex);
	try
	{
	    for (i = 0; i < (int)m_injectorsInUse.size(); i++)
		{
	        InjectorHost inj = m_injectorsInUse[i];
			for (int j = 0; j<arrInj.GetSize(); j++)
			{
				if(CString(inj.strHostName.c_str()) == arrInj.GetAt(j))
				{
					LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_HIGH, "Call IS->GetTaskSnapshot");
					TaskSnap2List vuSnaps;
					bRet = inj.pIS->GetTaskSnapshot(taskIDs, vuSnaps);
					if (bRet)
					{
						char strMsg[64];
					    sprintf(strMsg, "Got %d more snapshots", vuSnaps.length());
						LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_HIGH, strMsg);
						int curLen = taskSnapshots.length();
						taskSnapshots.length(curLen + vuSnaps.length());
		
						for (int snap = 0; snap < (int)vuSnaps.length(); snap++)
						{
							taskSnapshots[curLen + snap] = vuSnaps[snap];
						}
					}
				}
			}
	    }
	}
	catch(...)
	{
		LOG_MESSAGE(TRCCAT_STARTUP, TRCLEV_NONE, "Exception: GetTaskSnapshot");
	}

	return bRet;
}

