/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*
 *  tpg_getcmd.c:
 *  =============
 *
 *  This module contains the TPG_GETCMD routine. This retrieves and parses
 *  the TPG command line. 
 */

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 *  Function Prototypes for External Routines.
 */
#include <psl_rtndef.h>         /* PSL Routines */
#include "tpg_rtndef.h"         /* TPG Routines */

/*
 *  Local Function prototypes.
 */
static void INI_TPG_CMD (struct tpg_cmd *tpg_cmd_ptr);

/*
 *  Structure Definitions.
 */
#include <shrmsg.h>         /* PSL messages */
#include <getopt.h>         /* getopt symbols */
#include <psldiag_prmdef.h>     /* Diagnostic definitions */
#include "tpg_comdef.h"         /* Common Data Structure */
#include "tpg_cmd.h"            /* Command Line Data Structure */

/************************************************************************/
/*                                  */
/*  TPG_GETCMD - Get and Parse Command Line             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPG_GETCMD (tpg_com_ptr, argc, argv_ptr);     */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret        = Unsigned Long PSL Routine Return Status    */
/*     tpg_com_ptr = Pointer to Common Data Structure       */
/*     argc        = Argument Count                 */
/*     argv_ptr    = Pointer to Argument Vector         */
/*                                  */
/*  This routine parses the command line.               */
/*                                  */
/************************************************************************/

unsigned long TPG_GETCMD (struct tpg_com *tpg_com_ptr,
                          long argc,
                          char *argv_ptr [])
{
    /*
     *  Local Data.
     */
    unsigned long       iret = PSL_NORMAL;  /* Return Status - assume success */
    struct   tpg_cmd    *tpg_cmd_ptr;   /* Pointer to Command Line Struct */
    int        optchr;     /* Option character or -1 (end of options) */
    int        option_index;       /* `getopt_long' stores the option index here */
    static   char       *tpg_file = NULL;   /* Pointer to TPG filename */
    char       msgbuf [256];   /* Message buffer */

    static const struct option long_options[] =
        {
            /* name,    has_arg,           flag, value */
            {"help", no_argument, 0, 'h'},
            {"confirm", no_argument, 0, 'i'},
            {"list", required_argument, 0, 'l'},
            {"object", required_argument, 0, 'o'},
            {"log", no_argument, 0, 'v'},
            {0, 0, 0, 0}
        };

    const char *const usage = "Usage: tpg [OPTION] [TPG_FILE]\n"
                              "-h              --help\n"
                              "-i              --confirm\n"
                              "-l [on|off]     --list=[on|off]\n"
                              "-o obj_file     --object=obj_file\n"
                              "-v              --log\n";

    /*
     *  Copy pointers for ease of reference.
     */
    tpg_cmd_ptr = (*tpg_com_ptr).cmd_ptrps;

    /* Initialize Command Line Data */
    INI_TPG_CMD (tpg_cmd_ptr);

    /*
     *  If there are no arguments, just output the usage summary.
     */
    if (argc == 1)          /* If no arguments */
    {
        iret = PSL_BADPARAM;        /* Force an error */
        goto output_usage;      /* And output usage summary */
    }

    /*
     *  Parse the command line: Suppress error reporting in getopt_long so that
     *  we can report errors in our own style, when "?" is returned.
     */
    opterr = 0;             /* Suppress getopt_long error reporting */

    while (1)   /* Loop indefinitely - rely on 'break' and 'goto' to exit loop */
    {
        option_index = 0;       /* Reset option index for each option */

        /* Parse the next command line option */
        optchr = getopt_long (argc, argv_ptr, "hil:o:v", long_options, &option_index);

        /* Detect the end of the options */
        if (optchr == -1) break;

        /* Process this command line option */
        switch (optchr)
        {
        case 0:
            fprintf (stderr, "getopt_long returns 0 - should never happen\n");
            break;

        case 'h':           /* --help */
            /*puts("option -h\n"); */
            goto output_usage;

        case 'i':           /* --confirm */
            /*puts("option -i\n"); */
            (*tpg_cmd_ptr).confirmn = PTRUE;
            (*tpg_cmd_ptr).confirm_negatedn = PFALSE;
            (*tpg_cmd_ptr).confirm_valuen = PFALSE;
            break;

        case 'l':           /* --list=[on|off] */
            /*printf("option -l with value `%s'\n", optarg); */
            if (strcmp (optarg, "on") == 0) /* --list=on */
            {
                (*tpg_cmd_ptr).listn = PTRUE;
            }
            else if (strcmp (optarg, "off") == 0) /* --list=off */
            {
                (*tpg_cmd_ptr).listn = PFALSE;
            }
            else                /* Unrecognised --list keyword */
            {
                iret = PSL_IVKEYW;      /* Set up error: Unrecognised keyword */
                PSL_PUTDIAG (           /* Put Diagnostic */
                    (*tpg_com_ptr).diagstkps,   /* Use this stack */
                    1,              /* Number of message segments */
                    iret,           /* Message */
                    optarg);            /* Unrecognised keyword */
                PSL_OUTPUT_DIAG (       /* Output the diagnostic */
                    (*tpg_com_ptr).diagstkps,   /* From this stack */
                    0,              /* No Context */
                    DIOUFL_LSTERR);     /* Last Error */
                goto output_usage;      /* And output usage summary */
            }
            break;

        case 'o':           /* --object=obj_file */
            /*printf ("option -o with value `%s'\n", optarg); */
            (*tpg_cmd_ptr).objectn = PTRUE;
            (*tpg_cmd_ptr).object_negatedn = PFALSE;
            (*tpg_cmd_ptr).object_valuen = PTRUE;
            (*tpg_cmd_ptr).object_value_ptrpc = optarg;
            break;

        case 'v':           /* --log */
            /*puts("option -v\n"); */
            (*tpg_cmd_ptr).logn = PTRUE;
            (*tpg_cmd_ptr).log_negatedn = PFALSE;
            (*tpg_cmd_ptr).log_valuen = PFALSE;
            break;

        case '?':               /* Unrecognised option switch */
        default:                /* ... */
            msgbuf[0] = optopt;         /* Unrecognised option switch */
            msgbuf[1] = '\0';           /* ... */

            iret = PSL_IVSWIT;          /* Set up error: Unrecognised switch */
            PSL_PUTDIAG (           /* Put Diagnostic */
                (*tpg_com_ptr).diagstkps,   /* Use this stack */
                1,              /* Number of message segments */
                iret,           /* Message */
                msgbuf);            /* Unrecognised switch */
            PSL_OUTPUT_DIAG (           /* Output the diagnostic */
                (*tpg_com_ptr).diagstkps,   /* From this stack */
                0,              /* No Context */
                DIOUFL_LSTERR);     /* Last Error */
            goto output_usage;          /* And output usage summary */
        }
    }

    if (optind == argc - 1)     /* If correct number of parameters */
    {
        /*printf("filename %s \n", argv_ptr[optind]); */
        (*tpg_cmd_ptr).tpgn = PTRUE;
        (*tpg_cmd_ptr).tpg_negatedn = PFALSE;
        (*tpg_cmd_ptr).tpg_valuen = PTRUE;
        (*tpg_cmd_ptr).tpg_value_ptrppc = &tpg_file;
        tpg_file = argv_ptr[optind++];
        (*tpg_cmd_ptr).tpg_value_countl += 1;
    }
    else if (optind >= argc)        /* Else if too few parameters */
    {
        iret = PSL_INSFPRM;     /* Set up error: Insufficient parameters */
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tpg_com_ptr).diagstkps,   /* Use this stack */
            1,             /* Number of message segments */
            iret);         /* Message */
        PSL_OUTPUT_DIAG (       /* Output the diagnostic */
            (*tpg_com_ptr).diagstkps,   /* From this stack */
            0,              /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        goto end;           /* and exit! */
    }
    else                /* Else too many parameters */
    {
        iret = PSL_MAXPARM;     /* Set up error: Too many parameters */
        optind++;           /* Move to next option */
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tpg_com_ptr).diagstkps,   /* Use this stack */
            1,             /* Number of message segments */
            iret,          /* Message */
            argv_ptr[optind]);     /* First 'extra' Parameter */
        PSL_OUTPUT_DIAG (       /* Output the diagnostic */
            (*tpg_com_ptr).diagstkps,   /* From this stack */
            0,              /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        goto end;           /* and exit! */
    }

    /*
     *  Set up the listing environment, if required.
     */
    if ((*tpg_cmd_ptr).listn)
    {
        (void) LIST_ALLOCATE ((*tpg_com_ptr).diagstkps,
                              &(*tpg_com_ptr).list_ctx_ptrps);
    }

    /*
     *  OBJECT.
     *
     *  Unless explicitly negated, OBJECT is the default.
     */
    if (!(*tpg_cmd_ptr).object_negatedn)
    {
        (*tpg_cmd_ptr).objectn = PTRUE;
    }

    goto end;               /* And exit */

    /*
     *  OUTPUT_USAGE:
     *
     *  Output the usage summary.
     */
output_usage:
    puts (usage);

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return iret;
}

/************************************************************************/
/*                                  */
/*  INI_TPG_CMD - Initialize Command Line Data          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     INI_TPG_GETCMD (tpg_cmd_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tpg_cmd_ptr = Pointer to Command Line Data Structure     */
/*                                  */
/*  This routine initializes the Command Line Data Structure.   */
/*                                  */
/************************************************************************/
static void INI_TPG_CMD (struct tpg_cmd *tpg_cmd_ptr)
{
    (*tpg_cmd_ptr).confirmn = PFALSE;
    (*tpg_cmd_ptr).confirm_negatedn = PFALSE;
    (*tpg_cmd_ptr).confirm_valuen = PFALSE;

    (*tpg_cmd_ptr).listn = PFALSE;
    (*tpg_cmd_ptr).list_negatedn = PFALSE;
    (*tpg_cmd_ptr).list_valuen = PFALSE;
    (*tpg_cmd_ptr).list_value_ptrpc = NULL;

    (*tpg_cmd_ptr).objectn = PFALSE;
    (*tpg_cmd_ptr).object_negatedn = PFALSE;
    (*tpg_cmd_ptr).object_valuen = PFALSE;
    (*tpg_cmd_ptr).object_value_ptrpc = NULL;

    (*tpg_cmd_ptr).logn = PFALSE;
    (*tpg_cmd_ptr).log_negatedn = PFALSE;
    (*tpg_cmd_ptr).log_valuen = PFALSE;

    (*tpg_cmd_ptr).tpgn = PFALSE;
    (*tpg_cmd_ptr).tpg_negatedn = PFALSE;
    (*tpg_cmd_ptr).tpg_valuen = PFALSE;
    (*tpg_cmd_ptr).tpg_value_ptrppc = NULL;
    (*tpg_cmd_ptr).tpg_value_countl = 0;

    return;
}
