/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/////////////////////////////////////////////////////////////////////////
//
// CECRegistryTools.h
// ------------------
//
//    This is the implementation file for the CECRegistryTools class
//
//=========================================================================

#include "stdafx.h"
#include "CECRegistryTools.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CECRegistryTools

// Constructor
CECRegistryTools::CECRegistryTools(HKEY hKey /*=NULL*/)
{
    Attach(hKey);
}

// Destructor
CECRegistryTools::~CECRegistryTools()
{
    Detach();
}


/////////////////////////////////////////////////////////////////////////////
// CECRegistryTools message handlers

//
// DeleteKey
// =========
//
// Removes the specified key from the registry and explicitly removes any subkeys.
// If the key has subkeys, you must call this method under Windows NT in order to
// delete the key. Under Windows 95, you can call DeleteSubKey to remove the key
// and any subkeys.
//
// LPCTSTR | lpszKey | specifies the name of the key to delete.
// This name must be a subkey of m_hKey.
//
// ERROR_SUCCESS if successful, otherwise an error message
LONG CECRegistryTools::DeleteKey(LPCTSTR lpszKey)
{
    CECRegistryTools key;

    LONG lResult = key.Open(m_hKey, lpszKey,KEY_READ|KEY_WRITE);

    if (lResult != ERROR_SUCCESS)
        return lResult;

    FILETIME time;
    TCHAR szBuffer[256];
    DWORD dwSize = 256;

    // And loop through the subkeys ...
    while (RegEnumKeyEx(key.m_hKey,
                        0,
                        szBuffer,
                        &dwSize,
                        NULL,
                        NULL,
                        NULL,
                        &time)==ERROR_SUCCESS)
    {
        // Delete the subkey ...
        lResult = key.DeleteKey(szBuffer);

        if (lResult != ERROR_SUCCESS)
            return lResult;

        dwSize = 256;
    }

    key.Close();

    // And delete this key
    return DeleteSubKey(lpszKey);
}

//
// Open
// ====
//
// Opens the specified key and sets m_hKey to the handle of this key. If the
// lpszKeyName parameter is NULL or points to an empty string, Open opens a
// new handle of the key identified by hKeyParent, but does not close any
// previously opened handle.
//
// HKEY | hKeyParent | The handle of an open key
// LPCTSTR | lpszKeyName | Specifies the name of a key to be created or opened
// REGSAM | samDesired | The security access for the key
//
// ERROR_SUCCESS if successful, otherwise an error message
//
LONG CECRegistryTools::Open(HKEY hKeyParent, LPCTSTR lpszKeyName, REGSAM samDesired /*=KEY_ALL_ACCESS*/)
{
    _ASSERTE(hKeyParent != NULL);

    HKEY hKey = NULL;
    LONG lResult = ERROR_SUCCESS;

    // Open the key
    lResult = RegOpenKeyEx(hKeyParent, lpszKeyName, 0, samDesired, &hKey);
    if (lResult == ERROR_SUCCESS)
    {
        lResult = Close();

        m_hKey = hKey;
    }
    else
    {
        ReportError();
    }

    return lResult;
}

//
// DeleteSubKey
// ============
//
// removes the specified key from the registry
//
// LPCTSTR | lpszSubKey | Key to delete
//
// ERROR_SUCCESS if successful, otherwise an error message
//
LONG CECRegistryTools::DeleteSubKey(LPCTSTR lpszSubKey)
{

    _ASSERTE(m_hKey != NULL);
    return RegDeleteKey(m_hKey, lpszSubKey);

}

//
// Close
// =====
//
// Releases the m_hKey member handle and sets it to NULL.
//
// ERROR_SUCCESS if successful, otherwise an error message
//
LONG CECRegistryTools::Close()
{
    LONG lResult = ERROR_SUCCESS;

    if (m_hKey != NULL)
    {
        lResult = RegCloseKey(m_hKey);

        if (lResult != ERROR_SUCCESS)
            ReportError();

        m_hKey = NULL;
    }
    return lResult;
}

//
// Attach
// ======
//
// Attaches an HKEY (hKey) to the CECRegistryTools key object (m_hKey)
// by setting the m_hKey member handle to hKey.
//
// HKEY | hKey | HKEY object
//
void CECRegistryTools::Attach(HKEY hKey)
{
    m_hKey = hKey;
}

//
// Detach
// ======
//
// Detaches the m_hKey member handle from the CRegKey object and
// sets m_hKey to NULL.
//
// HKEY associated with the CECRegistryTools object.
//
HKEY CECRegistryTools::Detach()
{
    HKEY hKey = m_hKey;
    m_hKey = NULL;

    return hKey;
}

//
// ReportError
// ===========
//
// Report any error we have in accessing the registry keys
//
void CECRegistryTools::ReportError()
{
    char    chErrorBuffer[512];
    DWORD   dwErrorCode;

    // Get the last error
    dwErrorCode = GetLastError();

    if (dwErrorCode)
    {
        // Translate it into a string
        FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM,
                       0,
                       dwErrorCode,
                       0,
                       chErrorBuffer,
                       80,
                       NULL);

        // And output it
        AfxMessageBox (chErrorBuffer);
    }
}

//
// GetStringValue
// ==============
//
// Get the string value of ANY key value in the Registry
//
CString CECRegistryTools::GetStringValue(HKEY hKeyParent, LPCTSTR lpszKeyName, LPCTSTR lpszField)
{
    HKEY    hRegKey;
    char    szKeyValueData[256];
    DWORD   dwKeyLen = sizeof(szKeyValueData);

    //
    // Assume that the context has not been set so that we have to query the registry to obtain the status
    //
    CString szValue="";

    //
    // First lets open the key
    //
    if (RegOpenKeyEx (hKeyParent, lpszKeyName, 0, KEY_READ, &hRegKey) == ERROR_SUCCESS)
    {

        //
        // And query the Requested Field
        //
        if (RegQueryValueEx(hRegKey, lpszField, 0, NULL, (unsigned char *) &szKeyValueData, &dwKeyLen) == ERROR_SUCCESS)
        {
            szValue = szKeyValueData;
        }

        // then close the key again
        RegCloseKey(hRegKey);
    }
    return szValue;
}

//
// SetStringValue
// ==============
//
// Set the string value of ANY key value in the Registry
//
int CECRegistryTools::SetStringValue(HKEY hKeyParent, LPCTSTR lpszKeyName, LPCTSTR lpszField ,char* Value ,HKEY BaseKeyToAddTo)
{
    HKEY    hRegKey;

    //
    // Assume that the context has not been set so that we have to query the registry to obtain the status
    //
    CString szValue="";
    //
    // First lets open the key creating it if it does not already exist
    //
    if (RegCreateKeyEx (BaseKeyToAddTo
                        ,lpszKeyName
                        ,0
                        ,NULL
                        ,REG_OPTION_NON_VOLATILE
                        ,KEY_WRITE
                        ,NULL
                        ,&hRegKey
                        ,NULL) == ERROR_SUCCESS)
    {

        //
        // And query the Requested Field
        //
        CString KeyValue = *Value;
        int nRetVal = RegSetValueEx(hRegKey ,lpszField, 0, REG_SZ ,(const unsigned char*) Value, strlen(Value) + 1);
        RegCloseKey(hRegKey);
        return nRetVal;
    }

    return 0;
}
