/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_ASYNCHTTPREQUEST_H_
#define _OPENSTA_ASYNCHTTPREQUEST_H_

////////////////////////////////////////////////////////////////////////////////
// Asynchronous HTTP request definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include <windef.h>
#include <list>
#include <vector>

#include "AsyncRequest.hpp"
#include "AsyncSecureSocket.hpp"
#include "MemoryBufferHandler.hpp"
#include "Trace.hpp"

#include "Statistics.hpp"

#include "SharedMemBuffs.hpp"

typedef enum { NO_RESULT=0, HEADER_ONLY=1, ALL_DATA=2 } ERequiredResults;   /// TODO: used as well in TEHttpRecords.hpp.


// This class implements an HTTP request (GET, HEAD, POST, CONNECT).
// THREAD UNSAFE, except for CancelRequest(), which must be callable
// from a concurrent thread. However, CancelRequest() is not required to
// be synchronous. This allows "soft cancels".
class CAsyncHTTPRequest : public CAsyncRequest, public IocpEventDispatcher
{
public:
    typedef enum {  HTTP_SECURE, HTTP_NOT_SECURE }          ESecurityMode;

	typedef enum {  CLIENT_INITIAL=0, 
                    CLIENT_CONNECTING=1, CLIENT_CONNECTED=2, CLIENT_DISCONNECTING=3,
                    CLIENT_REQUEST_SENT=4, CLIENT_RECEIVING=5 }  EClientStateType;

    typedef enum {  CNX_UNKNOWN, CNX_DISCONNECTS, 
                    CNX_KEEP_ALIVE }                        EConnectionType;

	// Message-parsing values.
	typedef enum {  MSG_UNKNOWN_TYPE, MSG_NO_BODY, MSG_CONTINUE, 
                    MSG_NON_IDENTITY, MSG_CONTENT_LENGTH }      EMessageLengthType;

    class CSockBuffList;

public:
    //// Construction / Destruction.
    CAsyncHTTPRequest(CIOCP& iocp, IAsyncRequestEvent& requestEventListener, DWORD dwRequestNumber = 0, CTrace* pTrace = NULL);
    virtual ~CAsyncHTTPRequest();

    //// User methods.

    static BOOL InitClass(DWORD dwMaxConcurrentSSLReq, DWORD dwSSLGrowingBuffCount);
    static void UninitClass();

	// Initialize. Create the device object.
	virtual void Init()
	{
        m_pDevice = new CAsyncSecureSocket(m_iocp, NULL, 0, m_pTrace);
	}

    void SetWatchStatsTo(BOOL b)    { m_bWatchStats = b; }

    // Connect to the server.
    DWORD Connect(LPCSTR zServer, WORD wPort, IocpEvent* pEventTarget);

    // Re-connect to the server.
    DWORD ReConnect();

    // Disconnect from the server.
    DWORD Disconnect();

    // Are we connected to the server?
    BOOL IsConnected()  const;
    BOOL IsTCPConnected()   const   { return static_cast<CAsyncTCPSocket*>(m_pDevice)->CAsyncTCPSocket::IsOpen(); }

    DWORD   GetConnectionId()                       const   { return m_dwConnectionId; }
    void    SetConnectionId(DWORD dwConnectionId)           { m_dwConnectionId = dwConnectionId; }

    BOOL    GetUseProxy()           const   { return m_bUseProxy; }
    void    SetUseProxyTo(BOOL b)           { m_bUseProxy = b; m_cnxType = CNX_KEEP_ALIVE; }

    void    SetWillReconnectThroughProxy(BOOL b)            { m_bWillReconnectThroughProxy = b; }
    BOOL    GetWillReconnectThroughProxy()          const   { return m_bWillReconnectThroughProxy; }

    void            SetSecurityMode(ESecurityMode sec)          { m_securityMode = sec; }
    ESecurityMode   GetSecurityMode()                   const   { return m_securityMode;}

    void    SetCertificateName(const std::string &szCertificateName) { m_szCertificateName = szCertificateName; }

    void            SetConnectionType(EConnectionType cnx)          { m_cnxType = cnx;  }
    EConnectionType GetConnectionType()                     const   { return m_cnxType; }

    CDataBuffHandler &  GetRequestResults() { return m_messageBuffer; }

    BOOL    NeedConnectToProxy()    const   { return m_bUseProxy && !m_bIsTunnelingThroughProxy; }
    BOOL    FoundContinueHeader()   const   { return m_bContainContinueHeader; }

    EMessageLengthType  GetMessageLengthType()      const   { return m_messageLengthType; }
    BOOL                DoesContainContinueHeader() const   { return m_bContainContinueHeader; }

    EClientStateType    GetStatus() const   { return (EClientStateType) m_dwState; }

    void GetConnParams(std::string &zCurrentServer, DWORD &dwCurrentPort)   const   { zCurrentServer = m_zServer; dwCurrentPort = m_dwPort; }

    void SetRequestText(std::string &szRequest)  { m_szCurrentRequest = szRequest; }

    DWORD GetHTTPBytes() const {return m_dwHTTPBytes;}
    const std::string & GetTimerName()          const { return m_szRequestTimer; }
    clock_t             GetSendRequestTime()    const { return m_ctSendRequestTime; }
    void                SetTimerName(const std::string &szTimerName) { m_szRequestTimer = szTimerName; }


    // Initiates a GET request.
    DWORD InitiateGET   (ERequiredResults eResultsRequired = NO_RESULT);

    // Initiates a POST request.
    DWORD InitiatePOST  (ERequiredResults eResultsRequired = NO_RESULT);

    // Initiates a HEAD request.
    DWORD InitiateHEAD  (ERequiredResults eResultsRequired = NO_RESULT);

    // Initiates a HEAD request.
    DWORD InitiateCONNECT(ERequiredResults eResultsRequired = NO_RESULT);

    // PUT, LINK, DELETE and UNLINK currently not supported.

    //// IocpEvent methods.
    virtual void OnEvent      (DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType);
    virtual void OnEventFailed(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType, DWORD dwError);

    //// --------------------------------------------------------------------------------
    //// IocpEventDispatcher methods.
    //// --------------------------------------------------------------------------------
    IocpEvent* GetIocpEventTarget() {return this;}

    virtual void OnRequestClosed();

    virtual void OnRequestFinished();
    virtual void OnRequestCancelled();

    CStatistics &GetStats()  { return m_statistics; }

protected:

	typedef enum {  FT_UNUSED, FT_MESSAGE, 
                    FT_CONNECTION, FT_PROXY_CONNECTION, 
                    FT_CONTENT_LENGTH, FT_CONTENT_TYPE, 
                    FT_TRANSFER_ENCODING, FT_END_OF_HEADER }    EHeaderFieldType;

	//// Helper methods

    DWORD InitiateGenericRequest(ERequiredResults eResultsRequired, BOOL bSignal);

	// Will handle the buffer things: depending on the implemented solution, either copy the 
	// buffer to final buffer, or move the buffer pointer to an empty area.
	void ProcessReadData();

	// Will parse the results to check if all the buffers were received.
	// Decision will involve number of bytes received, data chunks processing, "Content-length" parsing,
	// and data read errors.
	BOOL IsDataTransferComplete();

	void InitParseStructures();

    void StartTimer();
    void EndTimer();

	//// Buffer-parsing methods. Return FALSE and moves cursor to next line if the line is not completely received.

	// Parse Header lines.
	// Returns FALSE if line not complete and stores remaining data for the next buffer parsing.
    BOOL ParseHeaderLine(EHeaderFieldType &headerFieldType, CDataBuffListAccessor &iterParam, CDataBuffListAccessor &iterEndOfParam);
	BOOL ParseLine		(CDataBuffListAccessor &iterBegOfLine, CDataBuffListAccessor &iterEndOfLine);
	BOOL ParseLine		(CDataBuffListAccessor &iterBegOfLine, CDataBuffListAccessor &iterEndOfLine, BOOL& bEndsWithEmptyLine);

	BOOL BufferParsingOver();

	BOOL IsLastChunk();
	BOOL CheckContentLength();

	//// Chunk-specific.

	BOOL StoreChunk();	// Return TRUE if all the chunk has been stored ...
	BOOL SkipChunk();	// ... or FALSE if the chunk is continued in the next buffer.
		DWORD	m_dwCurrentChunkSizeLeft;	// Chunk buffer left to download.
		enum { CK_READ_SIZE, CK_READ_DATA, CK_READ_END_OF_DATA, CK_READ_TRAILER } m_chunkStep;

	//// Message-length-specific.
	BOOL StoreMessage();
	BOOL SkipMessage();

	BOOL SkipHeader();
	BOOL StoreHeader();

    DWORD TCPDisconnect();
    DWORD SSLDisconnect();

protected:
    //// Data.
	
    // Received data storage buffers.
	DWORD			    m_dwBytesRead;					// Bytes read by the last WSARead. Less than READ_BUFF_SIZE-m_dwBytesRemaining.
	CDataBuffHandler	m_messageBuffer;                // Storage buffer receiving all the data for further use by the HTML parser.

    // Received-data parsing data.
    CDataBuffListAccessor   m_messageBuffIter;          // Object used to iterate through the whole buffers received.
    BOOL                    m_bMessageBufferChanged;    // Flag set to know if the DOM structure for the current m_messageBuffer is
                                                        // up to date.

	enum { RQ_GET, RQ_POST, RQ_HEAD, RQ_CONNECT } m_requestType;    // Request type.

    DWORD	            m_dwContentLengthLeft;		    // Used when the Content-length header field is specified. Bytes left to download.
	EMessageLengthType	m_messageLengthType;
	BOOL				m_bIsHeaderFullyParsed;		    // FALSE if header was not completely received.
	BOOL				m_bIsChunkFullyReceived;
	ERequiredResults    m_eResultsRequired;			    // CHTTPTofInterpreter (does not) need(s) the results.
    BOOL                m_bContainContinueHeader;       // Tells if need to skip the continue header when parsing the results.
    BOOL                m_bNoMeaningFullDataLeft;       // Set when we reached the end of buffers but miss some bytes to match a meaningfull string.

    // Request information data.
    DWORD               m_dwConnectionId;               // TOF connection ID associated with the current structure.
    BOOL                m_bUseProxy;
    BOOL                m_bWillReconnectThroughProxy;   // Used when need to reconnect through the proxy (connexion-close case).
    BOOL                m_bIsTunnelingThroughProxy;     // Keep trace of the HTTP-CONNECT request results.

    ESecurityMode       m_securityMode;                 // HTTPS ...
    EConnectionType     m_cnxType;
    std::string         m_szCertificateName;

    std::string         m_szCurrentRequest;
    std::string			m_szRequestTimer;
    clock_t				m_ctSendRequestTime;    // To compute the request duration.


    // Indicators and statistics.
    DWORD               m_dwHTTPBytes;  // For statistics snapshots: sent a received HTTP bytes.
    CStatistics         m_statistics;
    BOOL                m_bWatchStats;

    // Connection info. (for use by the ReConnect method).
    std::string m_zServer;
    DWORD       m_dwPort;
    IocpEvent*  m_pEventTarget;

#ifdef USE_PERF_DATA
    //// Pointers to performance counters.
    static LONG *   s_plPendingQueriesPerfCtr;
    static LONG *   s_plQueryRatePerfCtr;
    static LONG *   s_plOutThroughPutPerfCtr;
    static LONG *   s_plInThroughPutPerfCtr;
#endif

    static in_addr  s_localAddress;
    static DWORD    s_dwMinSSLBuffSize;

    static CSharedMemBuffs *s_pSSLBuffs;
private:
    //// Forbidden.
    CAsyncHTTPRequest();
    void operator=(const CAsyncHTTPRequest&);
};

#endif  // _OPENSTA_ASYNCHTTPREQUEST_H_
