/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

////////////////////////////////////////////////////////////////////////////////
// Asynchronous secure socket definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Common.hpp"
#include "AsyncSecureSocket.hpp"
#include <iostream>

// The old value 0x00090317L (see Microsoft's WebClient.c SSL sample) is still used by some Windows NT4.0 clients!
// The new value, SEC_E_CONTEXT_EXPIRED, has a value of 0x80090317L. Sorry for the difference!
#define SEC_E_CONTEXT_EXPIRED_OLD      _HRESULT_TYPEDEF_(0x00090317L)

// Network data representation.
#define DATA_REP SECURITY_NETWORK_DREP

// Statics
CFastLock* CAsyncSecureSocket::m_pSChannelBugLock           = NULL;
HMODULE CAsyncSecureSocket::m_hSecurDll                     = NULL;
SecurityFunctionTable CAsyncSecureSocket::m_securityFuncTbl = {0};
BOOL CAsyncSecureSocket::m_bRunAsSystem                     = FALSE;
CredHandle CAsyncSecureSocket::m_hAnonymCreds               = {0};
CFastLock* CAsyncSecureSocket::m_pLock                      = NULL;
CAsyncSecureSocket::CertificateToHandle CAsyncSecureSocket::m_credentialsCache = CAsyncSecureSocket::CertificateToHandle();

DWORD CAsyncSecureSocket::Init()
{
    OSVERSIONINFOEX vi = {0};
    vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
    if(!GetVersionEx((OSVERSIONINFO*)&vi))
    {
        // Failed: try with OSVERSIONINFO.
        vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
        GetVersionEx((OSVERSIONINFO*)&vi);
    }
    else if(vi.dwPlatformId == VER_PLATFORM_WIN32_NT && vi.dwMajorVersion >= 5)
    {
        // This is Windows 2000 or above.

        if(vi.dwMajorVersion < 5 || vi.dwMajorVersion == 5 && vi.wServicePackMajor < 2)
        {
            // This is Windows version prior to 2000 SP2 or above.
            // If we are on a multiprocessor machine, then create a special
            // lock to workaround the schannel.dll bug.
            if(!CFastLock::IsSingleProcessorMachine())
                m_pSChannelBugLock = new CFastLock(4000, TRUE);
        }
    }

    // Check the Windows version: at least NT 4.0 (SP4).
    if(vi.dwPlatformId != VER_PLATFORM_WIN32_NT || vi.dwMajorVersion < 4)
        return ERROR_OLD_WIN_VERSION;

    if(m_securityFuncTbl.dwVersion)
        return NO_ERROR;    // Security library already initialized.

    // Get the correct security library name.
    LPCTSTR zSecurLib = NULL;
    if(vi.dwMajorVersion == 4)
        zSecurLib = TEXT("security.dll");
    else if(vi.dwMajorVersion >= 5)
        zSecurLib = TEXT("secur32.dll");

    // Load the security library.
    m_hSecurDll = LoadLibrary(zSecurLib);
    if(m_hSecurDll == NULL)
        return GetLastError();

    // Initialize the security function table.
    INIT_SECURITY_INTERFACE pInitSecurityInterface =
        (INIT_SECURITY_INTERFACE)GetProcAddress(m_hSecurDll, SECURITY_ENTRYPOINT);
    if(pInitSecurityInterface == NULL)
        return GetLastError();
    m_securityFuncTbl = *(pInitSecurityInterface());

    // Check whether the code is begin executed under the LocalSystem account.
    TCHAR szTitle[1024];
    ULONG lSize = elemCount(szTitle);
    GetUserName(szTitle, &lSize);
    m_bRunAsSystem = (lstrcmpi(TEXT("system"), szTitle) == 0);

    // Create the lock for the credentials cache.
    m_pLock = new CFastLock(4000, TRUE);

    // Create the anonymous credentials handle.
    SCHANNEL_CRED sslAnonymousCredentials = {0};
    sslAnonymousCredentials.dwVersion = SCHANNEL_CRED_VERSION;
    sslAnonymousCredentials.dwFlags   = SCH_CRED_NO_DEFAULT_CREDS;
    TimeStamp tsExpires;
    SECURITY_STATUS ss = m_securityFuncTbl.AcquireCredentialsHandle(
        NULL,                   // Name of principal    
        UNISP_NAME,             // Name of package
        SECPKG_CRED_OUTBOUND,   // Flags indicating use
        NULL,                   // Pointer to logon ID
        &sslAnonymousCredentials,// Package specific data
        NULL,                   // Pointer to GetKey() func
        NULL,                   // Value to pass to GetKey()
        &m_hAnonymCreds,        // (out) Cred Handle
        &tsExpires);            // (out) Lifetime (optional)

    if(ss != SEC_E_OK)
        return (DWORD)ss;

    return NO_ERROR;
}

void CAsyncSecureSocket::Reset()
{
    m_dwState           = STATE_NONE;
    m_ss                = SEC_E_OK;
    m_pEventTarget      = NULL;
    m_bDoRead           = FALSE;
    m_bConnected        = FALSE;
    m_pClientBuffer     = NULL;
    m_dwBufferCount     = 0;
    m_bAutoAlloc        = FALSE;
    m_bRenegotiate      = FALSE;
    m_pDataBuffer       = NULL;
    m_pExtraBuffer      = NULL;
    m_dwNextChunk       = 0;
    m_zCertifName[0]    = 0;
    m_dwTotalBufLen     = 0;

    ZeroMemory(&m_hContext,     sizeof(m_hContext));
    ZeroMemory(&m_hCredentials, sizeof(m_hCredentials));
    ZeroMemory(m_buf1,          sizeof(m_buf1));
    ZeroMemory(m_buf2,          sizeof(m_buf2));
    ZeroMemory(&m_oneBuf,       sizeof(m_oneBuf));
}

CAsyncSecureSocket::CAsyncSecureSocket(CIOCP& iocp, PBYTE pbReadBuffer, DWORD dwSize, CTrace* pTrace)
    : CAsyncTCPSocket(iocp, pTrace),
      m_pReadBuffer(pbReadBuffer), m_dwReadBufLen(dwSize)
{
    assert(m_hSecurDll != NULL && m_pLock != NULL);    // Init() must have been called.
    Reset();
}

CAsyncSecureSocket::~CAsyncSecureSocket()
{
    if(m_bConnected)
        CompleteClose(FALSE);    // Brutal disconnect.
    else
        FreeReadBuffer();
}

DWORD CAsyncSecureSocket::Open(LPCSTR zServer, DWORD dwPort, IocpEvent *pEventTarget)
{
    DWORD dwRes = 0;
    SECURITY_STATUS ss = 0;

    // Allocate the read buffer.
    dwRes = AllocReadBuffer();
    if(dwRes)
        return dwRes;

    if(m_zCertifName[0] == 0)
    {
        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
        {
            EnableTraceStream(m_pTrace);
                TraceHeader(*pOstream) << "Using client credentials for an anonymous user.\r\n";
            ReleaseTraceStream();
        }

        // Get client credentials for an anonymous user.
        m_hCredentials = m_hAnonymCreds;
    }
    else
    {
        if(!MakeClientCertificateCredentials())
            return ERROR_ACCESS_DENIED;
    }

    // Connect to server.
    if(!CAsyncTCPSocket::IsOpen())
    {
        dwRes = CAsyncTCPSocket::Open(zServer, dwPort, this);
        if(dwRes)
            return dwRes;
    }
    else
        SetEventTarget(this);   // Set the current instance as the event target for IOCP events.


    //// Initialize the client handshake.

    SecBufferDesc   OutBuffer;
    DWORD           dwSSPIFlags;
    DWORD           dwSSPIOutFlags;
    TimeStamp       tsExpiry;

    dwSSPIFlags = ISC_REQ_SEQUENCE_DETECT   |
                  ISC_REQ_REPLAY_DETECT     |
                  ISC_REQ_CONFIDENTIALITY   |
                  ISC_RET_EXTENDED_ERROR    |
                  ISC_REQ_ALLOCATE_MEMORY   |
                  ISC_REQ_STREAM
                  ;

    m_buf2[0].BufferType = SECBUFFER_TOKEN;
    m_buf2[0].pvBuffer   = NULL;
    m_buf2[0].cbBuffer   = 0;

    OutBuffer.cBuffers = 1;
    OutBuffer.pBuffers = m_buf2;
    OutBuffer.ulVersion = SECBUFFER_VERSION;

    // Is zServer is an IP address, get the server principle name (suppose it's the server name).
    char *pServerName, zBuf[256];
    if(IsIPAddr(zServer))
    {
        if(GetRemoteHostName(zBuf, sizeof(zBuf)))
            pServerName = zBuf;
        else
            return (DWORD)SEC_E_INTERNAL_ERROR;
    }
    else
        pServerName = const_cast<char*>(zServer);

    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
    {
        EnableTraceStream(m_pTrace);
            TraceHeader(*pOstream) << "Emitting initial blob with InitializeSecurityContext(). Target principle name: " << pServerName << ".\r\n";
        ReleaseTraceStream();
    }

    // Initiate a ClientHello message and generate a token.
    ss = m_securityFuncTbl.InitializeSecurityContext(
                    &m_hCredentials,
                    NULL,
                    pServerName,
                    dwSSPIFlags,
                    0,
                    DATA_REP,
                    NULL,
                    0,
                    &m_hContext,
                    &OutBuffer,
                    &dwSSPIOutFlags,
                    &tsExpiry);

    if(ss != SEC_I_CONTINUE_NEEDED)
        return ss;

    // Cache the caller's event target, so we can later route asynchronous
    // events to the caller.
    m_pEventTarget = pEventTarget;

    // Send response to server if there is one.
    dwRes = 0;
    if(m_buf2[0].cbBuffer != 0 && m_buf2[0].pvBuffer != NULL)
    {
        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
        {
            EnableTraceStream(m_pTrace);
                TraceHeader(*pOstream) << "Sending initial blob: " << m_buf2[0].cbBuffer << " bytes...\r\n";
            ReleaseTraceStream();
        }

        m_dwState = STATE_INITIAL_BLOB_SENT;
        dwRes = CAsyncTCPSocket::Write((LPCSTR)m_buf2[0].pvBuffer, m_buf2[0].cbBuffer);
        if(dwRes)
        {
            m_securityFuncTbl.FreeContextBuffer(m_buf2[0].pvBuffer);
            m_buf2[0].pvBuffer = NULL;
            m_securityFuncTbl.DeleteSecurityContext(&m_hContext);
            ZeroMemory(&m_hContext, sizeof(m_hContext));
        }
        // else, waiting for I/O completion.
    }
    else
        dwRes = InitiateClientHandshakeLoop(FALSE);

    return (dwRes == 0 ? ERROR_IO_PENDING : dwRes);
}

DWORD CAsyncSecureSocket::Close()
{
    DWORD dwRes = 0;

    // If physical TCP socket is closed, then just complete the closure,
    // and do not care to notify.
    if(!CAsyncTCPSocket::IsOpen())
    {
        m_bConnected = FALSE;

        // Deallocate the read buffer.
        FreeReadBuffer();

        // Reinitialize data members.
        Reset();
    }
    else
    {
        if(m_bConnected)
        {
            SecBufferDesc OutBuffer;
            DWORD dwType = SCHANNEL_SHUTDOWN;

            m_buf1[0].pvBuffer   = &dwType;
            m_buf1[0].BufferType = SECBUFFER_TOKEN;
            m_buf1[0].cbBuffer   = sizeof(dwType);

            OutBuffer.cBuffers  = 1;
            OutBuffer.pBuffers  = m_buf1;
            OutBuffer.ulVersion = SECBUFFER_VERSION;

            DWORD dwStatus = m_securityFuncTbl.ApplyControlToken(&m_hContext, &OutBuffer);

            if(SUCCEEDED(dwStatus)) 
            {
                DWORD           dwSSPIFlags;
                DWORD           dwSSPIOutFlags;
                TimeStamp       tsExpiry;

                // Build an SSL close notify message.
                dwSSPIFlags = ISC_REQ_SEQUENCE_DETECT   |
                              ISC_REQ_REPLAY_DETECT     |
                              ISC_REQ_CONFIDENTIALITY   |
                              ISC_RET_EXTENDED_ERROR    |
                              ISC_REQ_ALLOCATE_MEMORY   |
                              ISC_REQ_STREAM;

                m_buf1[0].pvBuffer   = NULL;
                m_buf1[0].BufferType = SECBUFFER_TOKEN;
                m_buf1[0].cbBuffer   = 0;

                OutBuffer.cBuffers  = 1;
                OutBuffer.pBuffers  = m_buf1;
                OutBuffer.ulVersion = SECBUFFER_VERSION;

                dwStatus = m_securityFuncTbl.InitializeSecurityContext(
                    &m_hCredentials,
                    &m_hContext,
                    NULL,
                    dwSSPIFlags,
                    0,
                    DATA_REP,
                    NULL,
                    0,
                    &m_hContext,
                    &OutBuffer,
                    &dwSSPIOutFlags,
                    &tsExpiry);

                if(SUCCEEDED(dwStatus)) 
                {
                    // Send the close notify message to the server.
                    if(m_buf1[0].pvBuffer != NULL && m_buf1[0].cbBuffer != 0)
                    {
                        dwRes = WriteHandshakeData(STATE_CLOSE_DATA_SENT);
                        if(dwRes)
                            CompleteClose(TRUE);
                        // else, if WriteHandshakeData succeeded, wait for I/O completion.
                    }
                }
                else
                    dwRes = dwStatus;
            }
            else
                dwRes = dwStatus;
        }
        else
        {
            // Only the TCP connection is established: close it!
            CompleteClose(TRUE);
        }
    }

    return (dwRes == 0 ? ERROR_IO_PENDING : dwRes);
}

DWORD CAsyncSecureSocket::CompleteClose(BOOL bNotify)
{
    DWORD dwRes = 0;
    dwRes = CAsyncTCPSocket::Close();   // Close the socket.
    m_bConnected = FALSE;

    // Deallocate the read buffer.
    FreeReadBuffer();

    // Reinitialize data members.
    IocpEvent* pTarget = m_pEventTarget;    // Cache event target before reset.
    Reset();

    // Notify the caller.
    if(bNotify)
    {
        if(dwRes == 0)
            pTarget->OnEvent(0, FALSE, EVENT_TYPE_CONNECTION_CLOSED);
        else
            pTarget->OnEventFailed(0, FALSE, EVENT_TYPE_CONNECTION_CLOSED, dwRes);
    }

    return dwRes;
}

DWORD CAsyncSecureSocket::CompleteCloseAcknowledge()
{
    DWORD dwRes = 0;
    dwRes = CAsyncTCPSocket::Close();   // Close the socket.
    m_bConnected = FALSE;

    // Deallocate the read buffer.
    FreeReadBuffer();

    // Reinitialize data members.
    IocpEvent* pTarget = m_pEventTarget;    // Cache event target before reset.
    Reset();

    // Notify the caller.
    if(dwRes == 0)
        pTarget->OnEvent(0, FALSE, EVENT_TYPE_CONNECTION_CLOSED);
    else
        pTarget->OnEventFailed(0, FALSE, EVENT_TYPE_CONNECTION_CLOSED, dwRes);

    return dwRes;
}

DWORD CAsyncSecureSocket::Write(LPCSTR zBuffer, DWORD dwBufferLength)
{
    WSABUF wsaBuf = {dwBufferLength, (LPSTR)zBuffer};
    return Write(&wsaBuf, 1);
}

DWORD CAsyncSecureSocket::Write(LPWSABUF lpBuffers, DWORD dwBufferCount)
{
    if(dwBufferCount == 0 || lpBuffers == NULL)
        return ERROR_INVALID_PARAMETER; // There must be at least one buffer!

    // Process buffer lengths and compute total length.
    m_dwTotalBufLen = 0;
    for(DWORD dw = 0; dw < dwBufferCount; ++dw)
    {
        // If length is zero, then assume buffer is null terminated.
        if(lpBuffers[dw].len == 0)
            lpBuffers[dw].len = lstrlen(lpBuffers[dw].buf);
        m_dwTotalBufLen += lpBuffers[dw].len;
    }

    SECURITY_STATUS scRet = 0;
    SecPkgContext_StreamSizes Sizes;

    // Read stream encryption properties.
    scRet = m_securityFuncTbl.QueryContextAttributes(
        &m_hContext,
        SECPKG_ATTR_STREAM_SIZES,
        &Sizes);
    if(scRet == SEC_E_OK)
    {
        // The sendable buffer is limited in size!
        if(m_dwTotalBufLen > Sizes.cbMaximumMessage)
            m_dwTotalBufLen = Sizes.cbMaximumMessage;   // Only the Sizes.cbMaximumMessage first bytes will be sent.

        // Check that the working buffer's size is sufficient. The plaintext sent to EncryptMessage
        // should never be more than 'Sizes.cbMaximumMessage', so a buffer 
        // size of this plus the header and trailer sizes should be safe enough.
        DWORD dwIoBufferLength = Sizes.cbHeader + Sizes.cbMaximumMessage
                               + Sizes.cbTrailer;

        if(m_dwReadBufLen < dwIoBufferLength)
            return ERROR_NOT_ENOUGH_MEMORY;

        // Use the read buffer.
        m_buf2[0].pvBuffer     = m_pReadBuffer;
        m_buf2[0].cbBuffer     = Sizes.cbHeader;
        m_buf2[0].BufferType   = SECBUFFER_STREAM_HEADER;

        m_buf2[1].pvBuffer     = ((PBYTE)m_buf2[0].pvBuffer) + m_buf2[0].cbBuffer;
        m_buf2[1].cbBuffer     = m_dwTotalBufLen;
        m_buf2[1].BufferType   = SECBUFFER_DATA;

        CopyFromClientBuffers(lpBuffers, dwBufferCount, m_buf2[1].pvBuffer, m_dwTotalBufLen);

        m_buf2[2].pvBuffer     = ((PBYTE)m_buf2[1].pvBuffer) + m_buf2[1].cbBuffer;
        m_buf2[2].cbBuffer     = Sizes.cbTrailer;
        m_buf2[2].BufferType   = SECBUFFER_STREAM_TRAILER;

        m_buf2[3].BufferType   = SECBUFFER_EMPTY;

        SecBufferDesc Message;
        Message.ulVersion       = SECBUFFER_VERSION;
        Message.cBuffers        = 4;
        Message.pBuffers        = m_buf2;

        scRet = m_securityFuncTbl.EncryptMessage(&m_hContext, 0, &Message, 0);
        if(SUCCEEDED(scRet))
        {
            scRet = WriteData();
            // If WriteData succeeded, wait for I/O completion.
        }
    }

    return (DWORD)scRet;
}

DWORD CAsyncSecureSocket::WriteData(DWORD dwState /* = STATE_DATA_SENT */)
{
    DWORD dwRes = 0;

    DWORD dwCountToSend = m_buf2[0].cbBuffer + m_buf2[1].cbBuffer + m_buf2[2].cbBuffer;
    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
    {
        EnableTraceStream(m_pTrace);
            TraceHeader(*pOstream) << "Sending data: " << dwCountToSend << " bytes...\r\n";
        ReleaseTraceStream();
    }

    m_dwState = dwState;
    dwRes = CAsyncTCPSocket::Write((LPCSTR)m_buf2[0].pvBuffer, dwCountToSend);
    if(dwRes)
    {
        m_buf2[0].cbBuffer = 0;
        m_securityFuncTbl.DeleteSecurityContext(&m_hContext);
        ZeroMemory(&m_hContext, sizeof(m_hContext));
        m_ss = SEC_E_INTERNAL_ERROR;
    }
    // else, waiting for I/O completion.

    return dwRes;
}

DWORD CAsyncSecureSocket::Read(LPSTR zBuffer, DWORD dwBufferLength)
{
    m_oneBuf.buf = zBuffer;
    m_oneBuf.len = dwBufferLength;
    return Read(&m_oneBuf, 1);
}

DWORD CAsyncSecureSocket::Read(LPWSABUF lpBuffers, DWORD dwBufferCount)
{
    DWORD dwRes = 0;

    if(dwBufferCount == 0 || lpBuffers == NULL)
        return ERROR_INVALID_PARAMETER; // There must be at least one buffer!

    // Cache the client buffer's address and length.
    m_pClientBuffer = lpBuffers;
    m_dwBufferCount = dwBufferCount;
    m_dwTotalBufLen = 0;
    for(DWORD dw = 0; dw < m_dwBufferCount; dw++)
    {
        if(m_pClientBuffer[dw].buf)
            m_pClientBuffer[dw].buf[0] = 0;
        m_dwTotalBufLen += m_pClientBuffer[dw].len;
    }

    if(m_dwNextChunk && m_pDataBuffer)
    {
        // There is a next chunk to read from the read buffer, which has already
        // been decrypted. This is due to the fact that the dwBufferLength argument
        // above indicates a zBuffer too small to contain a whole decrypted message
        // from the read buffer.
        DWORD dwSize = m_pDataBuffer->cbBuffer - m_dwNextChunk;
        DWORD dwOffset = m_dwNextChunk;
        if(dwSize > m_dwTotalBufLen)
        {
            dwSize = m_dwTotalBufLen;
            m_dwNextChunk += dwSize;        // Offset of next chunk.
        }
        else
            m_dwNextChunk = 0;  // No next chunk.

        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
        {
            EnableTraceStream(m_pTrace);
                TraceHeader(*pOstream) << "Extracting chunk from read buffer: " << dwSize << " bytes\r\n";
            ReleaseTraceStream();
        }

        CopyToClientBuffers(((PBYTE)m_pDataBuffer->pvBuffer) + dwOffset, dwSize);

        // Notify the caller.
        m_pEventTarget->OnEvent(dwSize, FALSE, EVENT_TYPE_MESSAGE_RECEIVED);
    }
    else
    {
        if(m_pExtraBuffer)
        {
            if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
            {
                EnableTraceStream(m_pTrace);
                    TraceHeader(*pOstream) << "Moving extra data into read buffer\r\n";
                ReleaseTraceStream();
            }

            // There is extra data to read from the client buffer, which has
            // been filled during a previous read.
            MoveMemory(m_buf2[0].pvBuffer, m_pExtraBuffer->pvBuffer, m_pExtraBuffer->cbBuffer);
            m_dwBytesReceived = m_pExtraBuffer->cbBuffer;
        }
        else
        {
            m_dwBytesReceived = 0;

            m_ss = SEC_E_INCOMPLETE_MESSAGE;

            // Point to the read buffer.
            m_buf2[0].pvBuffer = m_pReadBuffer;
            m_buf2[0].cbBuffer = 0;
        }

        m_dwState = STATE_NONE;
        dwRes = ReadLoop();
    }

    return dwRes;
}

DWORD CAsyncSecureSocket::ReadLoop()
{
    DWORD dwRes = 0;
    DWORD dwSize = 0;
    BOOL  bHaveReadValidData = FALSE;
    m_pDataBuffer = m_pExtraBuffer = NULL;

    do
    {
        switch(m_dwState)
        {
            case STATE_NONE:
                if(m_dwBytesReceived == 0 || m_ss == SEC_E_INCOMPLETE_MESSAGE)
                {
                    DWORD dwCountToRead = m_dwReadBufLen - m_dwBytesReceived;
                    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                    {
                        EnableTraceStream(m_pTrace);
                            TraceHeader(*pOstream) << "Receiving data: " << dwCountToRead << " bytes...\r\n";
                        ReleaseTraceStream();
                    }

                    m_dwState = STATE_DATA_RECEIVED;
                    dwRes = CAsyncTCPSocket::Read(((LPSTR)m_buf2[0].pvBuffer) + m_dwBytesReceived, dwCountToRead);
                    if(dwRes)
                        return dwRes;
                    // else, waiting for I/O completion.
                    goto exitloop;
                }

                // Fall through.

            case STATE_DATA_RECEIVED:
                {
                    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                    {
                        EnableTraceStream(m_pTrace);
                            TraceHeader(*pOstream) << "Let's decrypt.\r\n";
                        ReleaseTraceStream();
                    }

                    SecBufferDesc SecBuffDesc = {0};

                    // Setup initial data buffer to point to extra data
                    m_buf2[0].BufferType = SECBUFFER_DATA;
                    m_buf2[0].cbBuffer   = m_dwBytesReceived;

                    // Setup 3 empty out buffers
                    m_buf2[1].BufferType = SECBUFFER_EMPTY;
                    m_buf2[2].BufferType = SECBUFFER_EMPTY;
                    m_buf2[3].BufferType = SECBUFFER_EMPTY;

                    // Setup descriptor
                    SecBuffDesc.ulVersion = SECBUFFER_VERSION;
                    SecBuffDesc.cBuffers  = 4;
                    SecBuffDesc.pBuffers  = m_buf2;

                    // Attempt to decrypt the received data.
                    m_ss = m_securityFuncTbl.DecryptMessage(&m_hContext, &SecBuffDesc, 0, NULL);

                    if(m_ss == SEC_E_INCOMPLETE_MESSAGE)
                    {
                        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                        {
                            EnableTraceStream(m_pTrace);
                                TraceHeader(*pOstream) << "Message incomplete: read more...\r\n";
                            ReleaseTraceStream();
                        }

                        // The input buffer contains only a fragment of an
                        // encrypted record. Loop around and read some more
                        // data.
                        m_dwState = STATE_NONE;
                        continue;
                    }

                    // Server signalled end of session. Watch out for old NT4.0 value!
                    if(m_ss == SEC_E_CONTEXT_EXPIRED || m_ss == SEC_E_CONTEXT_EXPIRED_OLD)
                    {
                        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                        {
                            EnableTraceStream(m_pTrace);
                                TraceHeader(*pOstream) << "Context expired: the server closed the SSL session.\r\n";
                            ReleaseTraceStream();
                        }

                        //// Acknowledge the disconnection.

                        // Read stream encryption properties.
                        SecPkgContext_StreamSizes Sizes;
                        SECURITY_STATUS scRet = m_securityFuncTbl.QueryContextAttributes(
                            &m_hContext,
                            SECPKG_ATTR_STREAM_SIZES,
                            &Sizes);
                        if(scRet == SEC_E_OK)
                        {
                            m_buf2[0].pvBuffer     = m_pReadBuffer;
                            m_buf2[0].cbBuffer     = Sizes.cbHeader;
                            m_buf2[0].BufferType   = SECBUFFER_STREAM_HEADER;

                            m_buf2[1].pvBuffer     = "";
                            m_buf2[1].cbBuffer     = 0;
                            m_buf2[1].BufferType   = SECBUFFER_DATA;

                            m_buf2[2].pvBuffer     = ((PBYTE)m_buf2[0].pvBuffer) + m_buf2[0].cbBuffer;
                            m_buf2[2].cbBuffer     = Sizes.cbTrailer;
                            m_buf2[2].BufferType   = SECBUFFER_STREAM_TRAILER;

                            m_buf2[3].BufferType   = SECBUFFER_EMPTY;

                            SecBufferDesc Message;
                            Message.ulVersion       = SECBUFFER_VERSION;
                            Message.cBuffers        = 4;
                            Message.pBuffers        = m_buf2;

                            scRet = m_securityFuncTbl.EncryptMessage(&m_hContext, 0, &Message, 0);
                            if(SUCCEEDED(scRet))
                            {
                                scRet = WriteData(STATE_CLOSE_ACKNOWLEDGE_DATA_SENT);
                                // If WriteData succeeded, wait for I/O completion.
                                if(scRet != 0)
                                    dwRes = scRet;
                            }
                            else
                            {
                                dwRes = scRet;
                                if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                                {
                                    EnableTraceStream(m_pTrace);
                                        TraceHeader(*pOstream) << "Error returned by EncryptMessage: 0x"
                                            << std::hex << scRet << std::dec << ".\r\n";
                                    ReleaseTraceStream();
                                }
                            }
                        }
                        else
                            dwRes = scRet;

                        goto exitloop;
                    }

                    // Error returned by DecryptMessage.
                    if(m_ss != SEC_E_OK && 
                       m_ss != SEC_I_RENEGOTIATE)
                    {
                        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                        {
                            EnableTraceStream(m_pTrace);
                                TraceHeader(*pOstream) << "Error returned by DecryptMessage: 0x"
                                    << std::hex << m_ss << std::dec << ".\r\n";
                            ReleaseTraceStream();
                        }

                        dwRes = m_ss;
                        goto exitloop;
                    }

                    // Locate data and optional extra buffers.
                    for(int i = 1; i < 4; i++)
                    {
                        if(m_pDataBuffer == NULL && m_buf2[i].BufferType == SECBUFFER_DATA)
                            m_pDataBuffer = m_buf2 + i;
                        if(m_pExtraBuffer == NULL && m_buf2[i].BufferType == SECBUFFER_EXTRA)
                            m_pExtraBuffer = m_buf2 + i;
                    }

                    // Copy the results to the client buffer.
                    if(m_pDataBuffer)
                    {
                        dwSize = m_pDataBuffer->cbBuffer;   // Size of first chunk to copy.
                        if(dwSize > m_dwTotalBufLen)
                        {
                            dwSize = m_dwTotalBufLen;
                            m_dwNextChunk = dwSize;         // Offset of second chunk.
                        }
                        else
                            m_dwNextChunk = 0;  // No next chunk.

                        CopyToClientBuffers(m_pDataBuffer->pvBuffer, dwSize);
                    }

                    if(m_ss == SEC_I_RENEGOTIATE)
                    {
                        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                        {
                            EnableTraceStream(m_pTrace);
                                TraceHeader(*pOstream) <<"Renegotiation requested!\r\n";
                            ReleaseTraceStream();
                        }

                        // Here, there is no data nor extra data.
                        dwRes = InitiateClientHandshakeLoop(TRUE);
                        if(dwRes)
                        {
                            // Renegotiation failed.
                            m_bRenegotiate = FALSE;
                        }
                        goto exitloop;  // Failure, or waiting for asynchronous I/O completion.
                    }

                    if(m_pDataBuffer)
                        bHaveReadValidData = TRUE;

                    m_dwState = STATE_NONE;
                }
        }
    } while(m_ss == SEC_E_INCOMPLETE_MESSAGE);

exitloop:

    if(dwRes)
    {
        m_pClientBuffer = NULL;
        m_dwBufferCount = 0;
    }
    else if(bHaveReadValidData)
    {
        // Notify the caller.
        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
        {
            EnableTraceStream(m_pTrace);
                TraceHeader(*pOstream) << "Notifying event target that "
                    << dwSize
                    << " bytes of decrypted message are available.\r\n";
            ReleaseTraceStream();
        }
        m_pEventTarget->OnEvent(dwSize, FALSE, EVENT_TYPE_MESSAGE_RECEIVED);
    }

    return dwRes;
}

BOOL CAsyncSecureSocket::IsOpen()
{
    return m_bConnected;
}

DWORD CAsyncSecureSocket::ReadHandshakeData()
{
    DWORD dwRes = 0;

    DWORD dwCountToRead = m_dwReadBufLen - m_dwBytesReceived;
    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
    {
        EnableTraceStream(m_pTrace);
            TraceHeader(*pOstream) << "Receiving handshake data: " << dwCountToRead << " bytes...\r\n";
        ReleaseTraceStream();
    }

    m_dwState = STATE_HANDSHAKE_DATA_RECEIVED;
    dwRes = CAsyncTCPSocket::Read(((LPSTR)m_buf2[0].pvBuffer) + m_dwBytesReceived, dwCountToRead);
    if(dwRes)
        m_ss = SEC_E_INTERNAL_ERROR;
    // else, waiting for I/O completion.

    return dwRes;
}

DWORD CAsyncSecureSocket::WriteHandshakeData(DWORD dwNextState)
{
    DWORD dwRes = 0;

    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
    {
        EnableTraceStream(m_pTrace);
            if(dwNextState == STATE_CLOSE_DATA_SENT)
                TraceHeader(*pOstream) << "Sending close handshake data: " << m_buf1[0].cbBuffer << " bytes...\r\n";
            else
                TraceHeader(*pOstream) << "Sending handshake data: " << m_buf1[0].cbBuffer << " bytes...\r\n";
        ReleaseTraceStream();
    }
    
    m_dwState = dwNextState;
    dwRes = CAsyncTCPSocket::Write((LPCSTR)m_buf1[0].pvBuffer, m_buf1[0].cbBuffer);
    if(dwRes)
    {
        m_securityFuncTbl.FreeContextBuffer(m_buf1[0].pvBuffer);
        m_securityFuncTbl.DeleteSecurityContext(&m_hContext);
        ZeroMemory(&m_hContext, sizeof(m_hContext));
        m_ss = SEC_E_INTERNAL_ERROR;
    }
    // else, waiting for I/O completion.

    return dwRes;
}

SECURITY_STATUS CAsyncSecureSocket::ProcessBlob(DWORD& dwSSPIOutFlags)
{
    SecBufferDesc   InBuffer;
    SecBufferDesc   OutBuffer;
    TimeStamp       tsExpiry;

    // Set up the input buffers. Buffer 0 is used to pass in data
    // received from the server. Schannel will consume some or all
    // of this. Leftover data (if any) will be placed in buffer 1 and
    // given a buffer type of SECBUFFER_EXTRA.

    m_buf2[0].BufferType = SECBUFFER_TOKEN;
    m_buf2[0].cbBuffer   = m_dwBytesReceived;

    m_buf2[1].pvBuffer   = NULL;
    m_buf2[1].cbBuffer   = 0;
    m_buf2[1].BufferType = SECBUFFER_EMPTY;

    InBuffer.cBuffers    = 2;
    InBuffer.pBuffers    = m_buf2;
    InBuffer.ulVersion   = SECBUFFER_VERSION;

    // Set up the output buffers. These are initialized to NULL
    // so as to make it less likely we'll attempt to free random
    // garbage later.

    m_buf1[0].pvBuffer  = NULL;
    m_buf1[0].BufferType= SECBUFFER_TOKEN;
    m_buf1[0].cbBuffer  = 0;

    OutBuffer.cBuffers  = 1;
    OutBuffer.pBuffers  = m_buf1;
    OutBuffer.ulVersion = SECBUFFER_VERSION;

    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
    {
        EnableTraceStream(m_pTrace);
            TraceHeader(*pOstream) << "Calling InitializeSecurityContext() to process "
                                   << m_buf2[0].cbBuffer << " bytes of received blob.\r\n";
        ReleaseTraceStream();
    }

    SECURITY_STATUS ss;
    if(m_pSChannelBugLock)
    {
        CLock lock(m_pSChannelBugLock);

        // Call InitializeSecurityContext.
        ss = m_securityFuncTbl.InitializeSecurityContext(
            &m_hCredentials,
            &m_hContext,
            NULL,
            ISC_REQ_MANUAL_CRED_VALIDATION |    // L.M. ### This is a temporary hack, to establish a permissive secure session.
            ISC_REQ_SEQUENCE_DETECT   |
            ISC_REQ_REPLAY_DETECT     |
            ISC_REQ_CONFIDENTIALITY   |
            ISC_RET_EXTENDED_ERROR    |
            ISC_REQ_ALLOCATE_MEMORY   |
            ISC_REQ_STREAM,
            0,
            DATA_REP,
            &InBuffer,
            0,
            NULL,
            &OutBuffer,
            &dwSSPIOutFlags,
            &tsExpiry);
    }
    else
    {
        // Call InitializeSecurityContext.
        ss = m_securityFuncTbl.InitializeSecurityContext(
            &m_hCredentials,
            &m_hContext,
            NULL,
            ISC_REQ_MANUAL_CRED_VALIDATION |    // L.M. ### This is a temporary hack, to establish a permissive secure session.
            ISC_REQ_SEQUENCE_DETECT   |
            ISC_REQ_REPLAY_DETECT     |
            ISC_REQ_CONFIDENTIALITY   |
            ISC_RET_EXTENDED_ERROR    |
            ISC_REQ_ALLOCATE_MEMORY   |
            ISC_REQ_STREAM,
            0,
            DATA_REP,
            &InBuffer,
            0,
            NULL,
            &OutBuffer,
            &dwSSPIOutFlags,
            &tsExpiry);
    }

    return ss;
}

void CAsyncSecureSocket::ProcessSuccessfulHandshake()
{
    // If the "extra" buffer contains data, this is encrypted application
    // protocol layer stuff. It needs to be saved. The application layer
    // will later decrypt it with DecryptMessage.
    // This is the case, for instance, when an attempt to decrypt has
    // yielded a renegotiation, which in turn has extra data on output. This
    // extra data will then be reused by a new decrypt attempt.

    MoveExtraHandshakeData();
}

void CAsyncSecureSocket::MoveExtraHandshakeData()
{
    if (m_buf2[1].BufferType == SECBUFFER_EXTRA )
    {
        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
        {
            EnableTraceStream(m_pTrace);
                TraceHeader(*pOstream) << "Moving extra handshake data. Next to read: "
                                       << m_buf2[1].cbBuffer << " bytes.\r\n";
            ReleaseTraceStream();
        }

        MoveMemory(m_buf2[0].pvBuffer,
                   ((PBYTE)m_buf2[0].pvBuffer) + (m_dwBytesReceived - m_buf2[1].cbBuffer),
                   m_buf2[1].cbBuffer);

        m_dwBytesReceived    = m_buf2[1].cbBuffer;
        m_buf2[0].BufferType = SECBUFFER_TOKEN;
    }
    else
    {
        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
        {
            EnableTraceStream(m_pTrace);
                TraceHeader(*pOstream) << "No extra handshake data.\r\n";
            ReleaseTraceStream();
        }

        m_buf2[0].cbBuffer = 0;
        m_dwBytesReceived  = 0;
    }
}

BOOL CAsyncSecureSocket::MakeClientCertificateCredentials()
{
    if(m_zCertifName[0] == 0)   // Anonymous user.
    {
        // Notify the caller, to synchronously give him a chance to provide a valid client
        // certificate.
        if(m_pEventTarget)
        {
            // The following call, is honored by the client, may have set a certificate name.
            m_pEventTarget->OnEvent(0, FALSE, EVENT_TYPE_PLEASE_PROVIDE_CLIENT_CERTIFICATE);
            if(m_zCertifName[0] == 0)
                return TRUE;    // No, a client certificate has not been set. Try to continue anonymously.
            else
            {
                if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                {
                    EnableTraceStream(m_pTrace);
                        TraceHeader(*pOstream) << "The caller provided a client certificate name!\r\n";
                    ReleaseTraceStream();
                }
            }
        }
        else
            return TRUE;    // Cannot notify the caller. Try to continue anonymously.
    }

    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
    {
        EnableTraceStream(m_pTrace);
            TraceHeader(*pOstream) << "Creating client credentials for the client certificate named: '"
                << m_zCertifName
                << "'.\r\n";
        ReleaseTraceStream();
    }

    BOOL bRet = FALSE;

    // Lookup the certificate in the certificate cache.
    CertificateToHandle::const_iterator pCert;
    BOOL bInCache;
    {
        CLock lock(m_pLock);
        pCert = m_credentialsCache.find(m_zCertifName);
        bInCache = (pCert != m_credentialsCache.end());
    }
    if(bInCache)
    {
        // The credentials handle for this certificate is already in the cache.
        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
        {
            EnableTraceStream(m_pTrace);
                TraceHeader(*pOstream) << "Reusing cached certificate credentials!\r\n";
            ReleaseTraceStream();
        }
        m_hCredentials = pCert->second;
        bRet = TRUE;                // Success: got new certificate credentials!
    }
    else
    {
        // The credentials handle for this certificate is NOT in the cache; we need
        // to create it.
        if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
        {
            EnableTraceStream(m_pTrace);
                TraceHeader(*pOstream) << "Building new certificate credentials\r\n";
            ReleaseTraceStream();
        }
        PCCERT_CONTEXT pCertContext = NULL;
        ULONG lStore = m_bRunAsSystem ? CERT_SYSTEM_STORE_LOCAL_MACHINE
            : CERT_SYSTEM_STORE_CURRENT_USER;
        HCERTSTORE hMyCertStore = CertOpenStore(CERT_STORE_PROV_SYSTEM_A,
            X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, 0, lStore, "MY");
        if(hMyCertStore != NULL)
        {
            // Fill in an attribute structure for the certificate common name.
            CHAR pszCommonName[1024];
    #ifdef UNICODE
            if(WideCharToMultiByte(CP_ACP, 0, m_zCertifName, -1, pszCommonName,
                chDIMOF(pszCommonName), NULL, NULL) != 0)
    #else
            lstrcpy(pszCommonName, m_zCertifName);
    #endif
            {
                CERT_RDN_ATTR certRDNAttr[1];
                certRDNAttr[0].pszObjId     = szOID_COMMON_NAME;
                certRDNAttr[0].dwValueType  = CERT_RDN_PRINTABLE_STRING;
                certRDNAttr[0].Value.pbData = (PBYTE) pszCommonName;
                certRDNAttr[0].Value.cbData = lstrlenA(pszCommonName);

                CERT_RDN certRDN = {1, certRDNAttr};

                // Find the certificate context
                pCertContext = CertFindCertificateInStore(hMyCertStore,
                    X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, 0, CERT_FIND_SUBJECT_ATTR,
                    &certRDN, NULL);
                if(pCertContext != NULL)
                {
                    // Fill in an SCHANNEL_CRED variable with certificate information.
                    SCHANNEL_CRED sslCredentials = {0};
                    sslCredentials.dwVersion = SCHANNEL_CRED_VERSION;
                    sslCredentials.cCreds    = 1;
                    sslCredentials.paCred    = &pCertContext;

                    // Get a credentials handle.
                    TimeStamp tsExpires;
                    CredHandle hCertCredentials;
                    SECURITY_STATUS ss = m_securityFuncTbl.AcquireCredentialsHandle(
                        NULL, UNISP_NAME, SECPKG_CRED_OUTBOUND, NULL,
                        &sslCredentials, NULL, NULL, &hCertCredentials, &tsExpires);
                    if (ss == SEC_E_OK)
                    {
                        m_hCredentials = hCertCredentials;

                        // Insert the new credentials handle into the cache.
                        {
                            CLock lock(m_pLock);
                            m_credentialsCache.insert(std::make_pair(std::string(m_zCertifName), m_hCredentials));
                        }

                        bRet = TRUE;    // Success: got new certificate credentials!
                    }
                }
            }
            CertCloseStore(hMyCertStore, 0);
        }

        // Free the certificate context
        if(pCertContext != NULL)
            CertFreeCertificateContext(pCertContext);
    }

    return bRet;
}

DWORD CAsyncSecureSocket::InitiateClientHandshakeLoop(BOOL bRenegotiate)
{
    DWORD dwRes = 0;

    // Point to the read buffer.
    m_buf2[0].pvBuffer = m_pReadBuffer;
    m_buf2[0].cbBuffer = 0;
    m_dwBytesReceived  = 0;

    m_bDoRead = !bRenegotiate;
    m_bRenegotiate = bRenegotiate;

    m_ss = SEC_I_CONTINUE_NEEDED;
    m_dwState = STATE_NONE;
    dwRes = ClientHandshakeLoop();

    return dwRes;
}

DWORD CAsyncSecureSocket::ClientHandshakeLoop()
{
    DWORD dwRes = 0;
    DWORD dwSSPIOutFlags = 0;

    //// Loop until the handshake is finished or an error occurs.

    do
	{
        switch(m_dwState)
        {
            case STATE_NONE:

                // Read data from server.
                if(m_dwBytesReceived == 0 || m_ss == SEC_E_INCOMPLETE_MESSAGE)
                {
                    if(m_bDoRead)
                    {
                        dwRes = ReadHandshakeData();
                        if(dwRes)
                            goto loopexit;
                        else
                            return 0;   // Waiting for I/O completion.
                    }
                    else
                        m_bDoRead = TRUE;
                }

                // Fall through.

            case STATE_HANDSHAKE_DATA_RECEIVED:

                // Reset the state.
                m_dwState = STATE_NONE;

                dwSSPIOutFlags = 0;
                m_ss = ProcessBlob(dwSSPIOutFlags);

                // If InitializeSecurityContext was successful (or if the error was 
                // one of the special extended ones), send the contents of the output
                // buffer to the server.
                if(m_ss == SEC_E_OK                ||
                   m_ss == SEC_I_CONTINUE_NEEDED   ||
                   FAILED(m_ss) && (dwSSPIOutFlags & ISC_RET_EXTENDED_ERROR))
                {
                    if(m_buf1[0].cbBuffer != 0 && m_buf1[0].pvBuffer != NULL)
                    {
                        dwRes = WriteHandshakeData(STATE_HANDSHAKE_DATA_SENT);
                        if(dwRes)
                        {
                            dwRes = m_ss;
                            goto loopexit;
                        }
                        else
                            return 0;   // Waiting for I/O completion.
                    }
                }

                // Fall through.

            case STATE_HANDSHAKE_DATA_SENT:

                // Reset the state.
                m_dwState = STATE_NONE;

                // If InitializeSecurityContext returned SEC_E_INCOMPLETE_MESSAGE,
                // then we need to read more data from the server and try again.
                if(m_ss == SEC_E_INCOMPLETE_MESSAGE)
                    continue;

                // If InitializeSecurityContext returned SEC_E_OK, then the 
                // handshake completed successfully.
                if(m_ss == SEC_E_OK)
                {
                    ProcessSuccessfulHandshake();

                    // Bail out to quit
                    goto loopexit;
                }

                // Check for fatal error.
                if(FAILED(m_ss))
                {
                    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                    {
                        const char *pMsg;
                        switch(m_ss)
                        {
                            case SEC_E_UNTRUSTED_ROOT:
                                pMsg = "The certificate chain was issued by an untrusted authority.\r\n";
                                break;

                            case SEC_E_CERT_EXPIRED:
                                pMsg = "The received certificate has expired.\r\n";
                                break;

                            case SEC_E_WRONG_PRINCIPAL:
                                pMsg = "The target principal name is incorrect.\r\n";
                                break;

                            default:
                                pMsg = "The handshake loop failed.\r\n";
                                break;
                        }
                        EnableTraceStream(m_pTrace);
                            TraceHeader(*pOstream) << "Error 0x" << std::hex << m_ss << std::dec << ". " << pMsg;
                        ReleaseTraceStream();
                    }
                    goto loopexit;
                }

                // If InitializeSecurityContext returned SEC_I_INCOMPLETE_CREDENTIALS,
                // then the server just requested client authentication. 
                if(m_ss == SEC_I_INCOMPLETE_CREDENTIALS)
                {
                    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                    {
                        EnableTraceStream(m_pTrace);
                            TraceHeader(*pOstream) << "The server reported that client credentials are incomplete.\r\n";
                        ReleaseTraceStream();
                    }
                    MakeClientCertificateCredentials();

                    // Go around again.
                    m_bDoRead = FALSE;
                    m_ss = SEC_I_CONTINUE_NEEDED;
                    continue;
                }

                // Copy any leftover data from the "extra" buffer, and go around again.
                MoveExtraHandshakeData();
        }

    } while(m_ss == SEC_I_CONTINUE_NEEDED        ||
            m_ss == SEC_E_INCOMPLETE_MESSAGE     ||
            m_ss == SEC_I_INCOMPLETE_CREDENTIALS);

loopexit:

    // Delete the security context in the case of a fatal error.
    if(FAILED(m_ss))
    {
        m_securityFuncTbl.DeleteSecurityContext(&m_hContext);
        ZeroMemory(&m_hContext, sizeof(m_hContext));
    }

    dwRes = m_ss;

    // Notify the caller.
    if(!m_bRenegotiate)
    {
        if(dwRes == 0)
        {
            m_bConnected = TRUE;
            m_pEventTarget->OnEvent(0, FALSE, EVENT_TYPE_CONNECTION_ESTABLISHED);
            // The buf2[0] buffer may (if buf2[0].cbBuffer != 0), contain extra data.
            // If this client handshake loop was for a renegotiation, then the extra
            // data is to be used by a Read().
        }
        else
            m_pEventTarget->OnEventFailed(0, FALSE, EVENT_TYPE_CONNECTION_ESTABLISHED, dwRes);
    }
    else
    {
        m_ss = SEC_E_INCOMPLETE_MESSAGE;
        DWORD dwRes2 = ReadLoop(); // Re-enter read loop!
        if(dwRes2)
            dwRes = dwRes2;
    }

    return dwRes;
}

void CAsyncSecureSocket::OnEvent(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType)
{
    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
    {
        EnableTraceStream(m_pTrace);
            TraceHeader(*pOstream) << ">> OnEvent enter with " << dwNumberOfBytesTransferred << " bytes transferred.\r\n";
        ReleaseTraceStream();
    }

    switch(m_dwState)
    {
        case STATE_INITIAL_BLOB_SENT:
            // Free output buffer, then initiate the client handshake loop.
            m_securityFuncTbl.FreeContextBuffer(m_buf2[0].pvBuffer);
            m_buf2[0].pvBuffer = NULL;
            if(dwNumberOfBytesTransferred)
            {
                if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                {
                    EnableTraceStream(m_pTrace);
                        TraceHeader(*pOstream) << "<-- Sent " << dwNumberOfBytesTransferred
                            << " bytes of initial handshake blob.\r\n";
                    ReleaseTraceStream();
                }
                InitiateClientHandshakeLoop(FALSE);
            }
            else
                m_pEventTarget->OnEventFailed(0, FALSE, EVENT_TYPE_CONNECTION_ESTABLISHED, ERROR_NOT_CONNECTED);
            break;

        case STATE_HANDSHAKE_DATA_RECEIVED:
            if(dwNumberOfBytesTransferred == 0)
            {
                // The server unexpectedly disconnected.
                m_pEventTarget->OnEventFailed(0, FALSE, EVENT_TYPE_CONNECTION_ESTABLISHED, ERROR_NOT_CONNECTED);
            }
            else
            {
                if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                {
                    EnableTraceStream(m_pTrace);
                        TraceHeader(*pOstream) << "--> Received " << dwNumberOfBytesTransferred
                            << " bytes of handshake data.\r\n";
                    ReleaseTraceStream();
                }
                m_dwBytesReceived += dwNumberOfBytesTransferred;
                ClientHandshakeLoop();  // Re-enter the client handshake loop!
            }
            break;

        case STATE_HANDSHAKE_DATA_SENT:
            if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
            {
                EnableTraceStream(m_pTrace);
                    TraceHeader(*pOstream) << "<-- Sent " << dwNumberOfBytesTransferred
                        << " bytes of handshake data.\r\n";
                ReleaseTraceStream();
            }
            m_securityFuncTbl.FreeContextBuffer(m_buf1[0].pvBuffer);
            m_buf1[0].pvBuffer = NULL;
            ClientHandshakeLoop();  // Re-enter the client handshake loop!
            break;

        case STATE_CLOSE_DATA_SENT:
            if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
            {
                EnableTraceStream(m_pTrace);
                    TraceHeader(*pOstream) << "<-- Sent " << dwNumberOfBytesTransferred
                        << " bytes of close handshake.\r\n";
                ReleaseTraceStream();
            }
            m_securityFuncTbl.FreeContextBuffer(m_buf1[0].pvBuffer);
            m_buf1[0].pvBuffer = NULL;
            m_securityFuncTbl.DeleteSecurityContext(&m_hContext);
            ZeroMemory(&m_hContext, sizeof(m_hContext));
            CompleteClose(TRUE);
            break;

        case STATE_DATA_SENT:
            if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
            {
                EnableTraceStream(m_pTrace);
                    TraceHeader(*pOstream) << "<-- Sent " << dwNumberOfBytesTransferred
                        << " bytes of data.\r\n";
                ReleaseTraceStream();
            }
            m_buf2[0].cbBuffer = 0;

            // Notify the caller. The number of bytes sent is m_dwTotalBufLen, because
            // dwNumberOfBytesTransferred is a physical number, and the caller doesn't
            // care about the SSL header, trailer and encryption.
            m_pEventTarget->OnEvent(m_dwTotalBufLen, FALSE, EVENT_TYPE_MESSAGE_SENT);
            break;

        case STATE_CLOSE_ACKNOWLEDGE_DATA_SENT:
            if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
            {
                EnableTraceStream(m_pTrace);
                    TraceHeader(*pOstream) << "<-- Sent (to acknowledge a close) " << dwNumberOfBytesTransferred
                        << " bytes of data.\r\n";
                ReleaseTraceStream();
            }
            {
                DWORD dwErr = m_securityFuncTbl.DeleteSecurityContext(&m_hContext);
                ZeroMemory(&m_hContext, sizeof(m_hContext));
                if(dwErr != SEC_E_OK)
                {
                    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                    {
                        EnableTraceStream(m_pTrace);
                            TraceHeader(*pOstream) << "Error returned by DeleteSecurityContext: 0x"
                                << std::hex << dwErr << std::dec << ".\r\n";
                        ReleaseTraceStream();
                    }
                }
            }

            CompleteCloseAcknowledge();
            break;

        case STATE_DATA_RECEIVED:
            if(dwNumberOfBytesTransferred == 0)
            {
                if(m_dwBytesReceived)
                {
                    // The server unexpectedly disconnected.
                    m_pEventTarget->OnEventFailed(0, FALSE, EVENT_TYPE_MESSAGE_RECEIVED, (DWORD)SEC_E_INTERNAL_ERROR);
                }
                else
                {
                    // The server disconnected.
                    m_pEventTarget->OnEvent(0, FALSE, EVENT_TYPE_MESSAGE_RECEIVED);
                    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                    {
                        EnableTraceStream(m_pTrace);
                            TraceHeader(*pOstream) << "The server disconnected.\r\n";
                        ReleaseTraceStream();
                    }
                }
            }
            else
            {
                if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
                {
                    EnableTraceStream(m_pTrace);
                        TraceHeader(*pOstream) << "--> Received " << dwNumberOfBytesTransferred
                            << " bytes of data.\r\n";
                    ReleaseTraceStream();
                }
                m_dwBytesReceived += dwNumberOfBytesTransferred;
                DWORD dwRes = ReadLoop(); // Re-enter read loop!
                if(dwRes)
                    m_pEventTarget->OnEventFailed(0, FALSE, EVENT_TYPE_MESSAGE_RECEIVED, dwRes);
            }
            break;

        default:
            // Forward the event to the caller!
            m_pEventTarget->OnEvent(dwNumberOfBytesTransferred, bIsIo, dwEventType);
            break;
    }

    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
    {
        EnableTraceStream(m_pTrace);
            TraceHeader(*pOstream) << "<< OnEvent exit.\r\n";
        ReleaseTraceStream();
    }
}

void CAsyncSecureSocket::OnEventFailed(DWORD /*dwNumberOfBytesTransferred*/, BOOL /*bIsIo*/, DWORD /*dwEventType*/, DWORD /*dwError*/)
{
    if(m_pTrace && m_pTrace->IsTrace("SecureSocket"))
    {
        EnableTraceStream(m_pTrace);
            TraceHeader(*pOstream) << "CAsyncSecureSocket::OnEventFailed().\r\n";
        ReleaseTraceStream();
    }
}

void CAsyncSecureSocket::CopyToClientBuffers(PVOID pvSource, DWORD dwLen)
{
    LPWSABUF pBuf = m_pClientBuffer;
    if(pBuf->buf)
    {
        while(dwLen)
        {
            // Copy data to the current buffer.
            DWORD dwChunk = pBuf->len;
            if(dwChunk > dwLen)
                dwChunk = dwLen;
            CopyMemory(pBuf->buf, pvSource, dwChunk);
            dwLen -= dwChunk;
            pvSource = (PVOID)(((PBYTE)pvSource) + dwChunk);
            ++pBuf;
        }
    }
}

DWORD CAsyncSecureSocket::CopyFromClientBuffers(LPWSABUF lpSource, DWORD dwBufferCount, PVOID pvTarget, DWORD dwLen)
{
    for(DWORD dw = 0; dwLen != 0 && dw < dwBufferCount; dw++)
    {
        DWORD dwChunk = lpSource->len;
        if(dwChunk > dwLen)
            dwChunk = dwLen;
        CopyMemory(pvTarget, lpSource->buf, dwChunk);
        dwLen -= dwChunk;
        ++lpSource;
        pvTarget = (PVOID)(((PBYTE)pvTarget) + dwChunk);
    }

    return dwLen;   // Return the number of bytes that couldn't be copied.
}

// Output a secure socket header.
std::ostream& CAsyncSecureSocket::TraceHeader(std::ostream& os)
{
    os << "TID " << std::hex  << GetCurrentThreadId() << " - #SS(" << (DWORD)this << std::dec << "): ";
    return os;
}
