/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _ASYNCVIRTUALUSERCONTROLLER_H_
#define _ASYNCVIRTUALUSERCONTROLLER_H_

#include "SharedPointer.hpp"
#include "AsyncVirtualUser.h"
#include "ThreadPoolController.hpp"
#include "TimerQueue.hpp"
#include "TaskGroup.hpp"
#include <vector>
#include <map>

#ifndef CYRANO_TP_STANDALONE
#   include "TestExecuter.h"
#endif


// ////////////////////////////////////////////////////////////////////////////
//
// Class used to maintain an array of CAsyncVirtualUsers.
// Access to instances of that class should be synchronised by callers.
//
// ////////////////////////////////////////////////////////////////////////////

class CVirtualUsersGroup
{
public:
    CVirtualUsersGroup(CTrace* pTrace = NULL);
    ~CVirtualUsersGroup()   { DropAllUsers(); }

    BOOL CreateUsers(DWORD nbUsers);
    BOOL InitUsers(CIOCP* pIocp, CAsyncVirtualUserController& controller, CTimerQueue* pTimerQueue, DWORD dwParallelism);

    DWORD   StartMoreUsers(SharedPtr<TaskGroupDefinition>& pTaskGroup, DWORD dwNbUsers);
    void    CancelAllUsers();
    void    DropAllUsers();

    void    AppendThreadIdents      (std::string &str)                      const;
    DWORD   GetMatchingThreadsCount (const std::string &sname)              const;
    BOOL    GetVuScriptName         (const char *ident, std::string &name)  const;
    BOOL    GetVuStatus             (const char *ident, EVuExecPhase &phase)const;
    BOOL    GetVuStartTime          (const char *ident, ULONG &ulTime)      const;

private:
    // Helpers.

    // Forbidden operators.
    CVirtualUsersGroup(const CVirtualUsersGroup &VUG);
    CVirtualUsersGroup &operator = (const CVirtualUsersGroup &VUG);

    // Data structures.
    CIOCP* m_pIocp;

    CAsyncVirtualUser *m_VirtualUsersArray;
    DWORD m_dwVirtualUsersCount;
    DWORD m_dwStartedUsersCount;

    CTrace *m_pTrace;
};  // class CVirtualUsersGroup

class CVirtualUsersList
{
public:
    CVirtualUsersList();
    ~CVirtualUsersList();

    DWORD   Init(CIOCP* pIocp, CTimerQueue* pTimerQueue, CAsyncVirtualUserController *pVUController, DWORD dwInitialVUCount, DWORD dwVUCountGrowBy, DWORD dwParallelism, CTrace *pTrace = NULL);

    DWORD   StartUsers(SharedPtr<TaskGroupDefinition>& pTaskGroup, DWORD dwNbUsers);
    void    CancelAllUsers();
    DWORD   DropVirtualUser(DWORD dwVuId);
    void    DeleteGroups();

    void    GetThreadIdents(std::string &str)                   const;
    DWORD   GetMatchingThreadsCount(const std::string &sname)   const;
    DWORD   GetUsersCount()                                     const   { return m_dwVirtualUsersCount; }

    EVuExecPhase    GetVuStatus(const char *ident)      const;
    std::string     GetVuScriptName(const char *ident)  const;
    ULONG           GetVuStartTime(const char *ident)   const;

protected:
    // Helpers.
    BOOL CreateGroup(DWORD dwNbUsers);

    // Data structures.
    CIOCP *         m_pIocp;
    CTimerQueue *   m_pTimerQueue;
    CTrace *        m_pTrace;
    CAsyncVirtualUserController *   m_pVUController;
    DWORD   m_dwParallelism;

    std::vector<CVirtualUsersGroup*>    m_GroupsList;
    mutable CFastLock                   m_userGroupListFl;
    DWORD                               m_dwVirtualUsersCount;

    vector<CVirtualUsersGroup*>::size_type   m_currGroup;

    DWORD   m_dwVUCountGrowBy;

private:
    //// Forbidden operators.
    CVirtualUsersList(const CVirtualUsersList&);
    CVirtualUsersList& operator = (const CVirtualUsersList&);
};  // class CVirtualUsersList


// -------------------------------------------------------------------------------

class CAsyncVirtualUserController : public CThreadPoolController
{
public:
    CAsyncVirtualUserController(TScriptRule* rule, CTrace *pTrace = NULL);
    ~CAsyncVirtualUserController();

    virtual DWORD Init(DWORD dwThreadPoolSize, DWORD dwNumberOfConcurrentThreadsInIocp, DWORD dwInitialVUCount, DWORD dwVUCountGrowBy, DWORD dwBrowserParallelism);

    //// IocpEvent methods.
    virtual void OnEvent      (DWORD /*dwNumberOfBytesTransferred*/, BOOL /*bIsIo*/, DWORD /*dwEventType*/);
    virtual void OnEventFailed(DWORD /*dwNumberOfBytesTransferred*/, BOOL /*bIsIo*/, DWORD /*dwEventType*/, DWORD /*dwError*/);

    ScriptCache& get_cache()        { return m_scriptCache; }
#ifndef CYRANO_TP_STANDALONE
    InjectorControl_ptr GetInjectorControl() const;
#endif
    const std::string& GetExeName() const;
    long  GetTimeOffset() const;
    long  GetCachedTestVUCount() const;
    void  StopController();

#ifndef CYRANO_TP_STANDALONE
    //// Task group stats type definitions.
    struct TaskKey
    {
        TaskKey() : m_lTaskGroupId(0), m_lTaskId(0) {}
        TaskKey(long lTaskGroupId, long lTaskId) : m_lTaskGroupId(lTaskGroupId), m_lTaskId(lTaskId) {}
        TaskKey(const TaskId& tid) : m_lTaskGroupId(tid.lTaskGroupId), m_lTaskId(tid.lTaskId) {}
        bool operator<(const TaskKey& tk) const
        {
            if(m_lTaskGroupId != tk.m_lTaskGroupId)
                return m_lTaskGroupId < tk.m_lTaskGroupId;
            else
                return m_lTaskId < tk.m_lTaskId;
        }

        long m_lTaskGroupId;
        long m_lTaskId;
    };
    struct VUSnapShot                       // VU stats.
    {
        long        m_lVUID;                // Task group-relative VU id
        long        m_lCurTask;             // Task group-relative ID of current task
        long        m_lStartTime;           // VU Start date/time
        long        m_lRunTime;             // Period for which VU has been running (seconds)
        long        m_lCompletedIterations; // Number of Task iterations completed
        std::string m_strNoteText;          // SCL NOTE command text
    };
    typedef std::map<long, SharedPtr<VUSnapShot> > VUSnapShots;    // Maps task group-relative VU IDs to VU snapshots.
    typedef std::map<TaskKey, VUSnapShots> TaskSnapShot; // Maps task IDs to VU snapshots collections.
    struct TaskGroupSnapShot                // Task group stats.
    {
        TaskGroupSnapShot() : m_lCompletedIterations(0),m_lRunTime(0),m_lRequests(0),m_lBytes(0) {}
        long        m_lCompletedIterations; // Grand total of task group iterations.
        long        m_lRunTime;             // In seconds. Only computed in GetStatsSnapShot().
        long        m_lRequests;            // Total since the beginning of the task group.
        long        m_lBytes;               // Total since the beginning of the task group.
        std::vector<long> m_activeUsers;    // Each entry is the number of active users
                                            // currently executing the task, regardless
                                            // of the task group iteration of each user.
                                            // Indexed by the task number.
        VUSnapShots m_VUSnapShots;          // VU snapshots.
    };
    typedef std::map<int,TaskGroupSnapShot> TaskGroupSnapShots; // Maps task group IDs to task group snapshots.
    struct TestExecuterSnapShot             // Test Executer stats.
    {
        TestExecuterSnapShot() : m_lNumberOfActiveVUs(0) {}
        std::string m_strTestExecuter;      // Test Executer object.
        long        m_lNumberOfActiveVUs;   // Only computed in GetStatsSnapShot().
        TaskGroupSnapShots m_taskGroupSnapShots;
    };
#endif

    //// Task Group related methods.
    BOOL    StartTaskGroup(SharedPtr<TaskGroupDefinition>& pTaskGroup);
    void    StopTaskGroup(int iTaskGroupNumber);
#ifndef CYRANO_TP_STANDALONE
    BOOL    GetStatsSnapShot(TestExecuterSnapShot& sn);
    BOOL    GetTaskSnapshot(const TaskIdList& taskIDs, TaskSnap2List& taskSnapshots);
    long    GetTestExecuterVUs() const;
#endif
    TaskGroupDefinition::TaskList::const_iterator GetFirstTask(SharedPtr<TaskGroupDefinition>& pTaskGroup);
    BOOL    GetNextTask(SharedPtr<TaskGroupDefinition>& pTaskGroup, TaskGroupDefinition::TaskList::const_iterator& pCurTask);
    DWORD   IncTaskGroupVUs(SharedPtr<TaskGroupDefinition>& pTaskGroup);
    BOOL    DecTaskGroupVUs(SharedPtr<TaskGroupDefinition>& pTaskGroup);

    //// Task Group related notifications.
    void    OnUserEnterTask(const CAsyncVirtualUser& vu, int iTaskGroupNumber, int iTaskNumber);
    void    OnUserExitTask(const CAsyncVirtualUser& vu, int iTaskGroupNumber, int iTaskNumber, BOOL bExitTaskGroup = FALSE);
#ifndef CYRANO_TP_STANDALONE
    void    OnTaskGroupComplete(int iTaskGroupNumber, TaskGroupStopReason eReason);
#else
    void    OnTaskGroupComplete(int iTaskGroupNumber);
#endif
    void    OnRequestComplete(int iTaskGroupNumber, DWORD dwBytes);

    BOOL    StartNewUsersGroup(SharedPtr<TaskGroupDefinition>& pTaskGroup, DWORD dwNumUsers);
    void    CancelAllUsers();
    void    DropUsers() { m_VirtualUsersList.DeleteGroups(); }
    // TODO: check if should serialize the calls to the users list to avoid maintain integrity. However, seems to be synchronised since CORBA calls are.

    void    GetThreadIdents(std::string &str)                   const   { m_VirtualUsersList.GetThreadIdents(str); }
    DWORD   GetMatchingThreadsCount(const std::string &sname)   const   { return m_VirtualUsersList.GetMatchingThreadsCount(sname); }  // Return the number of threads matching sname.
    DWORD   GetUsersCount()                                     const   { return m_VirtualUsersList.GetUsersCount(); } // Return the whole users count.
    void    OnVirtualUserOver(DWORD dwVUId);

    std::string     GetVuScriptName(const char *ident)  const   { return m_VirtualUsersList.GetVuScriptName(ident); }
    EVuExecPhase    GetVuStatus(const char *ident)      const   { return m_VirtualUsersList.GetVuStatus(ident);     }
    DWORD           GetVuStartTime(const char *ident)   const   { return m_VirtualUsersList.GetVuStartTime(ident);  }

    void    IncFailedUsers()        { InterlockedIncrement((LONG*)&m_dwFailedUsersCount);     }
    void    IncSuccessfullUsers()   { InterlockedIncrement((LONG*)&m_dwSuccessfullUsersCount);}

    DWORD   GetFailedUsersCount()       const   { return m_dwFailedUsersCount; }
    DWORD   GetSuccessfullUsersCount()  const   { return m_dwSuccessfullUsersCount; }

protected:

    // User-defined IOCP events.
    enum { EVENT_TYPE_TASKGROUP_STOPPED  = 0x8010,  // Notify Injector Control that task group stopped.
           EVENT_TYPE_EXECUTER_STOPPING  = 0x8011   // Notify Injector Control that Test Executer is stopping.
         };

    //// Helpers.
    BOOL RampUpUsers(SharedPtr<TaskGroupDefinition>& pTaskGroup);
    BOOL RampUpBatch(SharedPtr<TaskGroupDefinition>& pTaskGroup, DWORD dwNumBatchUsers);

    //// Data.
    typedef std::map<int, SharedPtr<TaskGroupDefinition> > TaskGroups; // Maps task group IDs to task group definitions.
    mutable CFastLock   m_taskGroupsLock;       // Controls access to m_taskGroups.
    TaskGroups          m_taskGroups;           // Collection of running task groups.
#ifndef CYRANO_TP_STANDALONE
    mutable CFastLock   m_taskGroupsStatsLock;  // Controls access to m_snapShotStats and m_VUSnapShots.
    TestExecuterSnapShot m_snapShotStats;       // Real-time task group stats.
    TaskSnapShot        m_TaskSnapShots;        // Real-time task stats.
#endif
    CVirtualUsersList   m_VirtualUsersList;
    CTimerQueue*        m_pTimerQueue;

    ScriptCache         m_scriptCache;

#ifdef USE_PERF_DATA
    static LONG*        m_spActiveUsers;
#endif

    DWORD   m_dwFailedUsersCount;
    DWORD   m_dwSuccessfullUsersCount;

private:
    //// Forbidden operators.
    CAsyncVirtualUserController& operator = (const CAsyncVirtualUserController&);
};

#endif
