/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "stdafx.h"

#include "Statistics.hpp"
#include "Fastlock.hpp"

using namespace std;

// When dwViewSize=0, that value is computed automatically in CMemMappedFile.
BOOL CStatistics::Init(const string &zStatsFileName, const string &zUrlsFileName, DWORD dwViewSize)
{
    if(sm_pWriteSl == NULL)
        sm_pWriteSl = new CFastLock();
    if(sm_pWriteSl == NULL)
        return FALSE;

	sm_tOffset = 0;

    // Query for the performance timer resolution (counts per second) to help compute elapsed time.
    if(QueryPerformanceFrequency(&sm_liPerfCtrFrequency) == 0)
    {
        // Perf. counters not supported by hardware. Will use GetSystemTimeAsFileTime.
    }

#ifndef CYRANO_TP_STANDALONE
    if(     sm_urlsFile.OpenFile(zUrlsFileName, dwViewSize, GENERIC_WRITE, FILE_SHARE_READ) == FALSE
        ||  sm_statsFile.OpenFile(zStatsFileName, dwViewSize, GENERIC_WRITE, FILE_SHARE_READ) == FALSE)
    {
        sm_bErrorDontWrite = TRUE;
        return FALSE;
    }
    else
#endif
    // Won't dump the stats file.
    return TRUE;
}

void CStatistics::Release()
{
    sm_urlsFile.CloseFile();
    sm_statsFile.CloseFile();

    if(sm_pWriteSl != NULL)
    {
        delete sm_pWriteSl;
        sm_pWriteSl = NULL;
    }
}

void CStatistics::DumpData()
{
    if(!sm_bErrorDontWrite)
    {
        CLock lock (sm_pWriteSl);

        // WARNING: assumes that 'sizeof SOutputStats' is smaller than m_dwViewSize.

        // Write statistics.
        if(sm_statsFile.WriteData((LPCSTR)&m_statsData, sizeof SOutputStats) == FALSE)
            sm_bErrorDontWrite = TRUE;

        // Write urls + \0
        if(sm_urlsFile.WriteData(m_szURL.c_str(), m_szURL.size() + 1) == FALSE)
            sm_bErrorDontWrite = TRUE;
    }

    m_statsData.m_dwDataFlow = 0;
}

void CStatistics::OnRequestStart()
{
    m_statsData.m_dwDataFlow = 0;
    time(&m_statsData.m_ttTimeStamp);

	m_statsData.m_ttTimeStamp = m_statsData.m_ttTimeStamp - sm_tOffset;

    if(QueryPerformanceCounter(&m_liRequestStartCounter) == 0)
    {
        // Hardware does not provide a high-resolution performance timer.
        GetSystemTimeAsFileTime(&m_ftRequestSystemTime);
    }
}

void CStatistics::OnRequestOver()
{
    LARGE_INTEGER liNowCounter;

    if(QueryPerformanceCounter(&liNowCounter) != 0)
    {
        LARGE_INTEGER liDuration;
        assert(liNowCounter.QuadPart > m_liRequestStartCounter.QuadPart);
        liDuration.QuadPart = ((liNowCounter.QuadPart - m_liRequestStartCounter.QuadPart) * 1000i64) / sm_liPerfCtrFrequency.QuadPart;
        m_statsData.m_ctElapsedTime = (clock_t) liDuration.QuadPart;
    }
    else
    {
        // Hardware does not provide a high-resolution performance timer.

        FILETIME   ftNow;
        unsigned __int64    qwNow;
        unsigned __int64    qwStart;
        unsigned __int64    qwDuration;

        GetSystemTimeAsFileTime(&ftNow);

        // Compute the time elapsed since request started.

        qwNow = ftNow.dwHighDateTime;
        qwNow <<= 32;
        qwNow |= ftNow.dwLowDateTime;

        qwStart = m_ftRequestSystemTime.dwHighDateTime;
        qwStart <<= 32;
        qwStart |= m_ftRequestSystemTime.dwLowDateTime;

        qwDuration = (qwNow - qwStart) / 10000;

        assert(qwDuration <= INT_MAX);
        m_statsData.m_ctElapsedTime = (clock_t)qwDuration;
    }
}

void CStatistics::SetTimeOffset(long t_offset)
{
	sm_tOffset = t_offset;
}

CMemMappedFile CStatistics::sm_statsFile;
CMemMappedFile CStatistics::sm_urlsFile;

CFastLock *     CStatistics::sm_pWriteSl            = NULL;
LARGE_INTEGER   CStatistics::sm_liPerfCtrFrequency;

#ifndef CYRANO_TP_STANDALONE
BOOL            CStatistics::sm_bErrorDontWrite     = FALSE;
#else
// Don't write the data to the file.
BOOL            CStatistics::sm_bErrorDontWrite     = TRUE;
#endif

long			CStatistics::sm_tOffset = 0;
