/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// This file defines the Task and Task Group classes, which hold the data and
// features required to manage Tasks and Task Groups.

#ifndef __TASK_GROUP_HPP__
#define __TASK_GROUP_HPP__

#include<string>
#include<vector>

class TaskGroupDefinition
{
public:
    enum LogLev { loglev_none, loglev_low, loglev_high };

    TaskGroupDefinition();
    TaskGroupDefinition(int iTaskGroupNumber, long lTaskGroupDuration,
        long lNumVirtualUsers, long lNumTaskGroupIterations,
        int  iInterBatchDelay, long lVUsPerBatch, int iBatchRampUpTime,
        LogLev iLoggingLevel, BOOL bUsePageTimers);
    TaskGroupDefinition(const TaskGroupDefinition& taskGroup);
    ~TaskGroupDefinition()
    {
    }
    TaskGroupDefinition& operator=(const TaskGroupDefinition& taskGroup);

    // Thread-safe method to disable a task group.
    void Disable() {InterlockedExchange((PLONG)&m_bTaskGroupEnabled, 0);}

    //// Task definition.
    class TaskDefinition
    {
    public:
        TaskDefinition();
        TaskDefinition(const std::string& strScriptName, long lDuration, long lIterations, int iFixedDelay);
        TaskDefinition(const std::string& strScriptName, long lDuration, long lIterations, int iMinDelay, int iMaxDelay);
        TaskDefinition(const TaskDefinition& task);
        TaskDefinition& operator=(const TaskDefinition& task);

        //// Data.
        std::string m_strScriptName;    // The name (and path) of the TOF file associated with the task.
        long        m_lDuration;        // Number of seconds that the task should last
                                        // for each virtual user. If > 0, the VU should
                                        // loop on the task until the duration expires.
                                        // A VU should always perform an integer number
                                        // of iterations. If the value is 0, then there
                                        // is no duration, and the execution of the task
                                        // should stop when m_lIterations have been executed.
        long        m_lIterations;      // Number of iterations to perform.
                                        // Ignored when m_lDuration is > 0.
                                        //
        BOOL        m_bFixedDelay;      // TRUE if the delay between iterations is fixed (discriminator).
        union
        {
            int     m_iFixedDelay;      // In seconds. Value for fixed delays.
            struct
            {
                int m_iMinDelay;        // In seconds. Value of minimum delay for variable delays.
                int m_iMaxDelay;        // In seconds. Value of maximum delay for variable delays.
            }       m_variableDelay;
        }           m_delay;
    };

    typedef std::vector<TaskDefinition> TaskList;

    //// Data.
    int     m_iTaskGroupNumber;         // Task Group identifier.
    BOOL    m_bTaskGroupEnabled;        // If FALSE, the task group execution must stop as
                                        // soon as possible.
    long    m_lTaskGroupDuration;       // Number of seconds the task group should last.
                                        // As soon as this duration expires, all virtual
                                        // users should complete the task group and exit.
                                        // Thus, all VUs should always perform an integer
                                        // number of task group iterations.
                                        // If the value is 0, then there is no duration,
                                        // and the execution of each VU should stop when 
                                        // m_lNumTaskGroupIterations have been executed.
    unsigned __int64 m_qwEndTime;       // The timestamp of when the task group should end.
                                        // This is computed using m_lTaskGroupDuration.
    long    m_lNumVirtualUsers;         // Number of virtual users to run the task group.
    long    m_lNumTaskGroupIterations;  // Number of task group iterations, i.e., the
                                        // number of times to run the task group during
                                        // its execution. Ingored when m_lTaskGroupDuration
                                        // is > 0.
    int     m_iInterBatchDelay;         // Delay (in seconds) between 2 consecutive batches
                                        // of VUs.
    long    m_lVUsPerBatch;             // Number of virtual users to launch in each batch.
    int     m_iBatchRampUpTime;         // Time (in seconds) given to ramp-up a batch of VUs.
    LogLev  m_iLoggingLevel;            // Number of virtual users for which the statistics
                                        // are to be logged.
    BOOL    m_bUsePageTimers;           // 'Use Page Timers' flag.
    TaskList m_tasks;                   // List of tasks in the task group.

    //// Statistics. (Not snapshots requests!)
    long    m_lVirtualUserID;           // Sequence counter used to generate the external VU IDs.
    long    m_lCurVirtualUsers;         // Number of virtual users currently running the task group.
    long    m_lDoneVirtualUsers;        // Number of virtual users done executing the task group.
};

#endif  // __TASK_GROUP_HPP__
