/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifdef  WIN32
#  pragma warning(disable : 4710)
#endif  // WIN32

#include <ctime.hpp>
#include "gwhttp.hpp"
#include "cstof.hpp"
#include <set>
using std::set ;
#include <Wincrypt.h>

CTofReference               tofReference;
CTargetEvaluation *         gTargetEvaluation = NULL ;
CTMutex                     CGWHttp::myMutex_;
string                      gListeDocuments = "";
string                      gDocumentPath = "";
FILE *                      gFdBody = NULL ;
FILE *                      gFdClientHeader = NULL ;
FILE *                      gFdServerHeader = NULL ;
FILE *                      gFdIndex = NULL ;
FILE *                      gFdAll = NULL ;
int                         giSpid = 0;
int                         giNbCon = 0 ;
map<string, string>			gClientCertificateMap ;

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////// Class CGWHttp //
///////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// CGWHttp::CGWHttp
//
// Constructor.
//
// Arguments : pConnectionId : connection identifier.
//             taskName      : name of the task.
//             bTraceOn      : reference on the trace mode.
//             logFileList   : reference on the log file list.
//             pClient       : Client socket (web browser)
//             pTarget       : Target socket (web server)
// Return    : none.
//-----------------------------------------------------------------------------
CGWHttp::CGWHttp (const void* pConnectionId, const string& taskName, bool& bTraceOn, CTOLogFileList& logFileList, CTSocket*& pClient, CTSocket*& pTarget)
        : CGWParse (pConnectionId, taskName, bTraceOn, logFileList, pClient, pTarget), pTarget_(pTarget), pClient_(pClient), pContext_(NULL)
{
    CTLockGuard <CTMutex> guard(myMutex_);

    ClientPreviousPacket_ = NULL ;
    lClientPreviousPacketLen_ = 0 ;
    ServerPreviousPacket_ = NULL ;
    lServerPreviousPacketLen_ = 0 ;
    lEndDocumentLen_ = 0 ;
    savedDocument_ = NULL;
    lSavedDocumentLen_ = 0 ;
    bServerSendHeader_ = FALSE ;
    clientHeader_ = "" ;
    serverHeader_ = "" ;
    iServerHeaderSize_ = 0;
    previousHost_ = "" ;
    iPreviousPort_ = 0 ;
    previousProtocolType_ = "" ;

    bFullServerDocument_ = TRUE ;

    memset(endDocument_, 0, sizeof(endDocument_) ) ;
    iIndexSslURL_ = string::npos;

    giNbCon++ ;

    ++giSpid ;
    iSpid_ = giSpid ;

    if (pContext_ == NULL)
        pContext_ = new CContext ();

    pContext_->setConnectId(iSpid_);

    tofReference.addContext(pContext_);
    tofReference.associateTofFile(pContext_);

    iServerIndexJavaScript_ = string::npos;
} // CGWHttp::CGWHttp

//-----------------------------------------------------------------------------
// CGWHttp::~CGWHttp
//
// Desstructor.
//-----------------------------------------------------------------------------
CGWHttp::~CGWHttp ()
{
    CTLockGuard <CTMutex>  guard(myMutex_);
    if (savedDocument_ != NULL)
    {
        saveDocumentAsFile();
        if (!savedDocument_)
            delete [] savedDocument_ ;
        savedDocument_ = NULL ;
        lSavedDocumentLen_ = 0 ;
    }

    giNbCon-- ;
    if (!tofReference.isOutOfOrder())
        tofReference.writeDisconnectInTofFile(pContext_, bFullServerDocument_);
    tofReference.removeContext(pContext_);
}


//-----------------------------------------------------------------------------
// CGWHttp::find_in_string
//
// Find text in string, return index if found or -1 if not.
//
// Arguments : lookin   : input string
//             lookfor  : string to search
//             matchcase: true to match case
//             startpos : start position
// Return    : lookfor position in lookin (or -1)
//-----------------------------------------------------------------------------
int CGWHttp::find_in_string(string lookin, string lookfor, bool matchcase, unsigned startpos)
{

    // Ensure not looking beyond end
    if(startpos >= lookin.size())
        return -1;
    // If case insensitve, convert to uppercase
    if(!matchcase)
    {
        // Temporary Hacking strings - needed to resolve possible STL bug
        // related to "copy on write" feature of strings (MSC 4.2).
        // For more information on this please refer to mdbvar.cxx

        // They prevent corruption to the original string
        string tmplookin = lookin.c_str();
        string tmplookfor = lookfor.c_str();
        int k;
        for (k=0; k<(int)lookin.size(); k++)
        {
            tmplookin[k] = toupper(lookin[k]);
        }
        for (k=0; k<(int)lookfor.size(); k++)
        {
            tmplookfor[k] = toupper(lookfor[k]);
        }


        // Do search
        string::size_type p = tmplookin.find(tmplookfor, startpos);
        if(p == tmplookin.npos)
            return -1;
        return (int)p;
    }
    else
    {
        // Do search
        string::size_type p = lookin.find(lookfor, startpos);
        if(p == lookin.npos)
            return -1;
        return (int)p;
    }
}
//=========================================================< Control methodes >
//-----------------------------------------------------------------------------
// CGWHttp::parseBuffer
//
// Parse the buffer.
//
// Arguments : iDataFrom   : data from client or target.
//             bUrgent     : true if packet received in urgent mode.
//             pBuf        : data buffer.
//             lBufLen     : reference to data buffer length (can be modified).
// Return    : lParsedData : number of data parsed.
//-----------------------------------------------------------------------------
long CGWHttp::parseBuffer (const int iDataFrom, bool& bUrgent, const void* pBuf, long& lBufLen, const long lOffset, const long lMaxLen, long& lMoreData)
{
    try
    {
		{
			CTLockGuard <CTMutex> guard(myMutex_);
			char buf[50] ;
			sprintf(buf, "%d,%d", iSpid_, lBufLen);
			traceLine (string ( (char*) ( (iDataFrom == FROM_TARGET) ? "nsprecv" : "nspsend") ) + string(":packet dump\t//spid ") + string(buf)) ;
			CTMemView memoryView ( (char*) pBuf, lBufLen, string ( (iDataFrom == FROM_TARGET) ? "nsprecv:" : "nspsend:"), 8, " |", '.', "|") ;
			traceFlush (memoryView.getAscii () ) ;
		}
        // If one of the previous requests was direct https://, the gateway is
        // set in "out of order" state until the next capture.
        if (tofReference.isOutOfOrder())
        {
            throw CTException(string("Out of order because has received an unsupported protocol packet"), __FILE__, __LINE__);
        }

        // ------------------------------------------------------------------
        //
        // Client packets
        //
        // ------------------------------------------------------------------
        if (iDataFrom == FROM_CLIENT)
        {
            //
            // reinitialize Server members
            //
            iServerHeaderSize_ = string::npos ;
            iServerIndexTextHtml_ = string::npos ;
            iServerContentLength_ = 0 ;
            iServerIndexSetCookie_ = string::npos ;
            lEndDocumentLen_ = 0 ;
            iDataFrom_ = iDataFrom ;
            bServerSendHeader_ = FALSE ;
            bFullServerDocument_ = TRUE ;
            // if a previous packet was saved (because it did not contain the end
            // of message), we concatenate it
            char *packet = NULL ;
            packet = new char[lClientPreviousPacketLen_ + lBufLen + 1] ;
            memset(packet, 0, lClientPreviousPacketLen_ + lBufLen + 1);
            if (lClientPreviousPacketLen_ > 0)
                memcpy(packet, ClientPreviousPacket_, lClientPreviousPacketLen_ );
            else
                // if it is a new request from client, set body size to 0
                iClientBodySize_ = 0;
            memcpy(&packet[lClientPreviousPacketLen_], pBuf, lBufLen);

            int iContentLength = 0 ;
            // retreive the eventual body length
            string strPacket(packet);
            int indexContentLength = 0;
            if ((indexContentLength = CGWHttp::find_in_string(strPacket,"Content-Length:", false, 0)) != string::npos)
            {
                iContentLength = atoi(&packet[indexContentLength] + strlen("Content-Length:")) ;
            }

            //
            // Check to see if it is a SSL request
            //
            int indexEndFirstHeader = CGWHttp::find_in_string(strPacket,"\r\n", false,0) ;

            string strFirstHeader = strPacket.substr(0, indexEndFirstHeader) ;
            iIndexSslURL_ = CGWHttp::find_in_string(strFirstHeader,CYRANO_SSL_URL, false,0);

            char * zBuf = new char [strFirstHeader.length() + 1];
            strcpy(zBuf, strFirstHeader.c_str()) ;
            pContext_->setMethod(strtok(zBuf, " \t\r\n")) ;

            // We receive a CONNECT that means that the application do https://
            // requests directly (instead of http://{)(ex : java applet)
            // We close the thread and set a global flag to make the gateway
            // out of order.
            // This flag will be reinitialized in the next capture (new record).
            if (pContext_->getMethod() == "CONNECT")
            {
                tofReference.setOutOfOrder(true) ;
                tofReference.writeCommentInTofFile(pContext_, "[ERROR]: Unsupported protocol packet received : cannot continue");
                throw CTException(string("ERROR : unsupported protocol packet received"), __FILE__, __LINE__);
            }

            pContext_->setRequestUrl(strtok(NULL, " \t\r\n")) ;
            pContext_->setHttpVersion(strtok(NULL, " \t\r\n")) ;
			
			// Set the client certificate name in the context if the 
			// certificate has already been selected (it is in the map)
			pContext_->setClientCertificateName("");
			if (iIndexSslURL_ != string::npos)
			{
		        // extract the host:port part of hte url
		        string strConnectAndUrl = pContext_->getRequestUrl().substr(strlen("https://"), pContext_->getRequestUrl().length() - strlen("https://")) ;
		        int iIndex = CGWHttp::find_in_string(strConnectAndUrl,"/",false,0) ;
				string strConnect = strConnectAndUrl.substr(0, iIndex );

				if (gClientCertificateMap.find(strConnect) != gClientCertificateMap.end())
					pContext_->setClientCertificateName(gClientCertificateMap[strConnect]);
			}

            delete [] zBuf ;

            // Try to find the end of the header
            char * pHeaderEnd = strstr(packet, "\r\n\r\n") + 4;
            int iHeaderSize = 0;
            if (pHeaderEnd != NULL)
            {
                iHeaderSize = (unsigned int)pHeaderEnd - (unsigned int)packet ;
                iClientBodySize_ = lClientPreviousPacketLen_ + lBufLen - iHeaderSize ;
            }

            // if everything is in the packet(header + body), we can write
            // in the tof
            // for POST requests (detected by iContentLength > 0), if only receive
            // the header, seems to be the case 100 continue
            if (iHeaderSize > 0 &&
                    iHeaderSize + iContentLength <= lBufLen + lClientPreviousPacketLen_
               )
            {
                string host ;
                int iPort ;
                bool bProxy = FALSE ;
                if (!pContext_->is100continue())
                    targetEvaluate(host, iPort, bProxy) ;

                // If the gateway send directly requests to the web server
                // (no proxy server), replace proxy headers by
                // direct headers (ex : proxy-connection: by Connection:)
                if (!bProxy)
                    replaceProxyClientHeaders(packet, iHeaderSize, iClientBodySize_) ;

                if (!pContext_->is100continue())
                {
                    // --------------------------
                    // Add cookies (it should be there only for Netscape)
                    // --------------------------
                    addCookies(packet, iHeaderSize);
                    // --------------------------
                }

                string newHeader = replaceCyranoSslFromHeader(string(packet, iHeaderSize - 4)) ;
                memcpy(packet, newHeader.c_str(), newHeader.length());

                // Save the client header to be able to save it in a file
                clientHeader_ = newHeader.substr(indexEndFirstHeader + 2, newHeader.length() - indexEndFirstHeader - 2 );

				getClientHeaderInformations(clientHeader_); 

                if (CGWHttp::find_in_string(pContext_->getMethod(),"GET",false,0) == 0 ||
                        CGWHttp::find_in_string(pContext_->getMethod(),"POST",false,0) == 0)
                    tofReference.writeRequestInTofFile(pContext_, clientHeader_, (iContentLength>0)?packet + iHeaderSize:NULL, iContentLength);


                // remove cache headers before sending to the target
                // Remove host from url already done in 100 continue case
                if (!pContext_->is100continue())
                {
                    removeClientHeaders(packet, iHeaderSize, iClientBodySize_);
                }

                if (iIndexSslURL_ != string::npos)
                {
                    sendSSL(string(packet), packet + iHeaderSize, iContentLength>0?iContentLength:iClientBodySize_) ;
                    lBufLen = 0 ;
                    iIndexSslURL_ = string::npos ;
                }
                else
                {
                    // Remove host from url already done in 100 continue case
                    if (!bProxy && !pContext_->is100continue())
                        removeHostFromUrl(packet, iHeaderSize, iClientBodySize_);
                    // For 100 continue, only send the body
                    // The header has already been sent
                    traceLine (string ("After")) ;
                    CTMemView memoryView ( (char*) packet, iHeaderSize + iClientBodySize_, string ( (iDataFrom == FROM_TARGET) ? "nsprecv:" : "nspsend:"), 8, " |", '.', "|") ;
                    traceFlush (memoryView.getAscii () ) ;
                    if (pContext_->is100continue())
                    {
                        sendNormal(pContext_->getServerName(), pContext_->getPort(), &packet[iHeaderSize], iContentLength) ;
                        pContext_->set100continue(FALSE) ;
                    }
                    else
                        sendNormal(pContext_->getServerName(), pContext_->getPort(), packet, iHeaderSize + iContentLength) ;
                    lBufLen = 0 ;
                }
                if (ClientPreviousPacket_)
                    delete [] ClientPreviousPacket_ ;
                ClientPreviousPacket_ = NULL;
                lClientPreviousPacketLen_ = 0 ;
                delete [] packet ;
            }
            // The packet does not contain the end of the message
            // we save the packet to wait for the next one.
            else
            {
                // 100 continue case
                if (iHeaderSize > 0 && iContentLength > 0 && lBufLen + lClientPreviousPacketLen_ == iHeaderSize)
                {
                    string host ;
                    int iPort ;
                    bool bProxy = FALSE ;
                    targetEvaluate(host, iPort, bProxy) ;
                    pContext_->set100continue(TRUE) ;
                    // remove cache headers before sending to the target
                    // remove proxy headers is connected directly to a web server.
                    removeClientHeaders(packet, iHeaderSize, 0);
                    // --------------------------
                    // Add cookies (it should be there only for Netscape)
                    // --------------------------
                    addCookies(packet, iHeaderSize);
                    // --------------------------

                    if (iIndexSslURL_ != string::npos)
                    {
                        sendSSL(string(packet), NULL, 0) ;
                        iIndexSslURL_ = string::npos ;
                    }
                    else
                    {
                        if (!bProxy)
                            removeHostFromUrl(packet, iHeaderSize, 0);
                        sendNormal(host, iPort, packet, iHeaderSize) ;
                    }
                }
                else if (pContext_->is100continue())
                {
                    iClientBodySize_ += lBufLen ;
                }

                if (ClientPreviousPacket_)
                    delete [] ClientPreviousPacket_ ;
                ClientPreviousPacket_ = packet ;
                lClientPreviousPacketLen_ = iHeaderSize + iClientBodySize_ ;
                lBufLen = 0 ;
            }
        }
        // ------------------------------------------------------------------
        //
        // Server packets
        //
        // ------------------------------------------------------------------
        else
        {
            iIndexSslURL_ = string::npos ;
            // We have decided to let the gateway send the rest of the packets
            // (previous packet was also a server packet and is not HTML)
            if (iDataFrom_ == iDataFrom
                    && iServerIndexTextHtml_ == string::npos
                    && iServerHeaderSize_ > 0
                    && iServerIndexJavaScript_ == string::npos
                    && statusCode_ != "100")
            {
                // if savedDocument==NULL, it means we have saved documents
                // (by doing stop capture) during a receive.
                if (savedDocument_ != NULL)
                    saveDocumentInMemory((char *)pBuf, lBufLen) ;
                return lBufLen;
            }

            // check that we do not receive something differnet from HTTP
            // ( in the first response packet)
            // In this case, the gateway transfers direcly
            if (iDataFrom_ != iDataFrom)
            {
                if (strnicmp((char *)pBuf, "HTTP", lBufLen >= 4?4:lBufLen) != 0)
                    return lBufLen ;
            }
            iDataFrom_ = iDataFrom ;
            // if a previous packet was saved (because it did not contain the end
            // of message), we concatenate it
            char *packet = NULL ;
            packet = new char[lServerPreviousPacketLen_ + lBufLen + 1] ;
            memset(packet, 0, lServerPreviousPacketLen_ + lBufLen + 1);
            if (lServerPreviousPacketLen_ > 0)
                memcpy(packet, ServerPreviousPacket_, lServerPreviousPacketLen_ );
            memcpy(&packet[lServerPreviousPacketLen_], pBuf, lBufLen);

            string strPacket ;

            // Try to find the end of the header
            // If we have all the header, we can extract
            // . http version, status code
            // . content-length header
            // . content-type header (specific action fot "text/html")
            string strBodyPacket = "";
            if (iServerHeaderSize_  == string::npos)
            {
                bool bHeaderFinishedByOneCRLF = false ;
                bool bHeaderFinishedByLFLF = false ;
                strPacket = packet ;
                iServerHeaderSize_=CGWHttp::find_in_string(strPacket,"\r\n\r\n",false,0);

                // sometimes, HTTP headers are finished by LFLF instead of CRLFCRLF
                if (iServerHeaderSize_ == string::npos)
                {
                    iServerHeaderSize_=CGWHttp::find_in_string(strPacket,"\n\n",false,0);
					if (iServerHeaderSize_ != string::npos)
	                    bHeaderFinishedByLFLF = true ;
                }

                // Some site send http server 302 headers with only one CRLF at the end
                // This case has been seen with www.hotmail.com
                if (iServerHeaderSize_ == string::npos)
                {
                    bHeaderFinishedByOneCRLF = (strPacket.substr(strPacket.size() - 2, 2) == "\r\n") ;
					if (bHeaderFinishedByOneCRLF)
						iServerHeaderSize_ = strPacket.size() - 2 ;
                }
                if (iServerHeaderSize_ != string::npos)
                {
                    // keep the header to be able to save it in a file
                    int indexFirstLine = CGWHttp::find_in_string(strPacket,"\r\n",false,0) + 2 ;
                    serverHeader_ = strPacket.substr(indexFirstLine, iServerHeaderSize_ - indexFirstLine) ;
                    getServerHeaderInformations(strPacket) ;
                    iServerHeaderSize_ += strlen("\r\n\r\n") ;

                    // One CRLF case (www.hotmail.com case)
                    // We have to check that this header is really complete, i.e that
                    // the status is 302 and the "Content-Length" header is equal to 0
                    // and there is a "Location:" header.
                    if (bHeaderFinishedByOneCRLF)
                    {
                        // if status 302 and Content-Length==0 and there is a Location header
                        // We consider that all the header is received
                        if (!(string(statusCode_) == "302" && iServerContentLength_ == 0 && iServerIndexLocation_ != string::npos ))
                        {
                            strBodyPacket = (char *)pBuf;
                            iServerHeaderSize_ = string::npos ;
                            iServerIndexSetCookie_ = string::npos ;
                        }
                        else
                        {
                            // We add a CRLF at the end of the header to make the classical
                            // algorithm work.
                            char *packetTmp = new char[lServerPreviousPacketLen_ + lBufLen + 3] ;
                            memcpy(packetTmp, packet, lServerPreviousPacketLen_ + lBufLen - 2) ;
                            packetTmp[lServerPreviousPacketLen_ + lBufLen - 2] = 0;
                            strcat(packetTmp, "\r\n\r\n");
                            delete [] packet ;
                            packet = packetTmp ;
                            lBufLen += 2 ;
                        }
                    }
                    // header finished bt LFLF instead of CRLFCRLF
                    else if (bHeaderFinishedByLFLF)
                    {
                        // replace LFLF by CRLFCRLF at the end of the header to make the classical
                        // algorithm work.
                        char *packetTmp = new char[lServerPreviousPacketLen_ + lBufLen + 3] ;
                        memset(packetTmp, 0, lServerPreviousPacketLen_ + lBufLen + 3) ;
                        memcpy(packetTmp, packet, iServerHeaderSize_ - 4) ;
                        strcat(packetTmp, "\r\n\r\n");
                        memcpy(&packetTmp[iServerHeaderSize_], &packet[iServerHeaderSize_ - 2], lServerPreviousPacketLen_ + lBufLen - iServerHeaderSize_ + 2) ;
                        delete [] packet ;
                        packet = packetTmp ;
                        lBufLen += 2 ;
                        strPacket = packet ;
                        serverHeader_ = strPacket.substr(indexFirstLine, iServerHeaderSize_ - indexFirstLine) ;
                    }

                    if (iServerIndexSetCookie_ != string::npos)
                    {
                        writeLoadResponseHeader(strPacket.substr(0, iServerHeaderSize_)) ;
                    }
                    // 100 Continue case
                    // Sometimes, the next reply is content in the same TCP packet
                    if (string(statusCode_) == "100")
                    {
                        pClient_->sendBuffer ((char *)packet, iServerHeaderSize_, "", false);
                        lBufLen -= iServerHeaderSize_ ;
                        if (lBufLen > 0)
                        {
                            int iPos = iServerHeaderSize_ ;
                            iServerHeaderSize_ = string::npos ;
                            char *pCharBuf = (char *)pBuf ;
							pCharBuf += iPos ;

                            int iLen = parseBuffer(iDataFrom, bUrgent, pCharBuf, lBufLen, lOffset, lMaxLen, lMoreData) ;
							if (iLen > 0)
								pClient_->sendBuffer ((char *)pCharBuf, iLen, "", false);
							lBufLen = 0 ;
							return 0 ;
                        }
                        iServerHeaderSize_ = string::npos ;
                        return 0 ;
                    }
                }
                // We only put the body in strPacket
                strBodyPacket = packet + iServerHeaderSize_ ;
            }
            else
            {
                // Only usefull to try to find the end of an HTML document
                strBodyPacket = (char *)pBuf;
            }


            // If the packet is not a HTML packet, we can send the packet
            // back to the client because we are sure there are no element to modify in
            // the body part, else we only send the header
            bool bJustSendHeader = FALSE ;
            if (iServerHeaderSize_ > 0)
            {
                // remove cache headers before sending to the client
                int lSaveHeaderSize = iServerHeaderSize_ ;
                if (!bServerSendHeader_ )
                    parseHTTPandRemoveSSL(packet, iServerHeaderSize_) ;
                removeServerCacheHeaders(packet, iServerHeaderSize_, lBufLen + lServerPreviousPacketLen_ - iServerHeaderSize_);
                lBufLen -= (lSaveHeaderSize - iServerHeaderSize_) ;


                if (!bServerSendHeader_  && ( (iServerIndexTextHtml_ != string::npos ) || (iServerIndexJavaScript_ != string::npos ) ) )
                {
                    pClient_->sendBuffer ((char *)packet, iServerHeaderSize_, "", false);
                    packet = &packet[iServerHeaderSize_] ;
                    bServerSendHeader_ = TRUE ;
                    bJustSendHeader = TRUE ;
                }
                else
                    if ( (iServerIndexTextHtml_ == string::npos) && (iServerIndexJavaScript_ == string::npos) )// || statusCode_[0] == '3')
                    {
                        pClient_->sendBuffer ((char *)packet, lBufLen + lServerPreviousPacketLen_, "", FALSE);
                        saveDocumentInMemory(&packet[iServerHeaderSize_], lBufLen + lServerPreviousPacketLen_ - iServerHeaderSize_) ;
                        lBufLen = 0 ;
                        if (ServerPreviousPacket_)
                            delete [] ServerPreviousPacket_ ;
                        ServerPreviousPacket_ = NULL;
                        lServerPreviousPacketLen_ = 0 ;
                        delete [] packet ;
                        return 0;
                    }
            }

            if (iServerHeaderSize_ > 0 && ( (iServerIndexTextHtml_ > 0) || (iServerIndexJavaScript_ > 0) ) )
            {
                //
                // parse and replace https URL to http URL
                //
                bool bModified = FALSE ;
                string newHtml = parseHTMLandRemoveSSL(packet, bModified) ;

                if (bJustSendHeader)
                    lBufLen -= iServerHeaderSize_ ;

                saveDocumentInMemory(packet, lBufLen + lServerPreviousPacketLen_) ;

                //
                // Modify pBuf to allow the gateway to send the good packet.
                //
                if (bModified)
                    // have modified https:// links
                    memcpy((void *)pBuf, (char *)newHtml.c_str(), lBufLen + lServerPreviousPacketLen_) ;
                else
                {
                    // no https links
                    if (bJustSendHeader)
                        memcpy((void *)pBuf, (char *)packet, lBufLen + lServerPreviousPacketLen_) ;
                }

                memset(endDocument_, 0, sizeof(endDocument_)) ;
                memcpy(endDocument_, &packet[lBufLen + lServerPreviousPacketLen_ - sizeof(endDocument_) + 1], sizeof(endDocument_) - 1) ;
                lEndDocumentLen_ = sizeof(endDocument_) - 1 ;
            }
            // The packet does not contain the end of the header
            // we save the packet to wait for the next one.
            else
            {
                if (ServerPreviousPacket_)
                    delete [] ServerPreviousPacket_ ;
                ServerPreviousPacket_ = packet ;
                lServerPreviousPacketLen_ += lBufLen ;
                lBufLen = 0 ;
            }
        }
    }
    catch (CTException& e)
    {
        throw CGWParseException ( (CGWParseException&) e) ;
    }

    return lBufLen ;
} // CGWHttp::parseBuffer


//-----------------------------------------------------------------------------
// CGWHttp::writeLoadResponseHeader
//
// This functions write load response header in the tof file
//
//
// Arguments : strHeader : buffer that contains the HTTP header
// Return    : None
//-----------------------------------------------------------------------------
void CGWHttp::writeLoadResponseHeader(string strHeader)
{
    int iLength = strHeader.length() ;
    int iIndexSetCookie = CGWHttp::find_in_string(strHeader,"Set-Cookie:",false, 0) ;
    while (iIndexSetCookie != string::npos)
    {
        string cookieHeader = strHeader.substr(iIndexSetCookie, iLength - iIndexSetCookie);
        int iEndLine = CGWHttp::find_in_string(cookieHeader,"\r\n",false,0);
        cookieHeader = cookieHeader.substr(strlen("Set-Cookie:"), iEndLine - strlen("Set-Cookie:"));
        // remove spaces
        int i = 0 ;
        while (cookieHeader[(size_t)i] == ' ')
            i++;
        cookieHeader = cookieHeader.substr(i, cookieHeader.length() - i);


        string fakedUrl = pContext_->getRequestUrl();
		InternetSetCookie(fakedUrl.c_str(), NULL, cookieHeader.c_str());
		int iIndexColonSlashSlash = fakedUrl.find("://") ;
		if (iIndexColonSlashSlash != string::npos)
		{
			fakedUrl.insert(iIndexColonSlashSlash + 3, CYRANO_SSL_STR) ;
			InternetSetCookie(fakedUrl.c_str(), NULL, cookieHeader.c_str());
		}

		tofReference.writeLoadResponseHeader(pContext_, cookieHeader, string("cookie"));

        strHeader = strHeader.substr(iIndexSetCookie + iEndLine + 2, iLength - iIndexSetCookie - iEndLine - 2);
        iLength = strHeader.length() ;
        iIndexSetCookie = CGWHttp::find_in_string(strHeader,"Set-Cookie:",false, 0) ;
    }
}

//-----------------------------------------------------------------------------
// CGWHttp::saveDocumentInMemory
//
// This functions save the document contained in packet in memory
//
//
// Arguments : packet : buffer that contains the request
//             lBufLen : buffer length
// Return    : None
//-----------------------------------------------------------------------------
void CGWHttp::saveDocumentInMemory(char *packet, long lBufLen, bool bDelete)
{
    CTLockGuard <CTMutex> guard(myMutex_);
    //
    // Save the document in memory
    // to put it in a file at the end
    //
    char * docTemp = new char[lSavedDocumentLen_ + lBufLen + 1] ;
    memset(docTemp, 0, lSavedDocumentLen_ + lBufLen + 1) ;
    memcpy(&docTemp[lSavedDocumentLen_], packet, lBufLen) ;

    if (lSavedDocumentLen_ > 0)
    {
        memcpy(docTemp, savedDocument_, lSavedDocumentLen_) ;
        delete [] savedDocument_ ;
    }
    savedDocument_ = docTemp ;
    lSavedDocumentLen_ += lBufLen ;

    if (iServerContentLength_ > 0 && iServerContentLength_ == lSavedDocumentLen_)
    {
	    if (tofReference.getSelectStatus() != ESL_STOP)
	        saveDocumentAsFile() ;
        if (bDelete)
        {
            if (!savedDocument_)
                delete [] savedDocument_ ;
            savedDocument_ = NULL ;
            lSavedDocumentLen_ = 0 ;
		}
    }
    else if (iServerIndexChunked_ != string::npos && isLastChunk())
    {
        if (bDelete)
        {
		    if (tofReference.getSelectStatus() != ESL_STOP)
				saveDocumentAsFile() ;
            if (!savedDocument_)
                delete [] savedDocument_ ;
            savedDocument_ = NULL ;
            lSavedDocumentLen_ = 0 ;
        }
    }
}

//-----------------------------------------------------------------------------
// CGWHttp::targetEvaluate
//
// This functions evalutae if the Request must betransferd through
// a proxy or directly to the web server
//
// Arguments : packet : buffer that contains the request
//             iHeaderSize : Size of the header
//             strConnect (out): contains the server to connect on
//             iPort (out) : port number to connect on
//             bPorxy (out) : proxy connection ?
// Return    : None
//-----------------------------------------------------------------------------
void CGWHttp::targetEvaluate(string & strConnect, int & iPort, bool & bProxy)
{
    int iIndexColon = CGWHttp::find_in_string(pContext_->getRequestUrl(),":",false, 0);
    string strProtocolType = pContext_->getRequestUrl().substr(0, iIndexColon) ;
    string strConnectAndUrl = pContext_->getRequestUrl().substr(strProtocolType.length() + 3, pContext_->getRequestUrl().length() - strProtocolType.length() - 3) ;

    // get the host (or host:port) from the URL
    int iIndex = CGWHttp::find_in_string(strConnectAndUrl,"/",false,0) ;
    string strConnectTmp = strConnectAndUrl.substr(0, iIndex );

    if (strConnectTmp[(size_t)0] == '{')
    {
        strProtocolType += "s" ;
        strConnectTmp = strConnectTmp.substr(1, strConnectTmp.size() - 1) ;
    }

    gTargetEvaluation->evaluate(strConnectTmp, strProtocolType, strConnect, iPort, bProxy) ;
    pContext_->setServerName(strConnect) ;
    pContext_->setPort(iPort) ;
    pContext_->setThroughProxy(bProxy) ;
    // The gateway is a proxy so, the browser can keep a same connection to
    // go to different URLS. If the gateway connect directly to the web server (with
    // no proxy), it must simulate (in the tof file) the creation of a new connection
    // (by incrementing the conid)
    if (previousHost_ == "" && iPreviousPort_ == 0)
    {
        if (bProxy)
            tofReference.writeConnectInTofFile(pContext_);
    }
    else if (previousHost_ != "" && (previousHost_ != strConnect || iPreviousPort_ != iPort || previousProtocolType_ != strProtocolType))
    {
        CTLockGuard <CTMutex> guard(myMutex_);
        tofReference.writeDisconnectInTofFile(pContext_, bFullServerDocument_);
        ++giSpid ;
        iSpid_ = giSpid ;
        pContext_->setConnectId(iSpid_);
        if (bProxy)
            tofReference.writeConnectInTofFile(pContext_);
    }

}


//-----------------------------------------------------------------------------
// CGWHttp::replaceCyranoSslFromHeader
//
// This functions replace all http://{ (our representation of SSL urls)
// present in the header by https://
//
// Arguments : header (in/out): HTTP header
// Return    : None
//-----------------------------------------------------------------------------
string CGWHttp::replaceCyranoSslFromHeader(string & header)
{
    // replace "http://{"  by https in the header
    int iIndex = 0 ;
    int iNewIndex = string::npos ;
    string newHeader = "";
    while ((iNewIndex = CGWHttp::find_in_string(header,CYRANO_SSL_URL,false,iIndex)) != string::npos)
    {
        newHeader += header.substr(iIndex, iNewIndex - iIndex) ;
        newHeader += "https://" ;
        // We must calculate the length to add to content length header
        iIndex = iNewIndex + strlen(CYRANO_SSL_URL);
    }
    newHeader += header.substr(iIndex, header.length() - iIndex) ;

    // replace the '{' in the "host:" header by ' '
    if  ((iIndex = CGWHttp::find_in_string(newHeader,"Host: {", false,0)) != string::npos)
    {
        newHeader[iIndex + strlen("Host: {") - 1] = ' ' ;
    }

    return newHeader ;
}


//-----------------------------------------------------------------------------
// CGWHttp::removeHostFromUrl
//
// For direct connections (in a normal case the browser do not connect through a proxy),
// This functions remove the protocol and hostname from the url.
//
// Arguments : header (in/out): HTTP header
// Return    : None
//-----------------------------------------------------------------------------
void CGWHttp::removeHostFromUrl(char * packet, int & iHeaderSize, int iBodySize)
{
    string strPacket = packet ;
    int indexEndFirstHeader = CGWHttp::find_in_string(strPacket,"\r\n",false, 0) ;
    string requestUrl = pContext_->getRequestUrl() ;

    string strConnectAndUrl = "";
    if (iIndexSslURL_ == string::npos)
        strConnectAndUrl = requestUrl.substr(strlen("http://"), requestUrl.length() - strlen("http://")) ;
    else
        strConnectAndUrl = requestUrl.substr(strlen("https://"), requestUrl.length() - strlen("https://")) ;

    // get the host and port to connect on (begining of the url)
    int iIndex = CGWHttp::find_in_string(strConnectAndUrl,"/",false,0) ;
    string url ;
    if (iIndex == string::npos)
        url = "/";
    else
        url = strConnectAndUrl.substr(iIndex, strConnectAndUrl.length() - iIndex ) ;

    string request = pContext_->getMethod() ;
    request += " ";
    request += url ;
    request += " ";
    request += pContext_->getHttpVersion() ;
    memcpy(packet, request.c_str(), request.length());
    memcpy(&packet[request.length()], &packet[indexEndFirstHeader], iHeaderSize - indexEndFirstHeader + iBodySize);
    iHeaderSize = iHeaderSize - indexEndFirstHeader + request.length() ;
}

//-----------------------------------------------------------------------------
// CGWHttp::removeClientHeaders
//
// This functions remove all the HTTP cache headers
// ("if-modified-since:", "if-None-Match:") ;
//
// Arguments : header (in/out): HTTP header
// Return    : None
//-----------------------------------------------------------------------------
void CGWHttp::removeClientHeaders(char * packet, int & iHeaderSize, int iBodySize)
{
    // *********************************
    // To disable client-caching,
    // remove "if-modified-since" header
    // *********************************
    int iOldHeaderSize = iHeaderSize ;
    string strPacket ;
    int indexEndHeader = 0 ;

    strPacket = packet ;
    int indexIfModifiedSince = 0;
    if ((indexIfModifiedSince = CGWHttp::find_in_string(strPacket,"if-modified-since:",false, 0)) != string::npos)
    {
        string strThisHeader = strPacket.substr(indexIfModifiedSince, iHeaderSize + iBodySize - indexIfModifiedSince);
        indexEndHeader = CGWHttp::find_in_string(strThisHeader,"\r\n",false, 0) + 2;
        memcpy(&packet[indexIfModifiedSince], &packet[indexIfModifiedSince + indexEndHeader], iHeaderSize - indexIfModifiedSince - indexEndHeader) ;
        iHeaderSize -= indexEndHeader ;
    }

    // remove "if-None-Match" header
    strPacket = packet ;
    int indexIfNoneMatch = 0;
    if ((indexIfNoneMatch = CGWHttp::find_in_string(strPacket,"if-None-Match:",false, 0)) != string::npos)
    {
        string strThisHeader = strPacket.substr(indexIfNoneMatch, iHeaderSize + iBodySize - indexIfNoneMatch);
        indexEndHeader = CGWHttp::find_in_string(strThisHeader,"\r\n",false, 0) + 2;
        memcpy(&packet[indexIfNoneMatch], &packet[indexIfNoneMatch + indexEndHeader], iHeaderSize - indexIfNoneMatch - indexEndHeader) ;
        iHeaderSize -= indexEndHeader ;
    }

    bClientKeepAlive_ = FALSE ;
    if (pContext_->getHttpVersion() == "HTTP/1.1" && CGWHttp::find_in_string(strPacket, "Connection: Keep-Alive", false, 0) != string::npos)
        bClientKeepAlive_ = TRUE ;

    if (iOldHeaderSize != iHeaderSize)
    {
        memcpy(&packet[iHeaderSize], &packet[iOldHeaderSize], iBodySize) ;
    }
    packet[iHeaderSize + iBodySize] = 0;
}

//-----------------------------------------------------------------------------
// CGWHttp::replaceProxyClientHeaders
//
// For direct connection configurations (no proxy
// It replace Proxy-Connection header by Connection header
//
// Arguments : header (in/out): HTTP header
// Return    : None
//-----------------------------------------------------------------------------
void CGWHttp::replaceProxyClientHeaders(char * packet, int & iHeaderSize, int iBodySize)
{
    int iOldHeaderSize = iHeaderSize ;
    string strPacket = packet;
    // remove "Proxy-Connection" header and replace it by "Connection" header
    int indexProxyConnection = 0;
    if ((indexProxyConnection = CGWHttp::find_in_string(strPacket,"Proxy-Connection:",false, 0)) != string::npos)
    {
        int lenP = strlen("Proxy-");
        memcpy(&packet[indexProxyConnection], &packet[indexProxyConnection + lenP], iHeaderSize - indexProxyConnection - lenP) ;
        iHeaderSize -= lenP ;
    }
    if (iOldHeaderSize != iHeaderSize)
    {
        memcpy(&packet[iHeaderSize], &packet[iOldHeaderSize], iBodySize) ;
    }
    packet[iHeaderSize + iBodySize] = 0;
}

//-----------------------------------------------------------------------------
// CGWHttp::removeServerCacheHeaders
//
// This functions remove all the HTTP cache headers
// ("last-modified:") ;
//
// Arguments : header (in/out): HTTP header
// Return    : None
//-----------------------------------------------------------------------------
void CGWHttp::removeServerCacheHeaders(char * packet, int & iHeaderSize, int iBodySize)
{
    // *********************************
    // remove "last-modified" header
    // *********************************
    int iOldHeaderSize = iHeaderSize ;
    string strPacket ;
    int indexEndHeader = 0 ;

    strPacket = packet ;
    int indexLastModified = 0;
    if ((indexLastModified = CGWHttp::find_in_string(strPacket,"last-modified:",false, 0)) != string::npos)
    {
        string strThisHeader = strPacket.substr(indexLastModified, iHeaderSize - indexLastModified);
        indexEndHeader = CGWHttp::find_in_string(strThisHeader,"\r\n",false, 0) + 2;
        memcpy(&packet[indexLastModified], &packet[indexLastModified + indexEndHeader], iHeaderSize - indexLastModified - indexEndHeader) ;
        iHeaderSize -= indexEndHeader ;
    }

    if (iOldHeaderSize != iHeaderSize)
    {
        memcpy(&packet[iHeaderSize], &packet[iOldHeaderSize], iBodySize) ;
    }
    packet[iHeaderSize + iBodySize] = 0;
}

//-----------------------------------------------------------------------------
// CGWHttp::getClientHeaderInformations
//
// This functions extract informations for client HTTP headers
// informations are saved in CGWHttp members.
// . UserAgent
//
// Arguments : header (in): HTTP header
// Return    : None
//-----------------------------------------------------------------------------
void CGWHttp::getClientHeaderInformations(string header)
{
    int indexUserAgent = CGWHttp::find_in_string(header,"User-Agent:",false, 0);
    if (indexUserAgent != string::npos)
    {
		int iBeginUserAgent = header.find_first_not_of(" \t", indexUserAgent + 11) ;
		int iEndHeader = header.find_first_of("\r\n", indexUserAgent) ;
		if (iEndHeader != string::npos)
			userAgent_ = header.substr(iBeginUserAgent, iEndHeader - iBeginUserAgent) ;
		else
			userAgent_ = header.substr(iBeginUserAgent, header.size() - iBeginUserAgent);
	}
	pContext_->setPrimary(true) ;
	// Check is the request is a primary request.
	int indexAccept = CGWHttp::find_in_string(header,"Accept:",false, 0);
	if (indexAccept != string::npos)
	{
		string accept ;
		int iBeginAccept = header.find_first_not_of(" \t", indexAccept + 7) ;
		int iEndHeader = header.find_first_of("\r\n", indexAccept) ;
		if (iEndHeader != string::npos)
			accept = header.substr(iBeginAccept, iEndHeader - iBeginAccept) ;
		else
			accept = header.substr(iBeginAccept, header.size() - iBeginAccept);

		if (accept == "*/*")
			pContext_->setPrimary(false) ;
	}
}

//-----------------------------------------------------------------------------
// CGWHttp::getServerHeaderInformations
//
// This functions extract informations for server HTTP headers
// informations are saved in CGWHttp members.
// . status code
// . body length
// . content-type
// . transfer-coding (chunked or not)
// . Set-cookie
//
// Arguments : header (in): HTTP header
// Return    : None
//-----------------------------------------------------------------------------
void CGWHttp::getServerHeaderInformations(string header)
{
    //
    // retreive the status code
    //
    char httpVersion[20] ;
    sscanf(header.c_str(), "%s %s ", httpVersion, statusCode_) ;
    //
    // retreive the eventual body length
    //
    char * packet = (char * )header.c_str();
    int indexContentLength = 0;
    if ((indexContentLength = CGWHttp::find_in_string(header,"Content-Length:",false,0)) != string::npos)
    {
        iServerContentLength_ = atoi(&packet[indexContentLength] + strlen("Content-Length:")) ;
    }
    //
    // Check if the content-type is html/text
    // or javascript or ... ??
    //
    iServerIndexTextHtml_ = string::npos ;
    iServerIndexJavaScript_ = string::npos ;
    int indexContentType = CGWHttp::find_in_string(header,"Content-Type:",false, 0);
    if (indexContentType != string::npos)
    {
        iServerIndexTextHtml_ = CGWHttp::find_in_string(header,"text/html",false, indexContentType + strlen("Content-Type:")) ;
        iServerIndexJavaScript_ = CGWHttp::find_in_string(header,"application/x-javascript",false, indexContentType + strlen("Content-Type:")) ;
        if (iServerIndexJavaScript_ == string::npos)
            iServerIndexJavaScript_ = CGWHttp::find_in_string(header,"application/javascript",false, indexContentType + strlen("Content-Type:")) ;
        // We found the "text/html" string but it seems it is not the good one
        if (iServerIndexTextHtml_ != string::npos && iServerIndexTextHtml_ > indexContentType + (int)strlen("Content-Type:") + 5)
        {
            iServerIndexTextHtml_ = string::npos ;
        }
        if (iServerIndexJavaScript_ != string::npos && iServerIndexJavaScript_ > indexContentType + (int)strlen("Content-Type:") + 5)
        {
            iServerIndexJavaScript_ = string::npos ;
        }
    }
    //
    // Get the transfer-coding header to check if it is
    // a chunk transfer-coding.
    //
    iServerIndexChunked_ = string::npos ;
    int indexTransferCoding = CGWHttp::find_in_string(header,"Transfer-Encoding:",false, 0);
    if (indexTransferCoding != string::npos)
    {
        iServerIndexChunked_ = CGWHttp::find_in_string(header,"chunked",false, indexTransferCoding + strlen("Transfer-Encoding:")) ;
        if (iServerIndexChunked_ != string::npos && iServerIndexChunked_ > indexTransferCoding + (int)strlen("Transfer-Encoding:") + 5)
        {
            iServerIndexChunked_ = string::npos ;
        }
    }

    bServerConnectionClose_ = FALSE;
    if (CGWHttp::find_in_string(header,"Connection: Close", false, 0) != string::npos)
        bServerConnectionClose_ = TRUE;

    //
    // Get the Set-cookie headers
    //
    iServerIndexSetCookie_ = CGWHttp::find_in_string(header,"Set-Cookie:",false, 0);

    //
    // Get the Location headers
    //
    iServerIndexLocation_ = CGWHttp::find_in_string(header,"Location:",false, 0);
}

//-----------------------------------------------------------------------------
// CGWHttp::extractDocumentFromChunkedDatas
//
// This function get the document from a HTTP body transfered by chunks
//
// Arguments : body (in): HTTP body
// Return    : Document without chunked coding
//-----------------------------------------------------------------------------
char * CGWHttp::extractDocumentFromChunkedDatas(char * pBody, int &iDocumentLength)
{
    char * pIndexEndFirstHeader = strstr(pBody, "\r\n") ;
    unsigned int uiIndexEndFirstHeader = (unsigned int)pIndexEndFirstHeader - (unsigned int)pBody ;
    char * pFirstHeader = new char[uiIndexEndFirstHeader + 1] ;
    memcpy(pFirstHeader, pBody, uiIndexEndFirstHeader) ;
    pFirstHeader[uiIndexEndFirstHeader] = '\0' ;
    string strFirstHeader = (char *)pFirstHeader ;
    int endLength = strFirstHeader.find_first_of(" ", 0);
    string strLength = strFirstHeader.substr(0, endLength) ;

    int iLength = 0;
    sscanf(strLength.c_str(), "%x", &iLength) ;

    int iPosition = 0 ;
    int iTotalLength = 0 ;
    char * pDocument = NULL;
    char * previousDocument = NULL;
    int iSavedLengthReste = lSavedDocumentLen_ ;
    while (iLength > 0)
    {
        // check to see if a cancel has been received
        if (iSavedLengthReste < (int)uiIndexEndFirstHeader + 2 + iLength + 2 )
        {
            iLength = iSavedLengthReste ;
            bFullServerDocument_ = FALSE ;
        }
        else
            iSavedLengthReste -= (uiIndexEndFirstHeader + 2 + iLength + 2) ;
        previousDocument = pDocument ;
        pDocument = new char[iTotalLength + iLength] ;
        if (previousDocument != NULL)
        {
            memcpy(pDocument, previousDocument, iTotalLength) ;
        }
        memcpy(&pDocument[iTotalLength], &pBody[uiIndexEndFirstHeader + 2], iLength);
        iTotalLength += iLength ;

        // A cancel has been received
        if (!bFullServerDocument_)
            break ;
        pBody += uiIndexEndFirstHeader + 2 + iLength + 2 ;

        char * pIndexEndFirstHeader = strstr(pBody, "\r\n") ;
        if (!pIndexEndFirstHeader)
            break ;
        uiIndexEndFirstHeader = (unsigned int)pIndexEndFirstHeader - (unsigned int)pBody ;
        char * pFirstHeader = new char[uiIndexEndFirstHeader + 1] ;
        memcpy(pFirstHeader, pBody, uiIndexEndFirstHeader) ;
        pFirstHeader[uiIndexEndFirstHeader] = '\0' ;
        string strFirstHeader = (char *)pFirstHeader ;
        endLength = strFirstHeader.find_first_of(" ", 0);
        string strLength = strFirstHeader.substr(0, endLength) ;
        sscanf(strLength.c_str(), "%x", &iLength) ;
    }
    iDocumentLength = iTotalLength ;
    return pDocument ;
}

//-----------------------------------------------------------------------------
// CGWHttp::isLastChunk
//
// This function return TRUE if the chunked datas are fnished
//
// Arguments :
// Return    : TRUE/FALSE
//-----------------------------------------------------------------------------
bool CGWHttp::isLastChunk()
{
    char *pBody = savedDocument_ ;
    char * pIndexEndFirstHeader = strstr(pBody, "\r\n") ;
    if (!pIndexEndFirstHeader)
        return FALSE ;
    unsigned int uiIndexEndFirstHeader = (unsigned int)pIndexEndFirstHeader - (unsigned int)pBody ;
    char * pFirstHeader = new char[uiIndexEndFirstHeader + 1] ;
    memcpy(pFirstHeader, pBody, uiIndexEndFirstHeader) ;
    pFirstHeader[uiIndexEndFirstHeader] = '\0' ;
    string strFirstHeader = (char *)pFirstHeader ;
    int endLength = strFirstHeader.find_first_of(" ", 0);
    string strLength = strFirstHeader.substr(0, endLength) ;
    int iLength = 0;
    sscanf(strLength.c_str(), "%x", &iLength) ;

    int iPosition = 0 ;
    int iTotalLength = 0 ;
    char * pDocument = NULL;
    char * previousDocument = NULL;
    int iSavedLengthReste = lSavedDocumentLen_ ;
    while (iLength != 0)
    {
        // Not all the chunk
        if (iSavedLengthReste < (int)uiIndexEndFirstHeader + 2 + iLength + 2 )
        {
            return FALSE ;
        }
        else
            iSavedLengthReste -= (iLength + 2 + uiIndexEndFirstHeader + 2) ;

        pBody += uiIndexEndFirstHeader + 2 + iLength + 2 ;

        char * pIndexEndFirstHeader = strstr(pBody, "\r\n") ;
        if (!pIndexEndFirstHeader)
            return FALSE ;
        uiIndexEndFirstHeader = (unsigned int)pIndexEndFirstHeader - (unsigned int)pBody ;

        char * pFirstHeader = new char[uiIndexEndFirstHeader + 1] ;
        memcpy(pFirstHeader, pBody, uiIndexEndFirstHeader) ;
        pFirstHeader[uiIndexEndFirstHeader] = '\0' ;
        string strFirstHeader = (char *)pFirstHeader ;
        int endLength = strFirstHeader.find_first_of(" ", 0);
        string strLength = strFirstHeader.substr(0, endLength) ;
        sscanf(strLength.c_str(), "%x", &iLength) ;
    }
    return TRUE ;
}

//-----------------------------------------------------------------------------
// CGWHttp::generateChunkedDatas
//
// This function code a HTTP body with chunks.
//
// Arguments : body (in): HTTP body
// Return    : Document without chunked coding
//-----------------------------------------------------------------------------
char * CGWHttp::generateChunkedDatas(char * pBuf, int iBufLength, int & iLen)
{
    char * pChunkedBuf = new char [iBufLength + 100] ;
	memset(pChunkedBuf, 0, iBufLength + 100);

    sprintf(pChunkedBuf, "%x\r\n", iBufLength);
    iLen = strlen(pChunkedBuf) ;
	if (iBufLength > 0)
	{
		memcpy(&pChunkedBuf[iLen], pBuf, iBufLength) ;
		iLen += iBufLength ;
		memcpy(&pChunkedBuf[iLen], "\r\n", 2);
		iLen += 2 ;
	}
    return pChunkedBuf ;
}




//-----------------------------------------------------------------------------
// CGWHttp::saveDocumentAsFile
//
// save an HTML document, its HTTP header (.(c/s)hdr), an index file (.idx)
//
// Arguments : None
// Return    : None
//-----------------------------------------------------------------------------
void CGWHttp::saveDocumentAsFile()
{
    CTLockGuard <CTMutex>  guard(myMutex_);
    if (tofReference.getSelectStatus() == ESL_STOP || gDocumentPath == "")
        return ;
    string fullPath = "";

    string documentName = pContext_->getRequestUrl() ;
    char bufSpid[20] ;
    sprintf(bufSpid, "%d", iSpid_);
    documentName += bufSpid ;

    // open the All concatenate File (not use until GWHttpList::flushDocumentsAndConcatenate
    if (gFdAll == NULL)
    {
        fullPath = gDocumentPath ;
        fullPath += ".all" ;
        if ((gFdAll = fopen(fullPath.c_str(), "wb+")) == NULL)
        {
            char bufError[100] ;
            sprintf(bufError, "%d", GetLastError()) ;
            traceLine (string(bufError) + string ( (char*) ":Cannot open ") + fullPath) ;
            return ;
        }
    }
    // open body file
    if (gFdBody == NULL)
    {
        fullPath = gDocumentPath ;
        fullPath += ".body" ;
        if ((gFdBody = fopen(fullPath.c_str(), "wb+")) == NULL)
        {
            char bufError[100] ;
            sprintf(bufError, "%d", GetLastError()) ;
            traceLine (string(bufError) + string ( (char*) ":Cannot open ") + fullPath) ;
            return ;
        }
    }
    // open client headers file
    if (gFdClientHeader == NULL)
    {
        fullPath = gDocumentPath ;
        fullPath += ".chdr" ;
        if ((gFdClientHeader = fopen(fullPath.c_str(), "wb+")) == NULL)
        {
            char bufError[100] ;
            sprintf(bufError, "%d", GetLastError()) ;
            traceLine (string(bufError) + string ( (char*) ":Cannot open ") + fullPath) ;
            return ;
        }
    }
    // open server headers file
    if (gFdServerHeader == NULL)
    {
        fullPath = gDocumentPath ;
        fullPath += ".shdr" ;
        if ((gFdServerHeader = fopen(fullPath.c_str(), "wb+")) == NULL)
        {
            traceLine (string ( (char*) "Cannot open ") + fullPath) ;
            return ;
        }
    }
    // open index file
    if (gFdIndex == NULL)
    {
        fullPath = gDocumentPath ;
        fullPath += ".idx" ;
        if ((gFdIndex = fopen(fullPath.c_str(), "wb+")) == NULL)
        {
            char bufError[100] ;
            sprintf(bufError, "%d", GetLastError()) ;
            traceLine (string(bufError) + string ( (char*) ":Cannot open ") + fullPath) ;
            return ;
        }
    }

    char * document = savedDocument_ ;
    unsigned int uiDocLength = lSavedDocumentLen_ ;
    ;

    if (iServerIndexChunked_ != string::npos && savedDocument_ != NULL)
    {
        int iLength = 0 ;
        document = extractDocumentFromChunkedDatas(savedDocument_, iLength) ;
        uiDocLength = iLength ;
        if (savedDocument_)
        {
            delete [] savedDocument_ ;
        }
        savedDocument_ = document ;
    }

    if (iServerContentLength_ > 0 && iServerContentLength_ > lSavedDocumentLen_)
    {
        bFullServerDocument_ = FALSE ;
    }

    // Write the document in the document file
    unsigned int uiDocOffset = ftell(gFdBody) ;
    fwrite(&uiDocLength, sizeof(unsigned int), 1, gFdBody);
    unsigned int uiDocOffset2 = ftell(gFdBody) ;
	if (document)
		fwrite(document, uiDocLength, 1, gFdBody);
    uiDocOffset2 = ftell(gFdBody) ;

    // Write the server header in the server headers file
    // If some headers are finished by LF instead of CRLF, add a CR
    int iIndex0A = 0;
    while ((iIndex0A = serverHeader_.find("\n", iIndex0A)) != string::npos)
    {
        if (iIndex0A > 0 && serverHeader_[iIndex0A - 1] != '\r')
        {
            serverHeader_.insert(iIndex0A, "\r");
            iIndex0A ++;
        }
        iIndex0A ++;
    }
    unsigned int uiSrvHdrLength = serverHeader_.length() ;
    unsigned int uiSrvHdrOffset = ftell(gFdServerHeader) ;
    fwrite(&uiSrvHdrLength, sizeof(unsigned int), 1, gFdServerHeader);
    fwrite(serverHeader_.c_str(), uiSrvHdrLength, 1, gFdServerHeader);

    // Write the client header in the client headers file
    unsigned int uiCltHdrLength = clientHeader_.length() ;
    unsigned int uiCltHdrOffset = ftell(gFdClientHeader) ;
    fwrite(&uiCltHdrLength, sizeof(unsigned int), 1, gFdClientHeader);
    fwrite(clientHeader_.c_str(), uiCltHdrLength, 1, gFdClientHeader);

    // declare the new document in the index file.
    // Format : URL+conid+\0+docOffset+cltHdrOffset+srvHdrOffset
    string sclUrl = pContext_->getRequestUrl() ;
    sclUrl = CTofReference::modifyStringAsScl(sclUrl) ;
    fwrite(sclUrl.c_str(), sclUrl.length(), 1, gFdIndex);
    char c = '\0' ;
    fwrite(&c, 1, 1, gFdIndex);
    fwrite(&iSpid_, sizeof(unsigned int), 1, gFdIndex);
    fwrite(&uiDocOffset, sizeof(unsigned int), 1, gFdIndex);
    fwrite(&uiCltHdrOffset, sizeof(unsigned int), 1, gFdIndex);
    fwrite(&uiSrvHdrOffset, sizeof(unsigned int), 1, gFdIndex);

}


//-----------------------------------------------------------------------------
// CGWHttp::Concatenate
//
// concatenate all document saved in saveDocumentAsFile in 1 big file
// with extension .all
//
// Arguments : None
// Return    : None
//-----------------------------------------------------------------------------
// ===================================================================================================
// Capture file format:
// -------------------
//
// <LastConnId><BodyBase><ClientHeaderBase><ServerHeaderBase><Index><Body><ClientHeader><ServerHeader>
//
// where:
// <LastConnId> = last connection id used in this capture. 4 bytes integer.
// <BodyBase>         = offset of the beginning of the bodies part within the file.
// <ClientHeaderBase> = offset of the beginning of the client headers part within the file.
// <ServerHeaderBase> = offset of the beginning of the server headers part within the file.
// <Index> = For each couple (URL, connection id), contains the offsets to the body and client
//           and server headers of the HTTP query. Format: list of records of the
//           form: <URL\0><Connid><BodyOffset><ClientHeaderOffset><ServerHeaderOffset>
// <URL\0> = null-terminated ASCII string representing a canonical HTTP URL.
// <Connid> = connection id of the HTTP query
// <BodyOffset> = offset of the HTTP query's body (HTML file ou resource file: .gif, .jpg...)
//                from <BodyBase>.
// <ClientHeaderOffset> = offset of the HTTP query's client headers from <ClientHeaderBase>.
// <ServerHeaderOffset> = offset of the HTTP query's server headers from <ServerHeaderBase>.
// <Body> = Contains the bodies of HTTP requests. Format: list of records of the form:
//          <BodySize><QueryBody>
// <BodySize> = size of the query's body in bytes. 4 bytes integer.
// <QueryBody> = the actual body of the query.
// <ClientHeader> = Contains the client headers of HTTP requests. Format: list of records of the form:
//          <ClientHeaderSize><QueryClientHeader>
// <ClientHeaderSize> = size of the query's client headers string in bytes. 4 bytes integer.
// <QueryClientHeader> = the actual client headers string.
// <ServerHeader> = Contains the server headers of HTTP requests. Format: list of records of the form:
//          <ServerHeaderSize><QueryServerHeader>
// <ServerHeaderSize> = size of the query's server headers string in bytes. 4 bytes integer.
// <QueryServerHeader> = the actual server headers string.
//
// This format can be viewed as the following diagram:
//
//                      --------------
//                      | LastConnId |
//                      --------------
//               +------| BodyBase   |
//               |      --------------------
//               | +----| ClientHeaderBase |
//               | |    --------------------
//               | | +--| ServerHeaderBase |
//               | | |  -------------------------------------------------------------------------------------------------------
// Index:        | | |  | URL                            | \0 | Connid | BodyOffset | ClientHeaderOffset | ServerHeaderOffset |
//               | | |  -------------------------------------------------------------------------------------------------------
//               | | |     .
//               | | |     .
//               | | |     .
//               | | |  ------
//               | | |  | \0 |
//               | | |  ----------------------------------------------
// Body:         +----->| BodySize | QueryBody                       |
//                 | |  ----------------------------------------------
//                 | |      .
//                 | |      .
//                 | |      .
//                 | |  ----------------------------------------------
// Client Header:  +--->| ClientHeaderSize | QueryClientHeader       |
//                   |  ----------------------------------------------
//                   |      .
//                   |      .
//                   |      .
//                   |  ----------------------------------------------
// Server Header:    +->| ServerHeaderSize | QueryServerHeader       |
//                      ----------------------------------------------
//                          .
//                          .
//                          .
//
// ===================================================================================================
void CGWHttp::Concatenate(void)
{
    CTLockGuard <CTMutex> guard(CGWHttp::myMutex_);
    if (!gFdBody || !gFdClientHeader || !gFdServerHeader || !gFdIndex)
        return ;

    int iBodySize = ftell(gFdBody) ;
    int iChdrSize = ftell(gFdClientHeader) ;
    int iShdrSize = ftell(gFdServerHeader) ;
    int iIdxSize = ftell(gFdIndex) ;

    fclose(gFdBody) ;
    fclose(gFdClientHeader) ;
    fclose(gFdServerHeader) ;
    fclose(gFdIndex) ;
    gFdBody = gFdClientHeader = gFdServerHeader = gFdIndex = NULL ;

    FILE * fdClientHeader = NULL ;
    FILE * fdServerHeader = NULL ;
    FILE * fdIndex = NULL ;
    FILE * fdBody = NULL ;

    string fullPath = "" ;
    fullPath = gDocumentPath ;
    fullPath += ".chdr" ;
    if ((fdClientHeader = fopen(fullPath.c_str(), "rb")) == NULL)
    {
        return ;
    }
    fullPath = gDocumentPath ;
    fullPath += ".shdr" ;
    if ((fdServerHeader = fopen(fullPath.c_str(), "rb")) == NULL)
    {
        return ;
    }
    fullPath = gDocumentPath ;
    fullPath += ".body" ;
    if ((fdBody = fopen(fullPath.c_str(), "rb")) == NULL)
    {
        return ;
    }
    fullPath = gDocumentPath ;
    fullPath += ".idx" ;
    if ((fdIndex = fopen(fullPath.c_str(), "rb")) == NULL)
    {
        return ;
    }
    gDocumentPath = "" ;

    // + 1 for a added 0 at the end of the index file.
    fwrite(&giSpid, sizeof(unsigned int), 1, gFdAll);
    int iOffset = iIdxSize + 1 + (4 * sizeof(int));
    fwrite(&iOffset, sizeof(unsigned int), 1, gFdAll);
    iOffset += iBodySize ;
    fwrite(&iOffset, sizeof(unsigned int), 1, gFdAll);
    iOffset += iChdrSize ;
    fwrite(&iOffset, sizeof(unsigned int), 1, gFdAll);

    size_t iRead = 0 ;
    BYTE zBuf[512] ;

    while (!feof(fdIndex))
    {
        iRead = fread(zBuf, 1, sizeof(zBuf), fdIndex) ;
        fwrite(zBuf, iRead, 1, gFdAll) ;
    }
    char c = '\0';
    fwrite(&c, 1, 1, gFdAll) ;

    while (!feof(fdBody))
    {
        iRead = fread(zBuf, 1, sizeof(zBuf), fdBody) ;
        fwrite(zBuf, iRead, 1, gFdAll) ;
    }
    while (!feof(fdClientHeader))
    {
        iRead = fread(zBuf, 1, sizeof(zBuf), fdClientHeader) ;
        fwrite(zBuf, iRead, 1, gFdAll) ;
    }
    while (!feof(fdServerHeader))
    {
        iRead = fread(zBuf, 1, sizeof(zBuf), fdServerHeader) ;
        fwrite(zBuf, iRead, 1, gFdAll) ;
    }
    fclose(gFdAll) ;
    fclose(fdBody) ;
    fclose(fdClientHeader) ;
    fclose(fdServerHeader) ;
    fclose(fdIndex) ;
    gFdAll = NULL ;
}


//-----------------------------------------------------------------------------
// CGWHttp::parseHTMLandRemoveSSL
//
// Replace https:// urls by http://{ urls.
//
// Arguments : html : html document
//             bMopdified (out) : the document has been modified ?
// Return    : Document modified
//-----------------------------------------------------------------------------
string CGWHttp::parseHTMLandRemoveSSL(string html, bool & bModified)
{
    string newHtml = "";
    int iIndex = 0 ;
    int iNewIndex = string::npos ;
    string htmlTemp ;

    if (lEndDocumentLen_ > 0)
    {
        htmlTemp = endDocument_ ;
        html = htmlTemp + html ;
    }

    while ((iNewIndex = CGWHttp::find_in_string(html,"https://",false, iIndex)) != string::npos)
    {
        newHtml += html.substr(iIndex, iNewIndex - iIndex) ;
        newHtml += CYRANO_SSL_URL ;
        // We must calculate the length to add to content length header
        iIndex = iNewIndex + strlen("https://");
        bModified = TRUE ;
    }
    newHtml += html.substr(iIndex, html.length() - iIndex) ;

    if (lEndDocumentLen_ > 0)
    {
        newHtml = newHtml.substr(lEndDocumentLen_, newHtml.length() - lEndDocumentLen_) ;
    }
    return newHtml;
}

//-----------------------------------------------------------------------------
// CGWHttp::parseHTTPandRemoveSSL
//
// Replace https:// urls by http://{ urls.
//
// Arguments : packet : http header
//             iHeaderSize : size of the header
//-----------------------------------------------------------------------------
void CGWHttp::parseHTTPandRemoveSSL(char * packet, int & iHeaderSize, bool bSsl)
{
    char * header = new char [iHeaderSize + 1] ;

    strncpy(header, packet, iHeaderSize);
    header[iHeaderSize] = 0 ;

    string newHeader = "";
    int iIndex = 0 ;
    int iNewIndex = string::npos ;
    string headerTemp = header;
    delete [] header ;

    while ((iNewIndex = CGWHttp::find_in_string(headerTemp,"https://",false, iIndex)) != string::npos)
    {
        newHeader += headerTemp.substr(iIndex, iNewIndex - iIndex) ;
        newHeader += CYRANO_SSL_URL ;
        // We must calculate the length to add to content length header
        iIndex = iNewIndex + strlen("https://");
    }
    newHeader += headerTemp.substr(iIndex, headerTemp.length() - iIndex) ;

    memcpy(packet, newHeader.c_str(), newHeader.length());
    iHeaderSize = newHeader.length();
}

//-----------------------------------------------------------------------------
// CGWHttp::extractClientHeadersForSSL
//
// extract client headers to resend via wininet for SSL
//
// Arguments : pHeader : HTTP header
// Return    : header to send with wininet
//-----------------------------------------------------------------------------
string CGWHttp::extractClientHeadersForSSL(char * pHeader)
{
    string inHeaders[2] = {"Content-type:", "Cookie:"} ;
    string header = pHeader ;
    string newHeader = "" ;
    int iIndex = 0 ;
    int iLength = header.length() ;
    int iPos = 0 ;

    for (int i=0; i<2; i++)
    {
        iPos = 0 ;
        while ((iIndex = CGWHttp::find_in_string(header,inHeaders[i], false, iPos)) != string::npos)
        {
            if (header[(size_t)(iIndex - 1)] == '\n')
            {
                string unHeader = header.substr(iIndex, iLength - iIndex) ;
                int iIndexRet = CGWHttp::find_in_string(unHeader,"\r\n",false,0) ;
                newHeader += unHeader.substr(0, iIndexRet);
                newHeader += "\r\n" ;
                iPos += iIndex + iIndexRet ;
            }
        }
    }
    return newHeader ;
}

//-----------------------------------------------------------------------------
// CGWHttp::removeClientHeadersForSSL
//
// extract client headers to resend via wininet for SSL
//
// Arguments : pHeader : HTTP header
// Return    : header to send with wininet
//-----------------------------------------------------------------------------
string CGWHttp::removeClientHeadersForSSL(char * pHeader)
{
    string inHeaders[9] = {"Content-length:",
                           "Accept:",
                           "Accept-Language:",
                           "Proxy-Connection:",
                           "Pragma:",
                           "User-Agent:",
                           "Referer:",
                           "Accept-Encoding:",
                           "Host:"} ;
    string header = pHeader ;
    string newHeader = "" ;
    int iIndex = 0 ;
    int iPos = 0 ;
    int iLength = 0 ;

    int iretret = CGWHttp::find_in_string(header,"\r\n\r\n",false,0);
    header = header.substr(0, iretret+2) ;

    iLength = header.length() ;
    // For the moment, only remove the "Content-length:" header,
    // It is why we loop until i=1 (not 9)
    for (int i=0; i<1; i++)
    {
        string tmpinhead=inHeaders[i].substr(iPos, inHeaders[i].length());

        while ((iIndex = CGWHttp::find_in_string(header,tmpinhead,false, iPos)) != string::npos)
        {
            string unHeader = header.substr(iIndex, iLength - iIndex) ;
            int iIndexRet = CGWHttp::find_in_string(unHeader,"\r\n",false,0) ;
            newHeader += header.substr(iPos, iIndex - iPos) ;
            iPos += iIndex + iIndexRet + 2;
        }
        newHeader += header.substr(iPos, iLength - iPos);
        header = newHeader ;
        iPos = 0 ;
        newHeader = "" ;
        iLength = header.length() ;
    }
    newHeader += header.substr(iPos, header.length() - iPos);
    return newHeader ;
}

//-----------------------------------------------------------------------------
// CGWHttp::sendNormal
//
// send a request in non secure mode
//
// Arguments : host : host to send the request to
//             iPort : port number to connect on
//             packet : packet to send
//             iSize : size of the packet
//-----------------------------------------------------------------------------
void CGWHttp::sendNormal(const string & host, int iPort, char * packet, int iSize)
{
    if (previousHost_ != host || iPreviousPort_ != iPort)
    {
        if (pTarget_ /*&& pTarget_->valid()*/)
        {
            pTarget_->close() ;
        }
        try
        {
            pTarget_->connect(host, iPort) ;
        }
        catch(CTException &)
        {
            bFullServerDocument_ = FALSE ;
            throw;
        }
    }
    previousHost_ = host ;
    iPreviousPort_ = iPort ;
    previousProtocolType_ = "http" ;
    pTarget_->sendBuffer (packet, iSize, "", FALSE) ;
}

//-----------------------------------------------------------------------------
// CGWHttp::sendSSL
//
// send a request in secure mode (SSL)
//
// Arguments : header : HTTP header
//             pBody : pody of the request
//             iBodySize : body size
// Return    : TRUE/FALSE
//-----------------------------------------------------------------------------
HHOOK hh = NULL;
LRESULT CALLBACK CallWndProc(int code, WPARAM wParam, LPARAM lParam);
char gzCertName[128] ;

bool CGWHttp::sendSSL(string & header, char * pBody, int iBodySize)
{
    HINTERNET hOpen = NULL, hConnect  = NULL, hRequest = NULL;
    TCHAR szErrMsg[200];
    DWORD dwSize = 0;

    LPTSTR AcceptTypes[2] = {TEXT("*/*"), NULL};

	gzCertName[0] = 0 ;

    if (bClientKeepAlive_)
    {
        if (pTarget_)
        {
            pTarget_->close() ;
        }
        try
        {
            pTarget_->connect(pContext_->getServerName(), pContext_->getPort()) ;
        }
        catch(CTException &)
        {}

    }

    previousHost_ = pContext_->getServerName() ;
    iPreviousPort_ = pContext_->getPort() ;
    previousProtocolType_ = "https" ;

    if (pContext_->getThroughProxy())
    {
        char portBuf[10] ;
        sprintf(portBuf, ":%d", pContext_->getPort()) ;
        string proxy = pContext_->getServerName() ;
        proxy += portBuf ;
        hOpen = InternetOpen(_T(userAgent_.c_str()),INTERNET_OPEN_TYPE_PROXY  , proxy.c_str(), 0, 0);
    }
    else
        hOpen = InternetOpen(_T(userAgent_.c_str()),INTERNET_OPEN_TYPE_DIRECT  , NULL, 0, 0);
    if(hOpen)
    {

        int indexEndFirstHeader = CGWHttp::find_in_string(header,"\r\n",false, 0) ;
        string strFirstHeader = header.substr(0, indexEndFirstHeader) ;

        string strMethod = pContext_->getMethod() ;
        string strUrl = pContext_->getRequestUrl();
        string strVersion = pContext_->getHttpVersion();


        string strConnectAndUrl = strUrl.substr(strlen("https://"), strUrl.length() - strlen("https://")) ;

        // get the host and port to connect on (begining of the url)
        int iIndex = CGWHttp::find_in_string(strConnectAndUrl,"/",false,0) ;
        string strConnect = strConnectAndUrl.substr(0, iIndex );
		string originalConnect = strConnect ;
        int iPort = 443 ;
        int iIndexColon = CGWHttp::find_in_string(strConnect,":",false, 0);
        if (iIndexColon != string::npos)
        {
            string strPort = strConnect.substr(iIndexColon+1, strConnect.length() - iIndexColon - 1) ;
            iPort = atoi(strPort.c_str());
            strConnect = strConnect.substr(0, iIndexColon) ;
        }

        // get the document to load (end of the url)
        string strObject = "";
        string strParameters = "" ;
        strObject = strConnectAndUrl.substr(iIndex + 1, strConnectAndUrl.length() - iIndex - 1) ;


        char *lpHeadersA;
        DWORD dwBytesRead = 0;
        int iServerIndexTextHtml = string::npos ;
        int iGlobalSize = 0 ;

        if (!(hConnect = InternetConnect (
                             hOpen,
                             strConnect.c_str(),
                             iPort,
                             NULL, NULL,
                             INTERNET_SERVICE_HTTP,
                             0, 0)))
        {
            sprintf (szErrMsg, TEXT("%s: %x"), TEXT("InternetConnect Error"),
                     GetLastError());
            throw CTException(string(szErrMsg), __FILE__, __LINE__);
        }

		DWORD flag = INTERNET_FLAG_NO_AUTO_REDIRECT|
                     INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTPS|
                     INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTP|
					 INTERNET_FLAG_IGNORE_CERT_CN_INVALID |
					 INTERNET_FLAG_IGNORE_CERT_DATE_INVALID |
                     INTERNET_FLAG_RELOAD|
                     INTERNET_FLAG_SECURE|
                     INTERNET_FLAG_NO_COOKIES ; // to allow to send the real cookie

		if (!(hRequest = HttpOpenRequest (
                             hConnect,
                             TEXT(strMethod.c_str()),
                             strObject.c_str(),
                             TEXT(strVersion.c_str()),
                             NULL,
                             (LPCTSTR*)AcceptTypes,
                             flag,
                             0)))
        {
            sprintf (szErrMsg, TEXT("%s: %x"), TEXT("HttpOpenRequest Error"),
                     GetLastError());
            throw CTException(string(szErrMsg), __FILE__, __LINE__);
        }
		
		char * pHeaders = (char * )header.c_str() ;
		string newHeader = removeClientHeadersForSSL(&pHeaders[indexEndFirstHeader + 2]) ;
		DWORD dwError = 0 ;
		for(;;)
		{
			BOOL bRetSendRequest ;

			// Sends a request to the HTTP server.
			char *ph = NULL ;
			if (dwError != ERROR_INTERNET_INVALID_CA)
				ph = (char *)newHeader.c_str() ;
			
			if (strMethod == "POST")
				bRetSendRequest = HttpSendRequest (hRequest, ph, -1, pBody , pBody?iBodySize:0) ;
			else if (strMethod == "GET")
				bRetSendRequest = HttpSendRequest (hRequest, ph, -1, NULL, 0) ;
			if (bRetSendRequest)
				break ;
			dwError = GetLastError() ;
			if (dwError != ERROR_INTERNET_CLIENT_AUTH_CERT_NEEDED 
				&& dwError != ERROR_INTERNET_INVALID_CA
				&& dwError != ERROR_INTERNET_SEC_CERT_CN_INVALID)
			{
				sprintf (szErrMsg, TEXT("%s:%s %x"), TEXT("POST HttpSendRequest Error"), TEXT(strUrl.c_str()), dwError);
				traceLine (string(szErrMsg)) ;
				throw CTException(string(szErrMsg), __FILE__, __LINE__);
			}
			else
			{
				DWORD dRet ;
				LPVOID pData ;
				// Install a hook to get the certificate name
				hh = SetWindowsHookEx(WH_CALLWNDPROC, CallWndProc, GetModuleHandle(NULL), GetCurrentThreadId());
				if( (dRet = InternetErrorDlg( GetForegroundWindow(), 
									 hRequest,
									 dwError,
									 FLAGS_ERROR_UI_FILTER_FOR_ERRORS       |
									 FLAGS_ERROR_UI_FLAGS_GENERATE_DATA     |
									 FLAGS_ERROR_UI_FLAGS_CHANGE_OPTIONS, 
									 &pData)) != ERROR_SUCCESS )
				{
					dwError = GetLastError() ;
					sprintf (szErrMsg, TEXT("%s:%s %x"), TEXT("InternetErrorDlg Error"),TEXT(strUrl.c_str()), dwError);
					traceLine (string(szErrMsg)) ;
					throw CTException(string(szErrMsg), __FILE__, __LINE__);
				}
				// Uninstall the hook
				// The certificate name should be saved in the gzCertName variable.
                UnhookWindowsHookEx(hh);

				// add the client certificate in the certificate map.
				if (dwError == ERROR_INTERNET_CLIENT_AUTH_CERT_NEEDED)
				{
				    CTLockGuard <CTMutex> guard(myMutex_);
					if (gClientCertificateMap.find(originalConnect) == gClientCertificateMap.end())
					{
						gClientCertificateMap[originalConnect] = gzCertName ;
						//"Certificate file name (.PFX or .PEM), password='opensta'" ;
						pContext_->setClientCertificateName(gClientCertificateMap[originalConnect]);
	                    tofReference.writeRequestInTofFile(pContext_, clientHeader_, (iBodySize>0)?pBody:NULL, iBodySize);
					}
				}

				if (dwError != ERROR_INTERNET_CLIENT_AUTH_CERT_NEEDED
					&& dwError != ERROR_INTERNET_INVALID_CA
					&& dwError != ERROR_INTERNET_SEC_CERT_CN_INVALID)
					break ;
			}
		}

        HttpQueryInfo (hRequest, HTTP_QUERY_RAW_HEADERS_CRLF, NULL, (unsigned long *)&iServerHeaderSize_,
                       NULL);
        // Allocates a block of memory for lpHeadersA.
        lpHeadersA = new char [iServerHeaderSize_ * 2 + 1];
        memset(lpHeadersA, 0, iServerHeaderSize_ * 2 + 1);

        // Call HttpQueryInfo again to get the headers.
        if (!HttpQueryInfo (
                    hRequest,
                    HTTP_QUERY_RAW_HEADERS_CRLF,
                    (LPVOID) lpHeadersA, (unsigned long *)&iServerHeaderSize_, NULL))
        {
            sprintf (szErrMsg, TEXT("%s: %x"), TEXT("HttpQueryInfo Error"),
                     GetLastError());
            throw CTException(string(szErrMsg), __FILE__, __LINE__);
        }

		// trace the HTTP header
		{
			CTLockGuard <CTMutex> guard(myMutex_);
			char buf[50] ;
			sprintf(buf, "%d,%d", iSpid_, iServerHeaderSize_);
			traceLine (string ( (char*) "nsprecv" ) + string(":packet dump\t//spid ") + string(buf)) ;
			CTMemView memoryView ( (char*) lpHeadersA, iServerHeaderSize_, string ( "nsprecv:"), 8, " |", '.', "|") ;
			traceFlush (memoryView.getAscii () ) ;
		}
        //
        // extract the header informations
        //
        string strPacket = lpHeadersA ;
        getServerHeaderInformations(strPacket) ;
        if (iServerIndexSetCookie_ != string::npos)
            writeLoadResponseHeader(strPacket) ;

        int indexFirstLine = CGWHttp::find_in_string(strPacket,"\r\n",false,0) + 2 ;
        serverHeader_ = strPacket.substr(indexFirstLine, iServerHeaderSize_ - indexFirstLine) ;

        // remove cache headers before sending to the client
        int lSaveHeaderSize = iServerHeaderSize_ ;
        parseHTTPandRemoveSSL(lpHeadersA, iServerHeaderSize_, TRUE) ;
        removeServerCacheHeaders(lpHeadersA, iServerHeaderSize_, 0);

        // send the header to the client for other types than HTML
        if ( (iServerIndexTextHtml_ == string::npos ) && (iServerIndexJavaScript_ == string::npos) )
            pClient_->sendBuffer ((char *)lpHeadersA, iServerHeaderSize_ ,"", FALSE);
		
        do
        {
            DWORD dwBytes;
            if(InternetQueryDataAvailable(hRequest, &dwBytes, 0, 0))
            {
                if (dwBytes == 0)
				{
					// if in chunk mode, save the socument in file.
					// It is not done automatically because datas are not chunked 
					// but the iServerIndexChunked is not string::npos
                    if (iServerIndexChunked_ != string::npos)
					{
						int iTmp = iServerIndexChunked_  ;
						iServerIndexChunked_ = string::npos ;
						saveDocumentAsFile() ;
						iServerIndexChunked_ = iTmp ;
	                    if ( (iServerIndexTextHtml_ == string::npos) && (iServerIndexJavaScript_ == string::npos) )
                            pClient_->sendBuffer ("0\r\n\r\n", 5,"", FALSE);
					}
                    break ;
				}
                LPSTR pData = new char[dwBytes+1];
                if(InternetReadFile(hRequest, pData, dwBytes, &dwBytesRead))
                {
					// trace the HTTP body
					{
						CTLockGuard <CTMutex> guard(myMutex_);
						char buf[50] ;
						sprintf(buf, "%d,%d", iSpid_, dwBytesRead);
						traceLine (string ( (char*) "nsprecv" ) + string(":packet dump\t//spid ") + string(buf)) ;
						CTMemView memoryView ( (char*) pData, dwBytesRead, string ( "nsprecv:"), 8, " |", '.', "|") ;
						traceFlush (memoryView.getAscii () ) ;
					}
					saveDocumentInMemory((char *)pData, dwBytesRead, FALSE) ;
                    if ( (iServerIndexTextHtml_ == string::npos) && (iServerIndexJavaScript_ == string::npos) )
                    {
                        if (iServerIndexChunked_ != string::npos)
						{
							char * pChunkData = NULL ;
							int iChunkLen = 0 ;
							pChunkData = generateChunkedDatas(pData, dwBytesRead, iChunkLen) ;
                            pClient_->sendBuffer ((char *)pChunkData, iChunkLen,"", FALSE);
							delete [] pChunkData ;
						}
                        else
                            pClient_->sendBuffer ((char *)pData, dwBytesRead,"", FALSE);
                    }
                }
                delete [] pData;
            }
        }
        while(dwBytesRead);
        InternetCloseHandle(hRequest);
        InternetCloseHandle(hConnect);
        InternetCloseHandle(hOpen);

        if ( (iServerIndexTextHtml_ != string::npos) || (iServerIndexJavaScript_ != string::npos) )
        {
            serverHeader_ = lpHeadersA ;
            bool bModified = FALSE ;
            string newHtml = "" ;
            if (savedDocument_ != NULL)
                newHtml = parseHTMLandRemoveSSL(savedDocument_, bModified) ;
			else
				saveDocumentAsFile() ;
            pClient_->sendBuffer ((char *)lpHeadersA, iServerHeaderSize_,"", FALSE);
            if (iServerIndexChunked_ != string::npos)
            {
                int iNewLen = 0 ;
				char * pChunkData = NULL ;
				if (newHtml.length() > 0)
					pChunkData = generateChunkedDatas((char *)newHtml.c_str(), newHtml.length(), iNewLen) ;
                if (iNewLen > 0)
				{
                    pClient_->sendBuffer (pChunkData, iNewLen,"", FALSE);
				}
				pClient_->sendBuffer ("0\r\n\r\n", 5,"", FALSE);

                if (pChunkData)
                    delete [] pChunkData ;
            }
            else
            {
                if (newHtml.length() > 0)
                    pClient_->sendBuffer ((char *)newHtml.c_str(), newHtml.length(),"", FALSE);
            }
            iServerIndexChunked_ = string::npos ;
            bool bFullServerDocument_ = TRUE ;
        }
        if (bClientKeepAlive_ == FALSE || bServerConnectionClose_ == TRUE)
        {
            if (pTarget_)
            {
                pTarget_->close() ;
            }
            pTarget_ = NULL ;

			if (iServerContentLength_ == 0 && iServerIndexChunked_ == string::npos)
				saveDocumentAsFile() ;

			if (savedDocument_ != NULL)
				delete [] savedDocument_ ;
			savedDocument_ = NULL ;
			lSavedDocumentLen_ = 0 ;
			throw CTException(string("SSL Normal exception"), __FILE__, __LINE__);
        }
        if (savedDocument_ != NULL)
            delete [] savedDocument_ ;
        savedDocument_ = NULL ;
        lSavedDocumentLen_ = 0 ;
    }
    return TRUE ;
}


//-----------------------------------------------------------------------------
// CGWHttp::addCookies
//
// declare cookies for the gateway process
//
// Arguments : packet : web server packet reply
//             iHeaderSize : HTTP header size
// Return    :
//-----------------------------------------------------------------------------
void CGWHttp::addCookies(char *& packet, int & iHeaderSize)
{
    DWORD dwSize = 0;               // variable to get the buffer size needed
    DWORD dwSizeAcol = 0;           // variable to get the buffer size needed
    LPSTR lpszData = NULL;   // buffer to hold the cookie data
    LPSTR lpszDataAcol = NULL;   // buffer to hold the '{' cookie data
 
    string strPacket = packet;
    string strCurrentUrl = pContext_->getRequestUrl();

	// Get the cookie for URL without '{'
    if (InternetGetCookie(strCurrentUrl.c_str(), NULL, NULL, &dwSize) )
	{
		lpszData = new char[dwSize];
		InternetGetCookie(strCurrentUrl.c_str(), NULL, lpszData, &dwSize);
	}

	// Get the cookie for URL with '{'
	int iIndexColonSlashSlash = strCurrentUrl.find("://") ;
	if (iIndexColonSlashSlash != string::npos)
	{
		strCurrentUrl.insert(iIndexColonSlashSlash + 3, CYRANO_SSL_STR) ;
		if (InternetGetCookie(strCurrentUrl.c_str(), NULL, NULL, &dwSizeAcol))
		{
			lpszDataAcol = new char[dwSizeAcol];
			InternetGetCookie(strCurrentUrl.c_str(), NULL, lpszDataAcol, &dwSizeAcol);
		}
	}

	char * cookieAll = new char[dwSize + dwSizeAcol + 3] ;
	cookieAll[0] = 0;
	if (lpszData != NULL)
	{
		strcat(cookieAll, lpszData);
	}
	if (lpszDataAcol != NULL)
	{
		strcat(cookieAll, "; ");
		strcat(cookieAll, lpszDataAcol);
	}

	typedef set<string> stringSet ;
	stringSet cookieSet ;
	char * token = strtok(cookieAll, "; \t");
	while (token != NULL)
	{
		// Add the cookies in the stl set container.
		// Do not want to add the faked cookies used for the Automatic modelling
		// in the request sent to the web server
		// The faked cookies name begin by the string defined in
		// AUTOMATIC_COOKIE_PREFIX (see gwhttp.hpp)
		if (strncmp(token, AUTOMATIC_COOKIE_PREFIX, strlen(AUTOMATIC_COOKIE_PREFIX)) != 0)
			cookieSet.insert(token) ;
		token = strtok(NULL, "; \t");
	}
	

    int indexCookie  = CGWHttp::find_in_string(strPacket,"Cookie:",false, 0) ;

    string newCookieHeader ;
    int indexInsert = 0 ;
    if (indexCookie == string::npos)
    {
        indexInsert = iHeaderSize - 4;
		for (stringSet::iterator itr = cookieSet.begin(); itr != cookieSet.end(); itr++)
		{
			if (itr != cookieSet.begin())
				newCookieHeader += "; " ;
			else
				newCookieHeader = "\r\nCookie: ";
			newCookieHeader += *itr;
		}
    }
    else
    {
        string cookieHeader = strPacket.substr(indexCookie, strPacket.length() - indexCookie);
        indexInsert = CGWHttp::find_in_string(cookieHeader,"\r\n",false,0) + indexCookie;
		cookieHeader.resize(indexInsert - indexCookie) ;
		
		for (stringSet::iterator itr = cookieSet.begin(); itr != cookieSet.end(); itr++)

		{
			string token = *itr ;
			string strToken = token ;

			// Extract the form name= from the form name=value
			// this form is saved in token (put a 0 after = in token)
			int iEqual = token.find_first_of("=") ;
			if (iEqual != string::npos)
				token = token.substr(0, iEqual + 1) ;
	
			int iIndex = 0 ;
			// Search if the cookie is already declared.
			while ((iIndex = cookieHeader.find(token, iIndex)) != string::npos)
			{
				if (cookieHeader[iIndex - 1] == ' ' 
					|| cookieHeader[iIndex - 1] == '\t'
					|| cookieHeader[iIndex - 1] == ':'
					|| cookieHeader[iIndex - 1] == ';')
				{
					break ;
				}
				// Find the next cookie
				if ((iIndex = cookieHeader.find_first_of(";\r\n", iIndex)) == string::npos)
					break ;
			}
			// If the cookie was not found, we add it.
			if (iIndex == string::npos)
			{
				newCookieHeader += "; " ;
				newCookieHeader += strToken ;
			}
		}
    }

    strPacket.insert(indexInsert, newCookieHeader);

    delete [] packet;
    packet = new char[strPacket.length() + 1];

    memcpy(packet, strPacket.c_str(), strPacket.length());
    packet[strPacket.length()] = '\0';

    iHeaderSize += newCookieHeader.length();

	if (cookieAll)
		delete [] cookieAll;
	if (lpszData)
		delete [] lpszData;
	if (lpszDataAcol)
		delete [] lpszDataAcol;
}



///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////// End of Class CGWHttp //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////// Globals Functions //
///////////////////////////////////////////////////////////////////////////////

//================================================================< Functions >
extern "C"
{

    void DLLSPEC_GWHTTP LGW_NewGatewayParse (const void* pConnectionId, const string& taskName, bool& bTraceOn, CTOLogFileList& logFileList, CTSocket*& pClient, CTSocket*& pTarget, CGWParse*& pGWParse)
    {
        try
        {
            pGWParse = (CGWParse*) new CGWHttp (pConnectionId, taskName, bTraceOn, logFileList, pClient, pTarget) ;
            if (pGWParse == NULL)
            {   // Error from new.
                throw CGWParseException ("LGW_NewGatewayParse: new failed.", __FILE__, __LINE__) ;
            }
        }
        catch (CTException& e)
        {
            pGWParse = NULL ;
            throw CGWParseException ( (CGWParseException&) e) ;
        }
    } // LGW_NewGatewayParse

} // extern "C"


///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// End of Globals Functions //
///////////////////////////////////////////////////////////////////////////////


