/* Copyright (C) 2000-2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     Smp
//
// @module  SNMP Class | This module contains the implementation of the SNMP
//          class. This class allows the retrieval of SNMP Statistics.
//
///////////////////////////////////////////////////////////////////////////////

#include "SNMPMain.h"
#include "Smp.h"
#include "OstaTraceDef.h"
#include <OSTAORB/OSTAORB.h>

#ifdef WIN32
#include <winsock.h>
#endif

#include <CECSimpleParse.h>
#include <iostream>
#include "../../include/ucd-snmp-includes.h"

///////////////////////////////////////////////////////////////////////////////
// Globals.
#define TIMEOUT 5000            /* milliseconds */
#define RETRIES 3

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor initialises the member variables.
//
OtaSnmp::OtaSnmp()
{
    // Error information
    const char *where = "OtaSnmp::OtaSnmp";
	m_lastVal = 0;

#ifdef WIN32
    WORD      VersionRequested;
    WSADATA   stWSAData;

    VersionRequested = MAKEWORD(1, 1);
    // FIXME - need to check return value.
    // TODO  - need to check that WSACleanup is called
    //         one for every startup. Should be ok -
    //         constructor/destructor.
    WSAStartup(VersionRequested, &stWSAData);
#endif

    // Define environment variable specifying the location of MIB files
    // NB On Unix, putenv requires a static string.
    char     *pHome = getenv("OPENSTA_HOME");

    if (pHome == NULL)
    {
        LOG_MESSAGE(TRCCAT_RUNNING, TRCLEV_NONE, "OPENSTA_HOME undefined");
    }
    else
    {
        // note that the separator in MIBDIRS is a colon in UNIX,
        // but a semicolon under Windows, to avoid conflict
        // with "drive letters"

        // only set MIBDIRS if not already set
        if (getenv("MIBDIRS") == NULL)
        {
            string    setmibdir = "MIBDIRS="
                                  + string(pHome) + string("/Engines/Mon/mibs");
            static char putenvString[512];

            strcpy(putenvString, setmibdir.data());
            putenv(putenvString);
        }
    }

    // only set MIBS if not already set
    if (getenv("MIBS") == NULL)
    {
        // must be static
        static char *string="MIBS=all";

        putenv(string);
    }

    // snmp_set_do_debugging(1);
    // Initialize the SNMP library
    // TODO more sensible name?
    // void - no return to check
    init_snmp("snmpapp");
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor.
//
OtaSnmp::~OtaSnmp()
{
#ifdef WIN32
    WSACleanup();
#endif
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method executes the passed string as an SNMP Query.
//
// @rdesc   true if query successful.
//
// @parm    String representing the requested SNMP Query: agent/community/oid
// @parm    The result of the query.
//
bool OtaSnmp::RunSimpleQuery(const string query, long *result)
{
    char      msg[2560];
    long      value = 0;

    // Split up the query
    CECSimpleParse token(query);
    string agent = token.getNext(' ');
    string community = token.getNext(' ');
	int nPos = query.find_last_of(' ');
    string oidString = query.substr(nPos+1);

    // Initialize a "session" that defines who we're going to talk to
    struct snmp_session session;
    void *sessp;

    // void - no return to check
    snmp_sess_init(&session);

    session.version = SNMP_VERSION_1;
    session.peername = (char *) agent.data();
    session.community = (unsigned char *) community.data();
    session.community_len = strlen((const char *) session.community);

    // Open the session
    sessp = snmp_sess_open(&session);
    if (sessp == NULL)
    {
        char     *err;

        snmp_sess_error(sessp, NULL, NULL, &err);
        sprintf(msg, "snmp_sess_open failed: %s", err);
        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE, msg);
        free(err);
        return false;
    }

    // Create the PDU for the data for our request.
    struct snmp_pdu *pdu;

    pdu = snmp_pdu_create(SNMP_MSG_GET);

    // Convert the OID string representation to an internal representation.
    oid       reqOid[MAX_OID_LEN];
    size_t    reqOid_len = MAX_OID_LEN;

    if (  get_node(oidString.data(), reqOid, &reqOid_len) != 1)
    {
        // try get_node instead
        if ( read_objid(oidString.data(), reqOid, &reqOid_len) != 1)
        {
            char     *err;

            snmp_sess_error(sessp, NULL, NULL, &err);
            sprintf(msg, "both read_objid and get_node failed: %s", err);
            LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE, msg);
            free(err);
            return false;
        }
    }

    // Add this oid, with a NULL value, to the PDU
    snmp_add_null_var(pdu, reqOid, reqOid_len);

    // Send the request and get the response
    int       status;
    struct snmp_pdu *response;

    status = snmp_sess_synch_response(sessp, pdu, &response);
    if (status != STAT_SUCCESS)
    {
        char     *err;

        snmp_sess_error(sessp, NULL, NULL, &err);
        sprintf(msg, "snmp_sess_synch_response failed: %s", err);
        free(err);
        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE, msg);
        return false;
    }
    else if (response->errstat != SNMP_ERR_NOERROR)
    {
        sprintf(msg, "Error in packet: %s",
                snmp_errstring(response->errstat));
        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE, msg);
        return false;
    }

    struct variable_list *vars;
    int       count = 0;

    for (vars = response->variables; vars; vars = vars->next_variable)
    {
        count += 1;
        if (count > 1)
        {
            LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE,
                        "More than 1 value returned for simple query");
            return false;
        }

        // TODO sort out signed/unsigned
        if (vars->type == ASN_INTEGER
                || vars->type == ASN_COUNTER )
        {
            value = *(vars->val.integer);
        }
        else
        {
            LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE,
                        "Non-integer value returned for simple query");
            return false;
        }
    }

    // Free the response and close the SNMP session
    if (response)
    {
        // void - no return to check
        snmp_free_pdu(response);
    }

    if ( snmp_sess_close(sessp) != 1 )
    {
        char     *err;

        snmp_sess_error(sessp, NULL, NULL, &err);
        sprintf(msg, "snmp_sess_close failed: %s", err);
        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE, msg);
        free(err);
        return false;
    }

    // Return result
    *result = value;
	m_lastVal = value;
    return true;
}

long OtaSnmp::GetLastVal()
{
	return m_lastVal;
}